Package: rio
Type: Package
Title: A Swiss-Army Knife for Data I/O
Version: 1.3.0
Authors@R: c(person("Jason", "Becker", role = "aut", email = "jason@jbecker.co"),
             person("Chung-hong", "Chan", role = c("aut", "cre"), email = "chainsawtiney@gmail.com",
	     	     comment = c(ORCID = "0000-0002-6232-7530")),
             person("David", "Schoch", email = "david@schochastics.net", role = c("aut"),
                     comment = c(ORCID = "0000-0003-2952-4812")),
             person("Geoffrey CH", "Chan", role = "ctb", email = "gefchchan@gmail.com"),
             person("Thomas J.", "Leeper",
                    role = "aut", 
                    email = "thosjleeper@gmail.com",
                    comment = c(ORCID = "0000-0003-4097-6326")),
             person("Christopher", "Gandrud", role = "ctb"),
             person("Andrew", "MacDonald", role = "ctb"),
             person("Ista", "Zahn", role = "ctb"),
             person("Stanislaus", "Stadlmann", role = "ctb"),
             person("Ruaridh", "Williamson", role = "ctb", email = "ruaridh.williamson@gmail.com"),
             person("Patrick", "Kennedy", role = "ctb"),
             person("Ryan", "Price", email = "ryapric@gmail.com", role = "ctb"),
             person("Trevor L", "Davis", email = "trevor.l.davis@gmail.com", role = "ctb"),
             person("Nathan", "Day", email = "nathancday@gmail.com", role = "ctb"),
             person("Bill", "Denney",
                    email="wdenney@humanpredictions.com",
                    role="ctb",
                    comment=c(ORCID="0000-0002-5759-428X")),
             person("Alex", "Bokov", email = "alex.bokov@gmail.com", role = "ctb",
                    comment=c(ORCID="0000-0002-0511-9815")),
             person("Hugo", "Gruson", role = "ctb", comment = c(ORCID = "0000-0002-4094-1476")),
             person("Jacob", "Mears", role = "ctb"))
Description: Streamlined data import and export by making assumptions that
    the user is probably willing to make: 'import()' and 'export()' determine
    the data format from the file extension, reasonable defaults are used for
    data import and export, web-based import is natively supported (including
    from SSL/HTTPS), compressed files can be read directly, and fast import
    packages are used where appropriate. An additional convenience function,
    'convert()', provides a simple method for converting between file types.
URL: https://gesistsa.github.io/rio/, https://github.com/gesistsa/rio
BugReports: https://github.com/gesistsa/rio/issues
Depends: R (>= 4.0)
Imports: tools, stats, utils, foreign, haven (>= 1.1.2), curl (>= 0.6),
        data.table (>= 1.11.2), readxl (>= 0.1.1), tibble, writexl,
        lifecycle, R.utils (>= 2.13.0), readr
Suggests: datasets, bit64, testthat, knitr, magrittr, clipr, fst,
        hexView, jsonlite, pzfx, readODS (>= 2.1.0), rmarkdown, rmatio,
        xml2 (>= 1.2.0), yaml, arrow (>= 0.17.0), stringi, withr,
        nanoparquet, qs2 (>= 0.1.5)
License: GPL-2
VignetteBuilder: knitr
Encoding: UTF-8
RoxygenNote: 7.3.2
Config/Needs/website: gesistsa/tsatemplate
NeedsCompilation: no
Packaged: 2026-04-01 15:37:49 UTC; chainsawriot
Author: Jason Becker [aut],
  Chung-hong Chan [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6232-7530>),
  David Schoch [aut] (ORCID: <https://orcid.org/0000-0003-2952-4812>),
  Geoffrey CH Chan [ctb],
  Thomas J. Leeper [aut] (ORCID: <https://orcid.org/0000-0003-4097-6326>),
  Christopher Gandrud [ctb],
  Andrew MacDonald [ctb],
  Ista Zahn [ctb],
  Stanislaus Stadlmann [ctb],
  Ruaridh Williamson [ctb],
  Patrick Kennedy [ctb],
  Ryan Price [ctb],
  Trevor L Davis [ctb],
  Nathan Day [ctb],
  Bill Denney [ctb] (ORCID: <https://orcid.org/0000-0002-5759-428X>),
  Alex Bokov [ctb] (ORCID: <https://orcid.org/0000-0002-0511-9815>),
  Hugo Gruson [ctb] (ORCID: <https://orcid.org/0000-0002-4094-1476>),
  Jacob Mears [ctb]
Maintainer: Chung-hong Chan <chainsawtiney@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-02 07:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(.export,rio_arff)
S3method(.export,rio_clipboard)
S3method(.export,rio_csv)
S3method(.export,rio_csv2)
S3method(.export,rio_csvy)
S3method(.export,rio_dbf)
S3method(.export,rio_dta)
S3method(.export,rio_dump)
S3method(.export,rio_feather)
S3method(.export,rio_fods)
S3method(.export,rio_fst)
S3method(.export,rio_fwf)
S3method(.export,rio_html)
S3method(.export,rio_json)
S3method(.export,rio_matlab)
S3method(.export,rio_ods)
S3method(.export,rio_parquet)
S3method(.export,rio_psv)
S3method(.export,rio_pzfx)
S3method(.export,rio_qs2)
S3method(.export,rio_r)
S3method(.export,rio_rda)
S3method(.export,rio_rdata)
S3method(.export,rio_rds)
S3method(.export,rio_sas7bdat)
S3method(.export,rio_sav)
S3method(.export,rio_tsv)
S3method(.export,rio_txt)
S3method(.export,rio_xlsx)
S3method(.export,rio_xml)
S3method(.export,rio_xpt)
S3method(.export,rio_yml)
S3method(.export,rio_zsav)
S3method(.import,rio_arff)
S3method(.import,rio_clipboard)
S3method(.import,rio_csv)
S3method(.import,rio_csv2)
S3method(.import,rio_csvy)
S3method(.import,rio_dat)
S3method(.import,rio_dbf)
S3method(.import,rio_dif)
S3method(.import,rio_dta)
S3method(.import,rio_dump)
S3method(.import,rio_eviews)
S3method(.import,rio_feather)
S3method(.import,rio_fods)
S3method(.import,rio_fortran)
S3method(.import,rio_fst)
S3method(.import,rio_fwf)
S3method(.import,rio_html)
S3method(.import,rio_json)
S3method(.import,rio_matlab)
S3method(.import,rio_mtp)
S3method(.import,rio_ods)
S3method(.import,rio_parquet)
S3method(.import,rio_psv)
S3method(.import,rio_pzfx)
S3method(.import,rio_qs2)
S3method(.import,rio_r)
S3method(.import,rio_rda)
S3method(.import,rio_rdata)
S3method(.import,rio_rds)
S3method(.import,rio_rec)
S3method(.import,rio_sas7bdat)
S3method(.import,rio_sav)
S3method(.import,rio_spss)
S3method(.import,rio_syd)
S3method(.import,rio_tsv)
S3method(.import,rio_txt)
S3method(.import,rio_xls)
S3method(.import,rio_xlsm)
S3method(.import,rio_xlsx)
S3method(.import,rio_xml)
S3method(.import,rio_xpt)
S3method(.import,rio_yml)
S3method(.import,rio_zsav)
S3method(characterize,data.frame)
S3method(characterize,default)
S3method(factorize,data.frame)
S3method(factorize,default)
export(characterize)
export(convert)
export(export)
export(export_list)
export(factorize)
export(gather_attrs)
export(get_ext)
export(get_info)
export(import)
export(import_list)
export(install_formats)
export(show_unsupported_formats)
export(spread_attrs)

# Rio terminal

> tl;dr: Rio is a terminal built to run everywhere, as a native desktop applications by Rust or even in the browser powered by WebAssembly.

<img src="misc/logo.svg" alt="Rio terminal logo" width="320px" />

[![Packaging status](https://repology.org/badge/vertical-allrepos/rio-terminal.svg)](https://repology.org/project/rio-terminal/versions)

## Platforms

| Name | Details |
| --- | --- |
| MacOs _as desktop application_ | [Installation guide](https://raphamorim.io/rio/docs/install/macos/) |
| Linux _as desktop application_ | [Installation guide](https://raphamorim.io/rio/docs/install/linux/) |
| Windows _as desktop application_ | [Installation guide](https://raphamorim.io/rio/docs/install/windows/) |
| Web Browser _(WebAssembly)_ | (Sugarloaf is ready but Rio still need to be ported) |

Demo usage of Rio terminal on MacOS:

![Demo Rio on MacOS](docs/static/assets/posts/0.1.0/demo-rio.png)

## About

Documentation: https://raphamorim.io/rio

If you are using or want to help in any way please consider to donate via [Github Sponsors](https://github.com/sponsors/raphamorim).

Rio would not be possible without [few acknowledgements](#acknowledgments) and specially [Alacritty](https://github.com/alacritty/alacritty/), since a lot of Rio functionalities (e.g: ANSI parser, events, grid system) was originally written (and still uses a good amount) of Alacritty code.

## Supporting the Project

If you use and like Rio, please consider sponsoring it: your support helps to cover the fees required to maintain the project and to validate the time spent working on it!

* [![Sponsor Rio terminal](https://img.shields.io/github/sponsors/raphamorim?label=Sponsor%20Rio&logo=github&style=for-the-badge)](https://github.com/sponsors/raphamorim)

## Acknowledgments

- Alacritty 🥇
- Since version 0.0.22, Sugarloaf ported glyph-brush code which was originally written by @alexheretic and licensed under Apache-2.0 license 🥇
- Components text render was originally from https://github.com/hecrj/wgpu_glyph
- The legacy Rio logo was made using _Adobe Sketchbook_ on iPad (between versions 0.0.1 between 0.0.18).
- WA was built originally from a fork from [Macroquad](https://github.com/not-fl3/macroquad) which is licensed under MIT license.
- https://github.com/servo/core-foundation-rs/blob/d4ce710182f1756c9d874ab917283fe1a1b7a011/cocoa/src/appkit.rs#L1447

# Rio term

<img src="misc/osx/rio-logo.png" alt="Rio terminal logo" width="320px" />

> tl;dr: Rio is a terminal built to run everywhere, as a native desktop applications by Rust/WebGPU or even in the browser powered by WebAssembly/WebGPU.

[![Packaging status](https://repology.org/badge/vertical-allrepos/rio-terminal.svg)](https://repology.org/project/rio-terminal/versions)

## Platforms

| Name | Details |
| --- | --- |
| MacOs _as desktop application_ | [Installation guide](https://raphamorim.io/rio/docs/install/macos) |
| Linux _as desktop application_ | [Installation guide](https://raphamorim.io/rio/docs/install/linux) |
| Windows _as desktop application_ | [Installation guide](https://raphamorim.io/rio/docs/install/windows) |
| Web Browser _(WebAssembly)_ | (Sugarloaf is ready but Rio still need to be ported) |
| Nintendo Switch * | (development hasn't started) |

_* Nintendo Switch development is just for fun, the goal is to have  the renderer working and the basic features of a terminal._

## Demo Gallery

| ![Demo rio](docs/static/assets/posts/0.0.11/demo-rio.png) | ![Demo tmux](docs/static/assets/demos/demo-tmux.png) |
| ----------- | ----------- |
| ![Demo linux x11](docs/static/assets/posts/0.0.15/demo-navigation-x11.png) | ![Demo linux wayland](docs/static/assets/posts/0.0.15/demo-navigation-wayland.png) |
| ![Demo Windows 10](docs/static/assets/posts/0.0.8/demo-windows-10.png) |<img src="docs/static/assets/demos/demo-emacs.png" alt="Demo emacs" width="500px"/> |
| ![Demo native tabs macos](docs/static/assets/posts/0.0.17/demo-native-tabs.png) | ![Demo error handling](docs/static/assets/posts/0.0.19/demo-error-handling.png) |

Note: Emojis are rendered with Noto Emoji.

## Sugarloaf

Rio is built over a custom renderer called [Sugarloaf](https://crates.io/crates/sugarloaf), which is responsible for font and style rendering. Sugarloaf demo:

| ![Demo sugarloaf 1](sugarloaf/resources/demo-sugarloaf-1.png) | ![Demo Sugarloaf wasm](sugarloaf/resources/demo-wasm-1.png) |
| ----------- | ----------- |
| ![Demo sugarloaf 3](sugarloaf/resources/demo-sugarloaf-3.png) | ![Demo sugarloaf 4](sugarloaf/resources/demo-sugarloaf-4.png) |
| ![Demo sugarloaf 5](sugarloaf/resources/demo-sugarloaf-5.png) | ![Demo sugarloaf 6](sugarloaf/resources/demo-sugarloaf-6.png) |

## About

Documentation: https://raphamorim.io/rio

If you are using or want to help in any way please consider to donate via [Github Sponsors](https://github.com/sponsors/raphamorim).

Rio would not be possible without [few acknowledgements](#acknowledgements) and specially [Alacritty](https://github.com/alacritty/alacritty/), since a lot of Rio functionalities (e.g: ANSI parser, events, grid system) was originally written (and still uses a good amount) of Alacritty code.

## Supporting the Project

If you use and like Rio, please consider sponsoring it: your support helps to cover the fees required to maintain the project and to validate the time spent working on it!

* [![Sponsor Rio terminal](https://img.shields.io/github/sponsors/raphamorim?label=Sponsor%20Rio&logo=github&style=for-the-badge)](https://github.com/sponsors/raphamorim)
* [Patreon](https://patreon.com/raphamorim)

## Acknowledgments

- Alacritty 🥇
- Since version 0.0.22, Sugarloaf ported glyph-brush code which was originally written by @alexheretic and licensed under Apache-2.0 license 🥇
- Components text render was originally from https://github.com/hecrj/wgpu_glyph
- The legacy Rio logo was made using _Adobe Sketchbook_ on iPad (between versions 0.0.1 between 0.0.18).
- https://github.com/madsmtm/menubar/

# rio: A Swiss-Army Knife for Data I/O <img src="man/figures/logo.png" align="right" height="139"/>

<!-- badges: start -->

[![CRAN
Version](https://www.r-pkg.org/badges/version/rio)](https://cran.r-project.org/package=rio)
![Downloads](https://cranlogs.r-pkg.org/badges/rio) [![Codecov test
coverage](https://codecov.io/gh/gesistsa/rio/branch/main/graph/badge.svg)](https://app.codecov.io/gh/gesistsa/rio/tree/main)
<!-- badges: end -->

## Overview

The aim of **rio** is to make data file I/O in R as easy as possible by
implementing two main functions in Swiss-army knife style:

  - `import()` provides a painless data import experience by
    automatically choosing the appropriate import/read function based on
    file extension (or a specified `format` argument)
  - `export()` provides the same painless file recognition for data
    export/write functionality

## Installation

The package is available on
[CRAN](https://cran.r-project.org/package=rio) and can be installed
directly in R using `install.packages()`.

``` r
install.packages("rio")
```

The latest development version on GitHub can be installed using:

``` r
if (!require("remotes")){
    install.packages("remotes")
}
remotes::install_github("gesistsa/rio")
```

Optional: Installation of additional formats (see below: **Supported
file formats**)

``` r
library(rio)
install_formats()
```

## Usage

Because **rio** is meant to streamline data I/O, the package is
extremely easy to use. Here are some examples of reading, writing, and
converting data files.

### Import

Importing data is handled with one function, `import()`:

``` r
library("rio")
import("starwars.xlsx")
```

    ##                  Name homeworld species
    ## 1      Luke Skywalker  Tatooine   Human
    ## 2               C-3PO  Tatooine   Human
    ## 3               R2-D2  Alderaan   Human
    ## 4         Darth Vader  Tatooine   Human
    ## 5         Leia Organa  Tatooine   Human
    ## 6           Owen Lars  Tatooine   Human
    ## 7  Beru Whitesun lars   Stewjon   Human
    ## 8               R5-D4  Tatooine   Human
    ## 9   Biggs Darklighter  Kashyyyk Wookiee
    ## 10     Obi-Wan Kenobi  Corellia   Human

``` r
import("starwars.csv")
```

    ##                  Name homeworld species
    ## 1      Luke Skywalker  Tatooine   Human
    ## 2               C-3PO  Tatooine   Human
    ## 3               R2-D2  Alderaan   Human
    ## 4         Darth Vader  Tatooine   Human
    ## 5         Leia Organa  Tatooine   Human
    ## 6           Owen Lars  Tatooine   Human
    ## 7  Beru Whitesun lars   Stewjon   Human
    ## 8               R5-D4  Tatooine   Human
    ## 9   Biggs Darklighter  Kashyyyk Wookiee
    ## 10     Obi-Wan Kenobi  Corellia   Human

### Export

Exporting data is handled with one function, `export()`:

``` r
export(mtcars, "mtcars.csv") # comma-separated values
export(mtcars, "mtcars.rds") # R serialized
export(mtcars, "mtcars.sav") # SPSS
```

A particularly useful feature of rio is the ability to import from and
export to compressed archives (e.g., zip), saving users the extra step
of compressing a large exported file, e.g.:

``` r
export(mtcars, "mtcars.tsv.zip")
```

`export()` can also write multiple data frames to respective sheets of
an Excel workbook or an HTML file:

``` r
export(list(mtcars = mtcars, iris = iris), file = "mtcars.xlsx")
```

## Supported file formats

**rio** supports a wide range of file formats. To keep the package slim,
several formats are supported via “Suggests” packages, which are not
installed (or loaded) by default. You can check which formats are
**not** supported via:

``` r
show_unsupported_formats()
```

You can install the suggested packages individually, depending your own
needs. If you want to install all suggested packages:

``` r
install_formats()
```

The full list of supported formats is below:

| Name                                | Extensions / “format”               | Import Package | Export Package | Type    | Note                           |
| :---------------------------------- | :---------------------------------- | :------------- | :------------- | :------ | :----------------------------- |
| Archive files (handled by tar)      | tar / tar.gz / tgz / tar.bz2 / tbz2 | utils          | utils          | Default |                                |
| Bzip2                               | bz2 / bzip2                         | base           | base           | Default |                                |
| Gzip                                | gz / gzip                           | base           | base           | Default |                                |
| Zip files                           | zip                                 | utils          | utils          | Default |                                |
| Ambiguous file format               | dat                                 | data.table     |                | Default | Attempt as delimited text data |
| CSVY (CSV + YAML metadata header)   | csvy                                | data.table     | data.table     | Default |                                |
| Comma-separated data                | csv                                 | data.table     | data.table     | Default |                                |
| Comma-separated data (European)     | csv2                                | data.table     | data.table     | Default |                                |
| Data Interchange Format             | dif                                 | utils          |                | Default |                                |
| Epiinfo                             | epiinfo / rec                       | foreign        |                | Default |                                |
| Excel                               | excel / xlsx                        | readxl         | writexl        | Default |                                |
| Excel (Legacy)                      | xls                                 | readxl         |                | Default |                                |
| Excel (Read only)                   | xlsm / xltx / xltm                  | readxl         |                | Default |                                |
| Fixed-width format data             | fwf                                 | readr          | utils          | Default |                                |
| Fortran data                        | fortran                             | utils          |                | Default | No recognized extension        |
| Google Sheets                       | googlesheets                        | data.table     |                | Default | As comma-separated data        |
| Minitab                             | minitab / mtp                       | foreign        |                | Default |                                |
| Pipe-separated data                 | psv                                 | data.table     | data.table     | Default |                                |
| R syntax                            | r                                   | base           | base           | Default |                                |
| SAS                                 | sas / sas7bdat                      | haven          | haven          | Default | Export is deprecated           |
| SAS XPORT                           | xport / xpt                         | haven          | haven          | Default |                                |
| SPSS                                | sav / spss                          | haven          | haven          | Default |                                |
| SPSS (compressed)                   | zsav                                | haven          | haven          | Default |                                |
| SPSS Portable                       | por                                 | haven          |                | Default |                                |
| Saved R objects                     | rda / rdata                         | base           | base           | Default |                                |
| Serialized R objects                | rds                                 | base           | base           | Default |                                |
| Stata                               | dta / stata                         | haven          | haven          | Default |                                |
| Systat                              | syd / systat                        | foreign        |                | Default |                                |
| Tab-separated data                  | / tsv / txt                         | data.table     | data.table     | Default |                                |
| Text Representations of R Objects   | dump                                | base           | base           | Default |                                |
| Weka Attribute-Relation File Format | arff / weka                         | foreign        | foreign        | Default |                                |
| XBASE database files                | dbf                                 | foreign        | foreign        | Default |                                |
| Apache Arrow (Parquet)              | parquet                             | nanoparquet    | nanoparquet    | Suggest |                                |
| Clipboard                           | clipboard                           | clipr          | clipr          | Suggest | default is tsv                 |
| EViews                              | eviews / wf1                        | hexView        |                | Suggest |                                |
| Fast Storage                        | fst                                 | fst            | fst            | Suggest |                                |
| Feather R/Python interchange format | feather                             | arrow          | arrow          | Suggest |                                |
| Graphpad Prism                      | pzfx                                | pzfx           | pzfx           | Suggest |                                |
| HTML Tables                         | htm / html                          | xml2           | xml2           | Suggest |                                |
| JSON                                | json                                | jsonlite       | jsonlite       | Suggest |                                |
| Matlab                              | mat / matlab                        | rmatio         | rmatio         | Suggest |                                |
| OpenDocument Spreadsheet            | ods                                 | readODS        | readODS        | Suggest |                                |
| OpenDocument Spreadsheet (Flat)     | fods                                | readODS        | readODS        | Suggest |                                |
| Serialized R objects (qs2)          | qs2                                 | qs2            | qs2            | Suggest |                                |
| Shallow XML documents               | xml                                 | xml2           | xml2           | Suggest |                                |
| YAML                                | yaml / yml                          | yaml           | yaml           | Suggest |                                |

Additionally, any format that is not supported by **rio** but that has a
known R implementation will produce an informative error message
pointing to a package and import or export function. Unrecognized
formats will yield a simple “Unrecognized file format” error.

## Other functions

### Convert

The `convert()` function links `import()` and `export()` by constructing
a dataframe from the imported file and immediately writing it back to
disk. `convert()` invisibly returns the file name of the exported file,
so that it can be used to programmatically access the new file.

``` r
convert("mtcars.sav", "mtcars.dta")
```

It is also possible to use **rio** on the command-line by calling
`Rscript` with the `-e` (expression) argument. For example, to convert a
file from Stata (.dta) to comma-separated values (.csv), simply do the
following:

    Rscript -e "rio::convert('iris.dta', 'iris.csv')"

### \*\_list

`import_list()` allows users to import a list of data frames from a
multi-object file (such as an Excel workbook, .Rdata file, zip
directory, or HTML file):

``` r
str(m <- import_list("mtcars.xlsx"))
```

    ## List of 2
    ##  $ mtcars:'data.frame':  32 obs. of  11 variables:
    ##   ..$ mpg : num [1:32] 21 21 22.8 21.4 18.7 18.1 14.3 24.4 22.8 19.2 ...
    ##   ..$ cyl : num [1:32] 6 6 4 6 8 6 8 4 4 6 ...
    ##   ..$ disp: num [1:32] 160 160 108 258 360 ...
    ##   ..$ hp  : num [1:32] 110 110 93 110 175 105 245 62 95 123 ...
    ##   ..$ drat: num [1:32] 3.9 3.9 3.85 3.08 3.15 2.76 3.21 3.69 3.92 3.92 ...
    ##   ..$ wt  : num [1:32] 2.62 2.88 2.32 3.21 3.44 ...
    ##   ..$ qsec: num [1:32] 16.5 17 18.6 19.4 17 ...
    ##   ..$ vs  : num [1:32] 0 0 1 1 0 1 0 1 1 1 ...
    ##   ..$ am  : num [1:32] 1 1 1 0 0 0 0 0 0 0 ...
    ##   ..$ gear: num [1:32] 4 4 4 3 3 3 3 4 4 4 ...
    ##   ..$ carb: num [1:32] 4 4 1 1 2 1 4 2 2 4 ...
    ##  $ iris  :'data.frame':  150 obs. of  5 variables:
    ##   ..$ Sepal.Length: num [1:150] 5.1 4.9 4.7 4.6 5 5.4 4.6 5 4.4 4.9 ...
    ##   ..$ Sepal.Width : num [1:150] 3.5 3 3.2 3.1 3.6 3.9 3.4 3.4 2.9 3.1 ...
    ##   ..$ Petal.Length: num [1:150] 1.4 1.4 1.3 1.5 1.4 1.7 1.4 1.5 1.4 1.5 ...
    ##   ..$ Petal.Width : num [1:150] 0.2 0.2 0.2 0.2 0.2 0.4 0.3 0.2 0.2 0.1 ...
    ##   ..$ Species     : chr [1:150] "setosa" "setosa" "setosa" "setosa" ...

`export_list()` makes it easy to export a list of (possibly named) data
frames to multiple files:

``` r
export_list(m, "%s.tsv")
c("mtcars.tsv", "iris.tsv") %in% dir()
```

    ## [1] TRUE TRUE

## Other projects

### GUIs

  - [**datamods**](https://cran.r-project.org/package=datamods) provides
    Shiny modules for importing data via `rio`.
  - [**rioweb**](https://github.com/lbraglia/rioweb) that provides
    access to the file conversion features of `rio`.
  - [**GREA**](https://github.com/Stan125/GREA/) is an RStudio add-in
    that provides an interactive interface for reading in data using
    `rio`.

### Similar packages

  - [**reader**](https://cran.r-project.org/package=reader) handles
    certain text formats and R binary files
  - [**io**](https://cran.r-project.org/package=io) offers a set of
    custom formats
  - [**ImportExport**](https://cran.r-project.org/package=ImportExport)
    focuses on select binary formats (Excel, SPSS, and Access files) and
    provides a Shiny interface.
  - [**SchemaOnRead**](https://cran.r-project.org/package=SchemaOnRead)
    iterates through a large number of possible import methods until one
    works successfully

<!-- LOGO -->
<h1>
<p align="center">
  <img src="https://rioterm.com/assets/rio-logo.png" alt="Rio terminal logo" width="128">
  <br>Rio Terminal
</h1>
  <p align="center">
    Rio is a modern terminal built to run everywhere.
    <br />
    <a href="#about">About</a>
    ·
    <a href="https://rioterm.com/docs/install">Install</a>
    ·
    <a href="https://rioterm.com/docs/config">Config</a>
    ·
    <a href="https://rioterm.com/changelog">Changelog</a>
    ·
    <a href="https://github.com/sponsors/raphamorim">Sponsor</a>
  </p>
</p>

Documentation: [rioterm.com](https://rioterm.com).

## Supporting the Project

If you use and like Rio, please consider sponsoring it: your support helps to cover the fees required to maintain the project and to validate the time spent working on it!

[![Sponsor Rio terminal](https://img.shields.io/github/sponsors/raphamorim?label=Sponsor%20Rio&logo=github&style=for-the-badge)](https://github.com/sponsors/raphamorim)

## Packaging

[![Packaging status](https://repology.org/badge/vertical-allrepos/rio-terminal.svg?columns=3)](https://repology.org/project/rio-terminal/versions)

> Demo with split and CRT on MacOS

![Demo Rio 0.2.0 on MacOS](https://rioterm.com/assets/posts/0.2.0/demo-rio.png)

> Demo with blurred background on Linux

![Demo blurred background](https://rioterm.com/assets/demos/demos-nixos-blur.png)

> Demo of Rio running on a Steam Deck

![Demo of Rio running on a Steam Deck](https://rioterm.com/assets/demos/demo-flatpak-steamdeck.jpg)

## Minimal stable rust version

Rio's MSRV is 1.96.0.

<!-- LOGO -->
<h1>
<p align="center">
  <img src="https://rioterm.com/assets/rio-logo.png" alt="Rio terminal logo" width="128">
  <br>Rio Terminal
</h1>
  <p align="center">
    Rio is a modern terminal built to run everywhere.
    <br />
    <a href="#about">About</a>
    ·
    <a href="https://rioterm.com/docs/install">Install</a>
    ·
    <a href="https://rioterm.com/docs/config">Config</a>
    ·
    <a href="https://rioterm.com/changelog">Changelog</a>
    ·
    <a href="https://github.com/sponsors/raphamorim">Sponsor</a>
  </p>
</p>

Documentation: [rioterm.com](https://rioterm.com).

## Supporting the Project

If you use and like Rio, please consider sponsoring it: your support helps to cover the fees required to maintain the project and to validate the time spent working on it!

[![Sponsor Rio terminal](https://img.shields.io/github/sponsors/raphamorim?label=Sponsor%20Rio&logo=github&style=for-the-badge)](https://github.com/sponsors/raphamorim)

## Packaging

[![Packaging status](https://repology.org/badge/vertical-allrepos/rio-terminal.svg?columns=3)](https://repology.org/project/rio-terminal/versions)

> Demo with split and CRT on MacOS

![Demo Rio 0.2.0 on MacOS](https://rioterm.com/assets/posts/0.2.0/demo-rio.png)

> Demo with blurred background on Linux

![Demo blurred background](https://rioterm.com/assets/demos/demos-nixos-blur.png)

> Demo of Rio running on a Steam Deck

![Demo of Rio running on a Steam Deck](https://rioterm.com/assets/demos/demo-flatpak-steamdeck.jpg)

## Minimal stable rust version

Rio's MSRV is 1.96.1.

# Corcovado

Corcovado is a maintained fork of mio 0.6.x along mio-signal-hook, mio-extras and using Windows API that works in Windows 11. It uses Rust edition 2021 instead of 2018.

Corcovado also uses Rust standard library for net and io instead of Mio 0.6.x.

# librio

Embeddable terminal core extracted from [Rio](https://github.com/raphamorim/rio):
PTY spawning, VT parsing, terminal state, and a host-pulled render-state API
with per-row dirty tracking. No drawing or windowing code — bring your own
renderer.

`librio` wraps [`rio-vt`](../rio-vt) as a C ABI for non-Rust consumers (Swift,
C, and anything that speaks C). It is **not published to crates.io**: it lives
in the Rio source tree and ships as the `RioKit.xcframework` static library in
Rio's GitHub releases. If you write Rust, depend on `rio-vt` directly.

The Rust surface it exposes over the C boundary:

```rust
use librio::{Engine, RenderState, SurfaceDelegate, SurfaceDesc};
use std::sync::Arc;

struct Delegate;
impl SurfaceDelegate for Delegate {
    fn wakeup(&self, _surface: usize) { /* schedule a render */ }
}

let engine = Engine::new(Arc::new(Delegate));
let surface = engine.create_surface(&SurfaceDesc::default())?;
let mut state = RenderState::new(&surface);
surface.text("ls\r");
state.update();
```

The C ABI is compiled by default (`crate-type = ["staticlib"]`); see
`librio/include` for the curated header and modulemap, and the `Makefile`
`librio-xcframework` target for packaging.

# rio-vt

Rio's embeddable terminal core: the VT state machine, ANSI/escape parser,
grid with scrollback, selection, search, and PTY-facing event model,
extracted from [Rio](https://github.com/raphamorim/rio) as a standalone,
dependency-light crate.

It carries **no rendering, GPU, or font-shaping dependencies**. You pair
it with a renderer of your choice (Rio uses `sugarloaf`) or drive it purely
from the pulled grid state. It is the safe-Rust sibling of
[`librio`](../librio), which wraps this same core in a C ABI for
embedding from Swift, C, Go (cgo), and other languages.

- **Rust embedders** depend on `rio-vt` directly (this crate).
- **Non-Rust embedders** link `librio` (C ABI) + its generated header.

## Add it

```toml
[dependencies]
rio-vt = "0.5"
```

`rio-vt` is lean by default. The `crosswords` VT logic, the parser, grid,
selection and search all compile with no windowing or renderer stack.

### Feature flags

| Feature | Default | Effect |
| --- | --- | --- |
| _(none)_ | ✓ | Lean core. No renderer, no windowing. |
| `rio-window` | | `From` conversions between the core's `WindowId` and `rio_window::window::WindowId`, plus the event-loop proxy listener. |
| `renderer` | | Expose the app-level `RioErrorType::FontsNotFound` variant (pulls a `sugarloaf` font type). Enabled transitively by `rio-backend`'s renderer. |
| `x11` / `wayland` | | Forward the matching clipboard backend feature. |

`WindowId` is **always** the core's own `WindowId(u64)` (with
`From<u64>`/`Into<u64>`); `rio-vt` never aliases it to the windowing
layer's type. The `rio-window` feature only adds the boundary `From`
conversions.

## Quick start (headless)

Create a grid, feed it bytes through the parser, and read cells back. No
PTY, no renderer, no threads:

```rust
use rio_vt::ansi::CursorShape;
use rio_vt::crosswords::grid::Dimensions;
use rio_vt::crosswords::pos::Column;
use rio_vt::crosswords::{Crosswords, CrosswordsSize};
use rio_vt::event::{VoidListener, WindowId};
use rio_vt::performer::handler::Processor;

// An 80x24 terminal with 10k lines of scrollback.
let size = CrosswordsSize::new(80, 24);
let cols = size.columns();
let mut term = Crosswords::new(
    size,
    CursorShape::Block,
    VoidListener,          // ignore events; swap for your own EventListener
    WindowId::from(0),
    0,                     // route id
    10_000,                // scrollback limit
);

// Feed raw PTY bytes (SGR red "hello", reset, then " world").
let mut parser = Processor::default();
parser.advance(&mut term, b"\x1b[31mhello\x1b[0m world");

// Pull the visible grid and read the first row's text.
let rows = term.visible_rows();
let line: String = (0..cols).map(|x| rows[0][Column(x)].c()).collect();
assert!(line.starts_with("hello world"));
```

> The full, runnable version of this snippet lives at
> [`examples/quickstart.rs`](examples/quickstart.rs)
> (`cargo run -p rio-vt --example quickstart`).

## Examples

Each is runnable with `cargo run -p rio-vt --example <name>`:

| Example | Shows |
| --- | --- |
| [`quickstart`](examples/quickstart.rs) | Create a grid, feed bytes, read cells back. |
| [`sixel`](examples/sixel.rs) | Decode a Sixel image, captured via `RioEvent::UpdateGraphics`. |
| [`kitty_image`](examples/kitty_image.rs) | Transmit a Kitty graphics-protocol image and capture it. |
| [`selection`](examples/selection.rs) | Select a range and read it back as text. |
| [`resize`](examples/resize.rs) | Resize the terminal and watch the grid follow. |

The `sixel` and `kitty_image` examples double as a reference for the
event-driven graphics flow: received images are drained into a
`RioEvent::UpdateGraphics` event (with `pending` for Sixel/iTerm2 and
`pending_images` for Kitty), which a frontend records through its
`EventListener`.

## Receiving events

Instead of `VoidListener`, implement `EventListener` to react to what the
terminal emits (PTY writes, bell, title changes, clipboard requests, and so
on). `event()` is the only required method; override the `send_*` hooks you
care about:

```rust
use rio_vt::event::{EventListener, RioEvent, WindowId};

#[derive(Clone)]
struct MyListener;

impl EventListener for MyListener {
    // Required: return a pending event (used by pull-style consumers).
    fn event(&self) -> (Option<RioEvent>, bool) {
        (None, false)
    }

    // Optional: react to events the terminal pushes out.
    fn send_event(&self, event: RioEvent, _window: WindowId) {
        match event {
            RioEvent::PtyWrite(_route_id, text) => { /* write `text` to the PTY */ }
            RioEvent::Title(title) => { /* update the window title */ }
            RioEvent::Bell => { /* ring */ }
            _ => {}
        }
    }
}
```

## Driving a real PTY

`performer::Machine` wires a spawned PTY to the parser and the grid on a
reader thread, so bytes from the child process flow into `Crosswords`
automatically. See [`librio/src/lib.rs`](../librio/src/lib.rs) for a
complete, working setup (`Crosswords::new` + `Machine::new` +
`teletypewriter`), which is also the reference consumer of this crate.

## Pull-based rendering

`rio-vt` does not draw anything. A frontend reads terminal state on demand:

- `term.visible_rows()` returns the current viewport as `Vec<Row<Square>>`.
- `Square::c()` gives a cell's character; foreground/background/flags come
  from its style.
- Damage tracking (`TerminalDamage`) lets a renderer repaint only changed
  rows instead of the whole grid each frame.

For a complete pull implementation (dirty-row tracking, per-cell readout,
cursor and selection state), see
[`librio/src/render_state.rs`](../librio/src/render_state.rs).

## C ABI

To embed from a non-Rust UI, use [`librio`](../librio). It builds this
crate as a `staticlib` and exposes an engine/surface/render-state C API
(`rio_engine_new`, `rio_surface_text`, `rio_render_state_cell`, ...). Rio's
SwiftUI + Metal frontend drives it that way.

## Architecture

```
rio-graphics   image/graphic value types + glyph decoder (leaf, no GPU)
     |
   rio-vt      this crate: VT core (crosswords, ansi, performer, event, ...)
     |     \
rio-backend   librio (C ABI)
     |
  rioterm     Rio's terminal app (adds the sugarloaf renderer + config)
```

`rio-backend` builds on `rio-vt` and re-exports every module at the same
path, so it stays the home of the user-facing config and the renderer glue
while this crate carries the reusable core.

## Safety

The public API of `rio-vt` is safe Rust. All `unsafe` and raw-pointer
handling lives in the separate `librio` C-ABI shim.

# WA

Cross-platform window assistant made primarily for Rio terminal.

WA was built for windowing for Games and Desktop applications.

```rust
let app = App::new(
	wa::Target::Application,
	Box::new(EventHandlerInstance::new(config))
);

// Available only on Target::Application
menu::create_menu();

app.run();
```

- On MacOS applications uses `NSView` and games uses [`MTKView`](https://developer.apple.com/documentation/metalkit/mtkview).

## Support

| Functionality | MacOS  | Windows  | Linux Wayland | Linux x11 |
| :-- | :--  | :--  | :-- | :-- |
| Multi window | YES (application only)  | NO | NO | NO |
| Tabs | YES (application only) | NO | NO | NO |
| Set background color | YES | NO | NO | NO |
| Set transparency | YES | NO | NO | NO |
| Open Url | YES (application only)  | NO | NO | NO |
| Theming | YES  | NO | NO | NO |

## Acknowledgments

- WA was built originally from a fork from [Macroquad](https://github.com/not-fl3/macroquad) which is licensed under MIT license.

## Reference

- https://developer.apple.com/documentation/metalkit/mtkview
- https://docs.rs/core-foundation/0.9.4/src/core_foundation/runloop.rs.html
- https://suelan.github.io/2021/02/13/20210213-dive-into-runloop-ios/

# Teletypewriter

Rust crate to create a pseudoterminal (pty) that emulates a tty, which is a command in Unix and Unix-like operating systems to print the file name of the terminal connected to standard input. tty stands for TeleTYpewriter. This project is created and maintained for [Rio terminal](https://github.com/raphamorim/rio) purposes but feel free to use it.

# Sugarloaf

Sugarloaf is Rio rendering engine, designed to be multiplatform. It is based on WebGPU, Rust library for Desktops and WebAssembly for Web (JavaScript). This project is created and maintained for Rio terminal purposes but feel free to use it.

```bash
cargo run --example text
```

## Examples

![Demo sugarloaf 1](https://github.com/raphamorim/rio/blob/main/sugarloaf/resources/demo-sugarloaf-1.png?raw=true)

s![Demo Sugarloaf wasm](https://github.com/raphamorim/rio/blob/main/sugarloaf/resources/demo-wasm-1.png?raw=true) |

## WASM Tests

### Setup

Install `wasm-bindgen-cli` globally: `cargo install wasm-bindgen-cli`.
`wasm-bindgen-cli` provides a test runner harness.

### Running Tests

Run (in the root sugarloaf directory):

```
CARGO_TARGET_WASM32_UNKNOWN_UNKNOWN_RUNNER=wasm-bindgen-test-runner cargo test --target wasm32-unknown-unknown -p sugarloaf --tests
```

Flag explanation:

- `CARGO_TARGET_WASM32_UNKNOWN_UNKNOWN_RUNNER=wasm-bindgen-test-runner`: Tells
  Cargo to use the test harness provided by `wasm-bindgen-cli`.
- `-p sugarloaf`: Only run tests in the sugarloaf directory.
- `--tests`: Only run tests; do not build examples. Many (possibly all) of the
  examples in sugarloaf/examples currently do not compile to WASM because they
  use networking.

# Copa

Copa is a fork of [Alacritty's VTE](https://github.com/alacritty/vte/) intended to extend [Paul Williams' ANSI parser state
machine] with custom instructions.

The state machine doesn't assign meaning to the parsed data and is
thus not itself sufficient for writing a terminal emulator. Instead, it is
expected that an implementation of the `Perform` trait which does something useful with the parsed data. The `Parser` handles the book keeping, and the `Perform` gets to simply handle actions.

See the [ansicode.txt](resources/ansicode.txt) for more info.

[Paul Williams' ANSI parser state machine]: https://vt100.net/emu/dec_ansi_parser
[docs]: https://docs.rs/crate/vte/

# Rio Terminal Man Pages

This directory contains manual pages for Rio terminal emulator in scdoc format.

## Files

- `rio.1.scd` - Main Rio terminal manual page (section 1)
- `rio.5.scd` - Rio configuration file format manual page (section 5)  
- `rio-bindings.5.scd` - Rio key bindings manual page (section 5)

## Building

To build the man pages, you need `scdoc` installed:

### Install scdoc

**macOS (Homebrew):**
```bash
brew install scdoc
```

**Ubuntu/Debian:**
```bash
sudo apt install scdoc
```

**Arch Linux:**
```bash
sudo pacman -S scdoc
```

**From source:**
```bash
git clone https://git.sr.ht/~sircmpwn/scdoc
cd scdoc
make
sudo make install
```

### Build man pages

```bash
# Build all man pages
make -C extra/man

# Or build individually
scdoc < extra/man/rio.1.scd > rio.1
scdoc < extra/man/rio.5.scd > rio.5
scdoc < extra/man/rio-bindings.5.scd > rio-bindings.5
```

### Install man pages

```bash
# Install to system man directory (requires sudo)
sudo cp rio.1 /usr/local/share/man/man1/
sudo cp rio.5 /usr/local/share/man/man5/
sudo cp rio-bindings.5 /usr/local/share/man/man5/

# Update man database
sudo mandb
```

### View man pages

```bash
man rio
man 5 rio
man 5 rio-bindings
```

## Format

The man pages are written in scdoc format, which is a simple markup language for writing man pages. See the [scdoc documentation](https://git.sr.ht/~sircmpwn/scdoc) for syntax details.

# Colors

## Conversion Enums

```rust
pub enum Format {
    SRGB0_255,
    SRGB0_1,
}
```

Enums are based on color conversion rules:

```bash
Hex #FFFFFF

sRGB   0-255 =  255.000  255.000  255.000
sRGB   0-1.0 =  1.00000  1.00000  1.00000
RGB Adobe 98 =  255.000  255.000  255.000
```

## Conversion to WGPU Color

[WGPU Color](https://docs.rs/wgpu/latest/wgpu/struct.Color.html)

```rust
let color: wgpu::Color =
    ColorBuilder::from_hex(String::from("#151515"), Format::SRGB0_1)
        .unwrap()
        .to_wgpu();

assert_eq!(
    color,
    Color {
        r: 0.08235294117647059,
        g: 0.08235294117647059,
        b: 0.08235294117647059,
        a: 1.0
    }
);
```

## ----include = FALSE----------------------------------------------------------
suppressPackageStartupMessages(library(data.table))

## ----featuretable, echo = FALSE-----------------------------------------------
rf <- data.table(rio:::rio_formats)[!input %in% c(",", ";", "|", "\\t") & type %in% c("import", "suggest", "archive"),]
short_rf <- rf[, paste(input, collapse = " / "), by = format_name]
type_rf <- unique(rf[,c("format_name", "type", "import_function", "export_function", "note")])

feature_table <- short_rf[type_rf, on = .(format_name)]

colnames(feature_table)[2] <- "signature"

setorder(feature_table, "type", "format_name")
feature_table$import_function <- stringi::stri_extract_first(feature_table$import_function, regex = "[a-zA-Z0-9\\.]+")
feature_table$import_function[is.na(feature_table$import_function)] <- ""
feature_table$export_function <- stringi::stri_extract_first(feature_table$export_function, regex = "[a-zA-Z0-9\\.]+")
feature_table$export_function[is.na(feature_table$export_function)] <- ""

feature_table$type <- ifelse(feature_table$type == "suggest", "Suggest", "Default")
feature_table <- feature_table[,c("format_name", "signature", "import_function", "export_function", "type", "note")]

colnames(feature_table) <- c("Name", "Extensions / \"format\"", "Import Package", "Export Package", "Type", "Note")
knitr::kable(feature_table)

## ----echo=FALSE, results='hide'-----------------------------------------------
library("rio")

export(mtcars, "mtcars.csv")
export(mtcars, "mtcars.dta")
export(mtcars, "mtcars_noext", format = "csv")

## -----------------------------------------------------------------------------
library("rio")

x <- import("mtcars.csv")
y <- import("mtcars.dta")

# confirm identical
all.equal(x, y, check.attributes = FALSE)

## -----------------------------------------------------------------------------
head(import("mtcars_noext", format = "csv"))

## ----echo=FALSE, results='hide'-----------------------------------------------
unlink("mtcars.csv")
unlink("mtcars.dta")
unlink("mtcars_noext")

## -----------------------------------------------------------------------------
library("rio")

export(mtcars, "mtcars.csv")
export(mtcars, "mtcars.dta")

## -----------------------------------------------------------------------------
library("magrittr")
mtcars %>%
  subset(hp > 100) %>%
  aggregate(. ~ cyl + am, data = ., FUN = mean) %>%
  export(file = "mtcars2.dta")

## -----------------------------------------------------------------------------
# export to sheets of an Excel workbook
export(list(mtcars = mtcars, iris = iris), "multi.xlsx")

## -----------------------------------------------------------------------------
export_list(list(mtcars = mtcars, iris = iris), "%s.tsv")

## -----------------------------------------------------------------------------
# create file to convert
export(mtcars, "mtcars.dta")

# convert Stata to SPSS
convert("mtcars.dta", "mtcars.sav")

## -----------------------------------------------------------------------------
# create an ambiguous file
fwf <- tempfile(fileext = ".fwf")
cat(file = fwf, "123456", "987654", sep = "\n")

# see two ways to read in the file
identical(import(fwf, widths = c(1, 2, 3)), import(fwf, widths = c(1, -2, 3)))

# convert to CSV
convert(fwf, "fwf.csv", in_opts = list(widths = c(1, 2, 3)))
import("fwf.csv") # check conversion

## ----echo=FALSE, results='hide'-----------------------------------------------
unlink("mtcars.dta")
unlink("mtcars.sav")
unlink("fwf.csv")
unlink(fwf)

## ----echo=FALSE, results='hide'-----------------------------------------------
unlink("mtcars.csv")
unlink("mtcars.dta")
unlink("multi.xlsx")
unlink("mtcars2.dta")
unlink("mtcars.tsv")
unlink("iris.tsv")

remote_to_local <- function(file, format) {
    if (grepl("docs.google.com/spreadsheets", file, fixed = TRUE)) {
        if (missing(format) || (!missing(format) && !format %in% c("csv", "tsv", "xlsx", "ods"))) {
            format <- "csv"
        }
        file <- .convert_google_url(file, export_as = format)
    }
    if (missing(format)) {
        ## try to extract format from URL, see below
        format <- .get_ext_temp(file)
    } else {
        format <- .standardize_format(format)
    }
    # save file locally
    temp_file <- tempfile(fileext = paste0(".", format))
    u <- curl::curl_fetch_memory(file)
    writeBin(object = u$content, con = temp_file)
    if (format != "TMP") { ## the happiest path
        return(temp_file)
    }
    ## fomart = "TMP": try to extract format from curl's final URL
    format <- .get_ext_temp(u$url)
    if (format != "TMP") { ## contain a file extension, also happy
        renamed_file <- sub("TMP$", format, temp_file)
        file.copy(from = temp_file, to = renamed_file)
        unlink(temp_file)
        return(renamed_file)
    }
    ## try to extract format from headers: read #403 about whether this code is doing anything
    h1 <- curl::parse_headers(u$headers)
    ## check `Content-Disposition` header
    if (!any(grepl("^Content-Disposition", h1))) {
        stop("Unrecognized file format. Try specifying with the format argument.")
    }
    h <- h1[grep("filename", h1, fixed = TRUE)]
    if (length(h)) {
        f <- regmatches(h, regexpr("(?<=\")(.*)(?<!\")", h, perl = TRUE))
        if (!length(f)) {
            f <- regmatches(h, regexpr("(?<=filename=)(.*)", h, perl = TRUE))
        }
        f <- file.path(dirname(temp_file), f)
        file.copy(from = temp_file, to = f)
        unlink(temp_file)
        return(f)
    }
}

.convert_google_url <- function(url, export_as = "csv") {
    ## convert a google sheets url to google csv export URL
    ## extract the doc-id and append /export?format = csv to it. (default)
    google_key <- regmatches(url, regexpr("[[:alnum:]_-]{30,}", url))
    if (grepl("gid=[[:digit:]]+", url)) {
        gidpart <- paste0(regmatches(url, regexpr("gid=[[:digit:]]+", url)))
    } else {
        gidpart <- "gid=0"
    }
    return(paste0("https://docs.google.com/spreadsheets/d/", google_key, "/export?", gidpart, "&format=", export_as))
}

.get_ext_temp <- function(file, temp_format = "TMP") {
    ## This is a version of get_ext for internal usage
    ## When file can't be queried, return `temp_format` instead of error
    format <- try(get_info(file)$format, silent = TRUE)
    if (inherits(format, "try-error")) {
        return(temp_format)
    }
    return(format)
}

skip_on_cran()

test_that("Import Remote Stata File", {
    f <- try(import("http://www.stata-press.com/data/r13/auto.dta"))
    if (!inherits(f, "try-error")) {
        expect_true(is.data.frame(f))
    }
})

test_that("Import Remote GitHub File", {
    rfile <- "https://raw.githubusercontent.com/gesistsa/rio/main/tests/testdata/noheader.csv"
    rfile_imported1 <- try(import(rfile))
    if (!inherits(rfile_imported1, "try-error")) {
        expect_true(inherits(rfile_imported1, "data.frame"), label = "Import remote file (implied format)")
    }
    rfile_imported2 <- try(import(rfile, format = "csv"))
    if (!inherits(rfile_imported2, "try-error")) {
        expect_true(inherits(rfile_imported2, "data.frame"), label = "Import remote file (explicit format)")
    }

    lfile <- remote_to_local(rfile)
    if (!inherits(lfile, "try-error")) {
        expect_true(file.exists(lfile), label = "Remote file copied successfully")
        expect_true(inherits(import(lfile), "data.frame"), label = "Import local copy successfully")
    }
    ## short url
    payload <- try(import("https://is.gd/NLAxtg"))
    if (!inherits(payload, "try-error")) {
        expect_true(inherits(payload, "data.frame"), label = "Import remote file from shorten url (implied format)")
    }
    ## no extension
    noextension_url <- "https://github.com/gesistsa/rio/raw/main/tests/testdata/iris_no_extension_xls"
    expect_error(import(noextension_url))
    expect_error(import(noextension_url, format = "xls"), NA)
})

test_that("Import from Google Sheets", {
    googleurl1 <- "https://docs.google.com/spreadsheets/d/1I9mJsS5QnXF2TNNntTy-HrcdHmIF9wJ8ONYvEJTXSNo/edit#gid=0"
    expect_true(inherits(import(googleurl1), "data.frame"), label = "Import google sheets (specified sheet)")

    googleurl2 <- "https://docs.google.com/spreadsheets/d/1I9mJsS5QnXF2TNNntTy-HrcdHmIF9wJ8ONYvEJTXSNo/edit"
    expect_true(inherits(import(googleurl2), "data.frame"), label = "Import google sheets (unspecified sheet)")

    expect_true(inherits(import(googleurl1, format = "tsv"), "data.frame"), label = "Import google sheets (specified sheet, specified format)")
})

.onAttach <- function(libname, pkgname) {
    if (interactive()) {
        w <- uninstalled_formats()
        if (length(w)) {
            msg <- "Some optional R packages were not installed and therefore some file formats are not supported. Check file support with show_unsupported_formats()"
            packageStartupMessage(msg)
        }
    }
}

#' @docType package
#' @name rio
#' @title A Swiss-Army Knife for Data I/O
#' @description The aim of rio is to make data file input and output as easy as possible. [export()] and [import()] serve as a Swiss-army knife for painless data I/O for data from almost any file format by inferring the data structure from the file extension, natively reading web-based data sources, setting reasonable defaults for import and export, and relying on efficient data import and export packages. An additional convenience function, [convert()], provides a simple method for converting between file types.
#'
#' Note that some of rio's functionality is provided by \sQuote{Suggests} dependendencies, meaning they are not installed by default. Use [install_formats()] to make sure these packages are available for use.
#'
#' @examples
#' # export
#' library("datasets")
#' export(mtcars, csv_file <- tempfile(fileext = ".csv")) # comma-separated values
#' export(mtcars, rds_file <- tempfile(fileext = ".rds")) # R serialized
#' export(mtcars, sav_file <- tempfile(fileext = ".sav")) # SPSS
#'
#' # import
#' x <- import(csv_file)
#' y <- import(rds_file)
#' z <- import(sav_file)
#'
#' # convert sav (SPSS) to dta (Stata)
#' convert(sav_file, dta_file <- tempfile(fileext = ".dta"))
#'
#' # cleanup
#' unlink(c(csv_file, rds_file, sav_file, dta_file))
#'
#' @references
#'   [datamods](https://cran.r-project.org/package=datamods) provides Shiny modules for importing data via `rio`.
#'
#'   [GREA](https://github.com/Stan125/GREA) provides an RStudio add-in to import data using rio.
#' @seealso [import()], [import_list()], [export()], [export_list()], [convert()], [install_formats()]
"_PACKAGE"

#' @rdname gather_attrs
#' @title Gather attributes from data frame variables
#' @description `gather_attrs` moves variable-level attributes to the data frame level and `spread_attrs` reverses that operation.
#' @details [import()] attempts to standardize the return value from the various import functions to the extent possible, thus providing a uniform data structure regardless of what import package or function is used. It achieves this by storing any optional variable-related attributes at the variable level (i.e., an attribute for `mtcars$mpg` is stored in `attributes(mtcars$mpg)` rather than `attributes(mtcars)`). `gather_attrs` moves these to the data frame level (i.e., in `attributes(mtcars)`). `spread_attrs` moves attributes back to the variable level.
#' @param x A data frame.
#' @return `x`, with variable-level attributes stored at the data frame level.
#' @seealso [import()], [characterize()]
#' @export
gather_attrs <- function(x) {
    if (!inherits(x, "data.frame")) {
        stop("'x' is not a data.frame")
    }
    dfattrs <- attributes(x)
    if ("label" %in% names(dfattrs)) {
        names(dfattrs)[names(dfattrs) == "label"] <- "title"
    }
    varattrs <- rep(list(list()), length(x))
    for (i in seq_along(x)) {
        a <- attributes(x[[i]])
        varattrs[[i]] <- a[!names(a) %in% c("levels", "class")]
        attr(x[[i]], "label") <- NULL
        if (any(grepl("labelled", class(x[[i]]), fixed = TRUE))) {
            x[[i]] <- haven::zap_labels(x[[i]])
        }
        f <- grep("^format", names(attributes(x[[i]])), value = TRUE)
        if (length(f)) {
            attr(x[[i]], f) <- NULL
        }
        rm(f)
    }
    if (any(lengths(varattrs))) {
        attrnames <- sort(unique(unlist(lapply(varattrs, names))))
        outattrs <- stats::setNames(lapply(attrnames, function(z) {
            stats::setNames(lapply(varattrs, `[[`, z), names(x))
        }), attrnames)
        attributes(x) <- c(dfattrs, outattrs)
    }
    x
}

#' @rdname gather_attrs
#' @export
spread_attrs <- function(x) {
    if (!inherits(x, "data.frame")) {
        stop("'x' is not a data.frame")
    }
    dfattrs <- attributes(x)
    d_level_attrs <- names(dfattrs) %in% c("row.names", "class", "names", "notes", "title")
    varattrs <- dfattrs[!d_level_attrs]
    for (i in seq_along(x)) {
        a <- attributes(x[[i]])
        attributes(x[[i]]) <- c(a, lapply(varattrs, `[[`, i))
    }
    if ("title" %in% names(dfattrs)) {
        names(dfattrs)[names(dfattrs) == "title"] <- "label"
    }
    attributes(x) <- dfattrs[d_level_attrs]
    x
}

e <- try(import("http://www.stata-press.com/data/r13/auto.dta"))

if (!inherits(e, "try-error")) {

    g <- gather_attrs(e)
    test_that("Gather attrs from Stata", {
        expect_true(length(attributes(e[[1]])) >= 1)
        expect_true(length(attributes(g[[1]])) == 0)
        expect_true(length(attributes(e)) == 5)
        expect_true(length(attributes(g)) == 8)
        expect_true("label" %in% names(attributes(e[[1]])))
        expect_true(!"label" %in% names(attributes(g[[1]])))
        expect_true("label" %in% names(attributes(g)))
        expect_true("labels" %in% names(attributes(g)))
        expect_true("format.stata" %in% names(attributes(g)))
        expect_true(!"format.stata" %in% names(attributes(g[[1]])))
    })

    test_that("Spread attrs from Stata", {
        s <- spread_attrs(g)
        # df-level attributes
        expect_true("label" %in% names(attributes(s)))
        expect_true("notes" %in% names(attributes(s)))
        # spread attributes
        expect_true("format.stata" %in% names(attributes(s[[1]])))
        expect_true(!"format.stata" %in% names(attributes(s)))
        expect_true("label" %in% names(attributes(s[[1]])))
        expect_true(!"labels" %in% names(attributes(s)))
    })

    test_that("Gather empty attributes", {
        require("datasets")
        g <- gather_attrs(iris)
        expect_true(length(attributes(iris[[1]])) == 0)
        expect_true(length(attributes(g[[1]])) == 0)
        expect_true(length(attributes(iris)) == 3)
        expect_true(length(attributes(g)) == 3)
    })

    test_that("gather_attrs() fails on non-data frame", {
        expect_error(gather_attrs(letters))
    })

    test_that("spread_attrs() fails on non-data frame", {
        expect_error(spread_attrs(letters))
    })
}

#' @title Export list of data frames to files
#' @description Use [export()] to export a list of data frames to a vector of file names or a filename pattern.
#' @param x A list of data frames to be written to files.
#' @param file A character vector string containing a single file name with a `\%s` wildcard placeholder, or a vector of file paths for multiple files to be imported. If `x` elements are named, these will be used in place of `\%s`, otherwise numbers will be used; all elements must be named for names to be used.
#' @param archive character. Either empty string (default) to save files in current
#' directory, a path to a (new) directory, or a .zip/.tar file to compress all
#' files into an archive.
#' @param \dots Additional arguments passed to [export()].
#' @return The name(s) of the output file(s) as a character vector (invisibly).
#' @details [export()] can export a list of data frames to a single multi-dataset file (e.g., an Rdata or Excel .xlsx file). Use `export_list` to export such a list to *multiple* files.
#' @examples
#' ## For demo, a temp. file path is created with the file extension .xlsx
#' xlsx_file <- tempfile(fileext = ".xlsx")
#' export(
#'     list(
#'         mtcars1 = mtcars[1:10, ],
#'         mtcars2 = mtcars[11:20, ],
#'         mtcars3 = mtcars[21:32, ]
#'     ),
#'     xlsx_file
#' )
#'
#' # import a single file from multi-object workbook
#' import(xlsx_file, sheet = "mtcars1")
#' # import all worksheets, the return value is a list
#' import_list(xlsx_file)

#' library('datasets')
#' export(list(mtcars1 = mtcars[1:10,],
#'             mtcars2 = mtcars[11:20,],
#'             mtcars3 = mtcars[21:32,]),
#'     xlsx_file <- tempfile(fileext = ".xlsx")
#' )
#'
#' # import all worksheets
#' list_of_dfs <- import_list(xlsx_file)
#'
#' # re-export as separate named files
#'
#' ## export_list(list_of_dfs, file = c("file1.csv", "file2.csv", "file3.csv"))
#'
#' # re-export as separate files using a name pattern; using the names in the list
#' ## This will be written as "mtcars1.csv", "mtcars2.csv", "mtcars3.csv"
#'
#' ## export_list(list_of_dfs, file = "%s.csv")
#' @seealso [import()], [import_list()], [export()]
#' @export
export_list <- function(x, file, archive = "", ...) {
    .check_file(file, single_only = FALSE)
    archive_format <- find_compress(archive)
    supported_archive_formats <- c("zip", "tar", "tar.gz", "tar.bz2")
    if (!is.na(archive_format$compress) && !archive_format$compress %in% supported_archive_formats) {
        stop("'archive' is specified but format is not supported. Only zip and tar formats are supported.", call. = FALSE)
    }
    if (inherits(x, "data.frame")) {
        stop("'x' must be a list. Perhaps you want export()?", call. = FALSE)
    }
    .check_tar_support(archive_format$compress, getRversion())

    outfiles <- .create_outfiles(file, x)

    if (is.na(archive_format$compress) && archive_format$file != "") {
        outfiles <- file.path(archive_format$file, outfiles)
    }
    outfiles_normalized <- normalizePath(outfiles, mustWork = FALSE)

    out <- list()
    for (f in seq_along(x)) {
        out[[f]] <- try(export(x[[f]], file = outfiles_normalized[f], ...), silent = TRUE)
        if (inherits(out[[f]], "try-error")) {
            warning(sprintf("Export failed for element %d, filename: %s", f, outfiles[f]))
        }
    }
    if (!is.na(archive_format$compress)) {
        .create_directory_if_not_exists(archive)
        compress_out(archive, outfiles_normalized, type = archive_format$compress)
        unlink(outfiles_normalized)
        return(invisible(archive))
    }
    return(invisible(outfiles))
}

test_that(".check_file", {
    data <- data.frame(
        x = sample(1:10, 10000, replace = TRUE),
        y = sample(1:10, 10000, replace = TRUE)
    )
    expect_error(.check_file(1))
    expect_error(.check_file(TRUE))
    expect_error(.check_file(data))
    expect_error(.check_file(iris))
    expect_error(.check_file(c("a.csv", "b.csv")))
    expect_error(.check_file(NA))
    expect_error(.check_file(NA_character_))
    expect_error(.check_file(c(NA, "a.csv")))
    expect_error(.check_file(c(NA_character_, "a.csv")))
    expect_error(.check_file("a.csv"), NA)
    expect_error(.check_file(), NA)
    ## single_only FALSE
    expect_error(.check_file(1, single_only = FALSE))
    expect_error(.check_file(TRUE, single_only = FALSE))
    expect_error(.check_file(data, single_only = FALSE))
    expect_error(.check_file(iris, single_only = FALSE))
    expect_error(.check_file(c("a.csv", "b.csv"), single_only = FALSE), NA)
    expect_error(.check_file("a.csv"), NA)
    expect_error(.check_file(single_only = FALSE), NA)
})

test_that("Invalid file argument - import(), #301", {
    data <- data.frame(
        x = sample(1:10, 10000, replace = TRUE),
        y = sample(1:10, 10000, replace = TRUE)
    )
    expect_error(import(data), "Invalid")
    expect_error(import(iris), "Invalid")
    expect_error(import(1), "Invalid")
    expect_error(import(TRUE), "Invalid")
    expect_error(import(c("a.csv", "b.csv")), "Invalid")
})

test_that("Invalid file argument - import_list(), #301", {
    data <- data.frame(
        x = sample(1:10, 10000, replace = TRUE),
        y = sample(1:10, 10000, replace = TRUE)
    )
    expect_error(import_list(data), "Invalid")
    expect_error(import_list(iris), "Invalid")
    expect_error(import_list(1), "Invalid")
    expect_error(import_list(TRUE), "Invalid")
})

test_that("Invalid file argument - export(), #301", {
    data <- data.frame(
        x = sample(1:10, 10000, replace = TRUE),
        y = sample(1:10, 10000, replace = TRUE)
    )
    expect_error(export(iris, data), "Invalid")
    expect_error(export(iris, iris), "Invalid")
    expect_error(export(iris, 1), "Invalid")
    expect_error(export(iris, TRUE), "Invalid")
    expect_error(export(iris, c("abc.csv", "123.csv")), "Invalid")
})

test_that("Invalid file argument - export_list(), #301", {
    data <- data.frame(
        x = sample(1:10, 10000, replace = TRUE),
        y = sample(1:10, 10000, replace = TRUE)
    )
    expect_error(export_list(iris, data), "Invalid")
    expect_error(export_list(iris, iris), "Invalid")
    expect_error(export_list(iris, 1), "Invalid")
    expect_error(export_list(iris, TRUE), "Invalid")
})

test_that("Data identical (import_list)", {
    withr::with_tempfile("mtcars_file", fileext = ".rds", code = {
        export(mtcars, mtcars_file)
        expect_equivalent(import_list(rep(mtcars_file, 2), trust = TRUE), list(mtcars, mtcars))
        mdat <- rbind(mtcars, mtcars)
        dat <- import_list(rep(mtcars_file, 2), rbind = TRUE, trust = TRUE)
        expect_true(ncol(dat) == ncol(mdat) + 1)
        expect_true(nrow(dat) == nrow(mdat))
        expect_true("_file" %in% names(dat))
    })
})

test_that("Import multi-object .Rdata in import_list()", {
    withr::with_tempfile("rdata_file", fileext = ".rdata", code = {
        export(list(mtcars = mtcars, iris = iris), rdata_file)
        dat <- import_list(rdata_file, trust = TRUE)
        expect_true(identical(dat[[1]], mtcars))
        expect_true(identical(dat[[2]], iris))
    })
})

test_that("Import multiple HTML tables in import_list()", {
    dat <- import_list("../testdata/twotables.html")
    expect_true(identical(dim(dat[[1]]), dim(mtcars)))
    expect_true(identical(names(dat[[1]]), names(mtcars)))
    expect_true(identical(dim(dat[[2]]), dim(iris)))
    expect_true(identical(names(dat[[2]]), names(iris)))
})

test_that("Import multiple HTML tables in import_list() but with htm #350", {
    withr::with_tempfile("temphtm", fileext = ".htm", code = {
        temphtm <- tempfile(fileext = ".htm")
        file.copy("../testdata/twotables.html", temphtm)
        dat <- import_list(temphtm)
        expect_true(identical(dim(dat[[1]]), dim(mtcars)))
        expect_true(identical(names(dat[[1]]), names(mtcars)))
        expect_true(identical(dim(dat[[2]]), dim(iris)))
        expect_true(identical(names(dat[[2]]), names(iris)))
    })
})

test_that("import_list() preserves 'which' names when specified", {
    withr::with_tempfile("data_file", fileext = ".xlsx", code = {
        export(list(a = mtcars, b = iris), data_file)
        expect_true(identical(names(import_list(data_file)), c("a", "b")))
        expect_true(identical(names(import_list(data_file, which = 1)), "a"))
        expect_true(identical(names(import_list(data_file, which = "a")), "a"))
        expect_true(identical(names(import_list(data_file, which = 2)), "b"))
        expect_true(identical(names(import_list(data_file, which = "b")), "b"))
        expect_true(identical(names(import_list(data_file, which = 1:2)), c("a", "b")))
        expect_true(identical(names(import_list(data_file, which = 2:1)), c("b", "a")))
        expect_true(identical(names(import_list(data_file, which = c("a", "b"))), c("a", "b")))
        expect_true(identical(names(import_list(data_file, which = c("b", "a"))), c("b", "a")))
    })
})

test_that("import_list() preserves 'which' names when specified ods", {
    skip_if_not_installed("readODS")
    withr::with_tempfile("data_file", fileext = ".ods", code = {
        export(list(a = mtcars, b = iris), data_file)
        expect_true(identical(names(import_list(data_file)), c("a", "b")))
        expect_true(identical(names(import_list(data_file, which = 1)), "a"))
        expect_true(identical(names(import_list(data_file, which = "a")), "a"))
        expect_true(identical(names(import_list(data_file, which = 2)), "b"))
        expect_true(identical(names(import_list(data_file, which = "b")), "b"))
        expect_true(identical(names(import_list(data_file, which = 1:2)), c("a", "b")))
        expect_true(identical(names(import_list(data_file, which = 2:1)), c("b", "a")))
        expect_true(identical(names(import_list(data_file, which = c("a", "b"))), c("a", "b")))
        expect_true(identical(names(import_list(data_file, which = c("b", "a"))), c("b", "a")))
    })
})


test_that("Import single file via import_list()", {
    withr::with_tempfile("data_file", fileext = ".rds", code = {
        export(mtcars, data_file)
        expect_true(identical(import_list(data_file, rbind = TRUE, trust = TRUE), mtcars))
    })
})

test_that("Import single file from zip via import_list()", {
    withr::with_tempfile("data_file", fileext = ".csv.zip", code = {
        export(mtcars, data_file, format = "csv")
        expect_true(is.data.frame(import_list(data_file)[[1L]]))
        expect_true(is.data.frame(import_list(data_file, which = 1)[[1L]]))
        basefile_name <- gsub(".zip$", "", basename(data_file))
        expect_true(is.data.frame(import_list(data_file, which = basefile_name)[[1L]]))
    })
})

test_that("Import multiple files from zip via import_list()", {
    withr::with_tempfile("data_file", fileext = ".csv.zip", code = {
        mylist <- list(mtcars3 = mtcars[1:10, ], mtcars2 = mtcars[11:20, ], mtcars1 = mtcars[21:32, ])
        expect_error(export_list(mylist, file = paste0("mtcars", 1:3, ".csv"), archive = data_file), NA)
        expect_error(res <- import_list(data_file), NA)
        expect_true(is.list(res))
        expect_equal(length(res), 3)
        expect_true(is.data.frame(res[[1]]))
        expect_true(is.data.frame(res[[2]]))
        expect_true(is.data.frame(res[[3]]))
    })
})

test_that("Import multiple files from zip via import_list()", {
    skip_if(getRversion() <= "4.0")
    withr::with_tempfile("data_file", fileext = ".csv.tar.gz", code = {
        mylist <- list(mtcars3 = mtcars[1:10, ], mtcars2 = mtcars[11:20, ], mtcars1 = mtcars[21:32, ])
        expect_error(export_list(mylist, file = paste0("mtcars", 1:3, ".csv"), archive = data_file), NA)
        expect_error(res <- import_list(data_file), NA)
        expect_true(is.list(res))
        expect_equal(length(res), 3)
        expect_true(is.data.frame(res[[1]]))
        expect_true(is.data.frame(res[[2]]))
        expect_true(is.data.frame(res[[3]]))
    })
})

test_that("Using setclass in import_list()", {
    withr::with_tempfile("data_file", fileext = ".rds", code = {
        export(mtcars, data_file)
        dat1 <- import_list(rep(data_file, 2), setclass = "data.table", rbind = TRUE, trust = TRUE)
        expect_true(inherits(dat1, "data.table"))
        dat2 <- import_list(rep(data_file, 2), setclass = "tbl", rbind = TRUE, trust = TRUE)
        expect_true(inherits(dat2, "tbl"))
    })
})

test_that("Object names are preserved by import_list()", {
    withr::with_tempdir(code = {
        export(list(mtcars1 = mtcars[1:10,],
                    mtcars2 = mtcars[11:20,],
                    mtcars3 = mtcars[21:32,]), "mtcars.xlsx")
        export(list(mtcars1 = mtcars[1:10,],
                    mtcars2 = mtcars[11:20,],
                    mtcars3 = mtcars[21:32,]), "mtcars.ods")
        export(list(mtcars1 = mtcars[1:10,],
                    mtcars2 = mtcars[11:20,],
                    mtcars3 = mtcars[21:32,]), "mtcars.fods")
        export(mtcars[1:10,],  "mtcars1.csv")
        export(mtcars[11:20,], "mtcars2.tsv")
        export(mtcars[21:32,], "mtcars3.csv")
        expected_names <- c("mtcars1", "mtcars2", "mtcars3")
        dat_xls <- import_list("mtcars.xlsx")
        dat_csv <- import_list(c("mtcars1.csv","mtcars2.tsv","mtcars3.csv"))
        dat_ods <- import_list("mtcars.ods")
        dat_fods <- import_list("mtcars.fods")
        expect_identical(names(dat_xls), expected_names)
        expect_identical(names(dat_csv), expected_names)
        expect_identical(names(dat_ods), expected_names)
        expect_identical(names(dat_fods), expected_names)
    })
})

test_that("File names are added as attributes by import_list()", {
    withr::with_tempdir(code = {
        export(mtcars[1:10,],  "mtcars.csv")
        export(mtcars[11:20,], "mtcars.tsv")
        expected_names <- c("mtcars", "mtcars")
        expected_attrs <- c(mtcars = "mtcars.csv", mtcars = "mtcars.tsv")
        dat <- import_list(c("mtcars.csv","mtcars.tsv"))
        expect_identical(names(dat), expected_names)
        expect_identical(unlist(lapply(dat, attr, "filename")), expected_attrs)
    })
})

test_that("URL #294", {
    skip_on_cran()
    ## url <- "https://evs.nci.nih.gov/ftp1/CDISC/SDTM/SDTM%20Terminology.xls" That's 10MB!
    url <- "https://github.com/tidyverse/readxl/raw/main/tests/testthat/sheets/sheet-xml-lookup.xlsx"
    expect_error(x <- import_list(url), NA)
    expect_true(inherits(x, "list"))
    expect_true("Asia" %in% names(x))
    expect_true("Africa" %in% x[[1]]$continent)
    expect_false("Africa" %in% x[[2]]$continent)
    ## double URLs; it reads twice the first sheet by default
    urls <- c(url, url)
    expect_error(x2 <- import_list(urls), NA)
    expect_true("sheet-xml-lookup" %in% names(x2))
    expect_true("Africa" %in% x2[[1]]$continent)
    expect_true("Africa" %in% x2[[2]]$continent)
})

test_that("Universal dummy `which` #326", {
    skip_if_not_installed("readODS")
    formats <- c("ods", "fods", "xlsx", "dta", "sav", "csv", "csv2")
    for (format in formats) {
        withr::with_tempfile("tempzip", fileext = paste0(".", format, ".zip"), code = {
            rio::export(mtcars, tempzip, format = format)
            expect_warning(rio::import(tempzip), NA)
            expect_warning(rio::import_list(tempzip), NA)
        })
    }
})

test_that("Universal dummy `which` (Suggests) #326", {
    skip_if_not_installed("qs2")
    skip_if_not_installed("arrow")
    skip_if_not_installed("readODS")
    skip_on_os("mac") ## apache/arrow#40991
    formats <- c("qs2", "parquet", "ods")
    for (format in formats) {
        withr::with_tempfile("tempzip", fileext = paste0(".", format, ".zip"), code = {
            rio::export(mtcars, tempzip, format = format)
            expect_warning(rio::import(tempzip), NA)
            expect_warning(rio::import_list(tempzip), NA)
        })
    }
})

test_that("Informative message when files are not found #389", {
    withr::with_tempfile("mtcars_file", fileext = ".rds", code = {
        export(mtcars, mtcars_file)
        expect_true(file.exists(mtcars_file))
        expect_false(file.exists("nonexisting.rds"))
        expect_warning(import_list(c(mtcars_file, "nonexisting.rds"), trust = TRUE), "^Import failed for nonexisting")
    })
})

test_that("Missing files and rbind", {
    withr::with_tempfile("mtcars_file", fileext = ".rds", code = {
        export(mtcars, mtcars_file)
        expect_true(file.exists(mtcars_file))
        expect_false(file.exists("nonexisting.rds"))
        expect_false(file.exists("nonexisting2.rds"))
        expect_warning(x <- import_list(c(mtcars_file, "nonexisting.rds"), rbind = TRUE, trust = TRUE), "^Import failed for nonexisting")
        expect_true(is.data.frame(x))
        expect_warning(x <- import_list(c("nonexisting.rds", "nonexisting2.rds"), rbind = TRUE, trust = TRUE), "^Import failed for nonexisting")
    })
})

library("datasets")

test_that("export_list() works", {
    withr::with_tempdir({
        export(list(
            mtcars3 = mtcars[1:10, ],
            mtcars2 = mtcars[11:20, ],
            mtcars1 = mtcars[21:32, ]
        ), "mtcars.xlsx")
        mylist <- import_list("mtcars.xlsx")

        expect_error(export_list(mtcars), label = "export_list() fails on exporting single data frame")
        expect_error(export_list(mylist, file = NULL), label = "export_list() fails when file is NULL")

        expect_true(identical(export_list(mylist, file = paste0("mtcars_", 3:1, ".csv")), paste0("mtcars_", 3:1, ".csv")))
        expect_true(identical(export_list(mylist, file = "%s.csv"), paste0("mtcars", 3:1, ".csv")))

        expect_true(identical(export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "archive.zip"), "archive.zip"))
        expect_true(identical(export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "arch/archive.zip"), "arch/archive.zip"))

        expect_true(all.equal(mylist[["mtcars1"]], import("mtcars1.csv")))
        expect_true(all.equal(mylist[["mtcars2"]], import("mtcars2.csv")))
        expect_true(all.equal(mylist[["mtcars3"]], import("mtcars3.csv")))

        names(mylist) <- NULL
        expect_true(identical(export_list(mylist, file = "mtcars_%s.csv"), paste0("mtcars_", 1:3, ".csv")))

        names(mylist) <- c("a", "", "c")
        expect_error(export_list(mylist), label = "export_list() fails without 'file' argument")
        expect_error(export_list(mylist, file = "%.csv"), label = "export_list() fails without missing names")
        expect_error(export_list(mylist, file = c("a.csv", "b.csv")), label = "export_list() fails with mismatched argument lengths")

        names(mylist) <- c("a", "a", "c")
        expect_error(export_list(mylist, file = "mtcars_%s.csv"), label = "export_list() fails with duplicated data frame names")
        expect_error(export_list(mylist, file = c("mtcars1.csv", "mtcars1.csv", "mtcars3.csv")), label = "export_list() fails with duplicated data frame names")
    })
})

test_that("archive formats, #415", {
    skip_if(getRversion() <= "4.0")
    withr::with_tempdir({
        mylist <- list(mtcars3 = mtcars[1:10, ], mtcars2 = mtcars[11:20, ], mtcars1 = mtcars[21:32, ])
        expect_error(export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "archive.csv.gz"), "specified but format is not supported")
        expect_error(export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "archive.csv.bz2"), "specified but format is not supported")
        expect_error(export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "archive.csv.zip"), NA)
        expect_error(export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "archive.csv.tar"), NA)
        expect_error(export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "archive.csv.tar.gz"), NA)
        expect_error(export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "archive.csv.tar.bz2"), NA)
    })
})

test_that("List length of one, #385", {
    withr::with_tempdir({
        example1 <- list("iris" = iris)
        tempfile <- tempfile(fileext = ".csv")
        expect_error(export(example1, tempfile), NA)
        tempfile <- tempfile(fileext = ".xlsx")
        expect_error(export(example1, tempfile), NA)
        expect_equal(readxl::excel_sheets(tempfile), "iris") ## name is retained
        tempfile <- tempfile(fileext = ".rds")
        expect_error(export(example1, tempfile), NA)
        expect_true(is.list(readRDS(tempfile)) && !is.data.frame(readRDS(tempfile)))
    })
})

test_that("tar export error for R < 4.0.3", {
    skip_if(getRversion() >= "4.0.3")
    withr::with_tempdir({
        mylist <- list(mtcars3 = mtcars[1:10, ], mtcars2 = mtcars[11:20, ], mtcars1 = mtcars[21:32, ])
        expect_error(export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "archive.csv.tar"), "^Exporting")
    })
})

#' @rdname export
#' @title Export
#' @description Write data.frame to a file
#' @param x A data frame, matrix or a single-item list of data frame to be written into a file. Exceptions to this rule are that `x` can be a list of multiple data frames if the output file format is an OpenDocument Spreadsheet (.ods, .fods), Excel .xlsx workbook, .Rdata file, or HTML file, or a variety of R objects if the output file format is RDS or JSON. See examples.) To export a list of data frames to multiple files, use [export_list()] instead.
#' @param file A character string naming a file. Must specify `file` and/or `format`.
#' @param format An optional character string containing the file format, which can be used to override the format inferred from `file` or, in lieu of specifying `file`, a file with the symbol name of `x` and the specified file extension will be created. Must specify `file` and/or `format`. Shortcuts include: \dQuote{,} (for comma-separated values), \dQuote{;} (for semicolon-separated values), \dQuote{|} (for pipe-separated values), and \dQuote{dump} for [base::dump()].
#' @param \dots Additional arguments for the underlying export functions. This can be used to specify non-standard arguments. See examples.
#' @return The name of the output file as a character string (invisibly).
#' @details This function exports a data frame or matrix into a file with file format based on the file extension (or the manually specified format, if `format` is specified).
#'
#' The output file can be to a compressed directory, simply by adding an appropriate additional extensiont to the `file` argument, such as: \dQuote{mtcars.csv.tar}, \dQuote{mtcars.csv.zip}, or \dQuote{mtcars.csv.gz}.
#'
#' `export` supports many file formats. See the documentation for the underlying export functions for optional arguments that can be passed via `...`
#'
#' \itemize{
#'     \item Comma-separated data (.csv), using [data.table::fwrite()]
#'     \item Pipe-separated data (.psv), using [data.table::fwrite()]
#'     \item Tab-separated data (.tsv), using [data.table::fwrite()]
#'     \item SAS (.sas7bdat), using [haven::write_sas()].
#'     \item SAS XPORT (.xpt), using [haven::write_xpt()].
#'     \item SPSS (.sav), using [haven::write_sav()]
#'     \item SPSS compressed (.zsav), using [haven::write_sav()]
#'     \item Stata (.dta), using [haven::write_dta()]. Note that variable/column names containing dots (.) are not allowed and will produce an error.
#'     \item Excel (.xlsx), using [writexl::write_xlsx()]. `x` can also be a list of data frames; the list entry names are used as sheet names.
#'     \item R syntax object (.R), using [base::dput()] (by default) or [base::dump()] (if `format = 'dump'`)
#'     \item Saved R objects (.RData,.rda), using [base::save()]. In this case, `x` can be a data frame, a named list of objects, an R environment, or a character vector containing the names of objects if a corresponding `envir` argument is specified.
#'     \item Serialized R objects (.rds), using [base::saveRDS()]. In this case, `x` can be any serializable R object.
#'     \item Serialized R objects (.qs2), using [qs2::qs_save()].
#'     \item "XBASE" database files (.dbf), using [foreign::write.dbf()]
#'     \item Weka Attribute-Relation File Format (.arff), using [foreign::write.arff()]
#'     \item Fixed-width format data (.fwf), using [utils::write.table()] with `row.names = FALSE`, `quote = FALSE`, and `col.names = FALSE`
#'     \item [CSVY](https://github.com/csvy) (CSV with a YAML metadata header) using [data.table::fwrite()].
#'     \item Apache Arrow Parquet (.parquet), using [nanoparquet::write_parquet()]
#'     \item Feather R/Python interchange format (.feather), using [arrow::write_feather()]
#'     \item Fast storage (.fst), using [fst::write.fst()]
#'     \item JSON (.json), using [jsonlite::toJSON()]. In this case, `x` can be a variety of R objects, based on class mapping conventions in this paper: [https://arxiv.org/abs/1403.2805](https://arxiv.org/abs/1403.2805).
#'     \item Matlab (.mat), using [rmatio::write.mat()]
#'     \item OpenDocument Spreadsheet (.ods, .fods), using [readODS::write_ods()] or [readODS::write_fods()].
#'     \item HTML (.html), using a custom method based on [xml2::xml_add_child()] to create a simple HTML table and [xml2::write_xml()] to write to disk.
#'     \item XML (.xml), using a custom method based on [xml2::xml_add_child()] to create a simple XML tree and [xml2::write_xml()] to write to disk.
#'     \item YAML (.yml), using [yaml::write_yaml()], default to write the content with UTF-8. Might not work on some older systems, e.g. default Windows locale for R <= 4.2.
#'     \item Clipboard export (on Windows and Mac OS), using [utils::write.table()] with `row.names = FALSE`
#' }
#'
#' When exporting a data set that contains label attributes (e.g., if imported from an SPSS or Stata file) to a plain text file, [characterize()] can be a useful pre-processing step that records value labels into the resulting file (e.g., `export(characterize(x), "file.csv")`) rather than the numeric values.
#'
#' Use [export_list()] to export a list of dataframes to separate files.
#'
#' @examples
#' ## For demo, a temp. file path is created with the file extension .csv
#' csv_file <- tempfile(fileext = ".csv")
#' ## .xlsx
#' xlsx_file <- tempfile(fileext = ".xlsx")
#'
#' ## create CSV to import
#' export(iris, csv_file)
#'
#' ## You can certainly export your data with the file name, which is not a variable:
#' ## import(mtcars, "car_data.csv")
#'
#' ## pass arguments to the underlying function
#' ## data.table::fwrite is the underlying function and `col.names` is an argument
#' export(iris, csv_file, col.names = FALSE)
#'
#' ## export a list of data frames as worksheets
#' export(list(a = mtcars, b = iris), xlsx_file)
#'
#' # NOT RECOMMENDED
#'
#' ## specify `format` to override default format
#' export(iris, xlsx_file, format = "csv") ## That's confusing
#' ## You can also specify only the format; in the following case
#' ## "mtcars.dta" is written [also confusing]
#'
#' ## export(mtcars, format = "stata")
#' @seealso [characterize()], [import()], [convert()], [export_list()]
#' @export
export <- function(x, file, format, ...) {
    .check_file(file, single_only = TRUE)
    if (missing(file) && missing(format)) {
        stop("Must specify 'file' and/or 'format'")
    }
    if (!missing(file)) {
        cfile <- file
        f <- find_compress(file)
        file <- f$file
        compress <- f$compress
        format <- ifelse(isFALSE(missing(format)), tolower(format), get_info(file)$input)
    } else {
        format <- .standardize_format(format)
        file <- paste0(as.character(substitute(x)), ".", format)
        compress <- NA_character_
    }
    .check_tar_support(compress, getRversion())
    format <- .standardize_format(format)
    outfile <- file
    if (is.matrix(x) || inherits(x, "ArrowTabular")) {
        x <- as.data.frame(x)
    }
    if (!is.data.frame(x) && is.list(x) && length(x) == 1 && is.data.frame(x[[1]]) &&
        !format %in% c("xlsx", "html", "rdata", "rds", "json", "qs", "qs2", "fods", "ods")) {
        x <- x[[1]] ## fix 385
    }
    if (!is.data.frame(x) && !format %in% c("xlsx", "html", "rdata", "rds", "json", "qs", "qs2", "fods", "ods")) {
        stop("'x' is not a data.frame or matrix", call. = FALSE)
    }
    if (format == "gz") {
        format <- get_info(tools::file_path_sans_ext(file, compression = FALSE))$format
        if (format != "csv") {
            stop("gz is only supported for csv (for now).", call. = FALSE)
        }
    }
    .create_directory_if_not_exists(file = file) ## fix 347
    class(file) <- c(paste0("rio_", format), class(file))
    .export(file = file, x = x, ...)
    if (!is.na(compress)) {
        cfile <- compress_out(cfile = cfile, filename = file, type = compress)
        unlink(file)
        return(invisible(cfile))
    }
    invisible(unclass(outfile))
}

require("datasets")

test_that("Export / Import to .R dump file", {
    withr::with_tempfile("iris_file", fileext = ".R", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, trust = TRUE)))

    })
    withr::with_tempfile("iris_file", fileext = ".dump", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, trust = TRUE)))
    })
})

test_that("Export to and import from SAS (.xpt)", {
    withr::with_tempfile("mtcars_file", fileext = ".xpt", code = {
        export(mtcars, mtcars_file)
        expect_true(file.exists(mtcars_file))
        expect_true(is.data.frame(import(mtcars_file)))
    })
})

test_that("Export SAS (.sas7bdat)", {
    withr::with_tempfile("mtcars_file", fileext = ".sas7bdat", code = {
        suppressWarnings(export(mtcars, mtcars_file))
        expect_true(file.exists(mtcars_file))
    })
})

skip_if_not_installed("pzfx")

test_that("Export to and import from pzfx", {
    ## pzfx support only numeric data
    iris_numeric <- iris
    iris_numeric$Species <- as.numeric(iris_numeric$Species)
    withr::with_tempfile("iris_file", fileext = ".pzfx", code = {
        export(iris_numeric, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
        ## Note that the dim test is only true as long as the data are exported with
        ## write_pzfx(..., row_names=FALSE) which is the default in the export
        ## method, but it is not default in pzfx::write_pzfx()
        expect_true(identical(dim(import(iris_file)), dim(iris_numeric)))
        expect_true(identical(dim(import(iris_file, which = "Data 1")), dim(iris_numeric)))
        expect_true(identical(dim(import(iris_file, table = "Data 1")), dim(iris_numeric)))
    })
})

simport <- function(...) {
    suppressWarnings(import(...))
}

test_that("Export to and import from FWF .fwf", {
    withr::with_tempfile("iris_file", fileext = ".fwf", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(simport(iris_file, widths = c(3, 3, 3, 3, 1))))
        expect_true(is.data.frame(simport(iris_file, widths = list(c(3, 3, 3, 3, 1)))))
        expect_true(is.data.frame(simport(iris_file, widths = c(3, 3, 3, 3, 1), col.names = names(iris))))
        ## negative column widths
        expect_true(is.data.frame(simport(iris_file, widths = c(-3, 3, 3, 3, 1))))
        ##use col_position instead
        expect_error(x <- import(iris_file, col_position = readr::fwf_widths(c(1, 3, 3, 3, 1), col_names = names(iris))), NA)
        expect_equal(x[1,1, drop = TRUE], 5)
    })
})

test_that("Export to and import from FWF .txt", {
    withr::with_tempfile("iris_file", fileext = ".txt", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(simport(iris_file, widths = c(3, 3, 3, 3, 1), format = "fwf")))
    })
})

test_that("Deprecation of `width` and `col.names`", {
    withr::with_tempfile("iris_file", fileext = ".fwf", code = {
        export(iris, iris_file)
        lifecycle::expect_deprecated(import(iris_file, widths = c(-3, 3, 3, 3, 1)))
        lifecycle::expect_deprecated(import(iris_file, col.names = names(iris)))
    })
})

skip_if_not_installed("jsonlite")

test_that("Export to and import from JSON", {
    withr::with_tempfile("iris_file", fileext = ".json", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
    })
})

test_that("Export to JSON (non-data frame)", {
    withr::with_tempfile("list_file", fileext = ".json", code = {
        export(list(1:10, letters), list_file)
        expect_true(file.exists(list_file))
        expect_true(inherits(import(list_file), "list"))
        expect_false(inherits(import(list_file), "data.frame"))
        expect_true(length(import(list_file)) == 2L)
    })
})

test_that("utf-8", {
    content <- c("\"", "\u010d", "\u0161", "\u00c4", "\u5b57", "\u30a2", "\u30a2\u30e0\u30ed")
    x <- data.frame(col = content)
    withr::with_tempfile("tempjson", fileext = ".json", code = {
        y <- import(export(x, tempjson))
        testthat::expect_equal(content, y$col)
    })
})

skip_if_not_installed("xml2")

test_xml <- function(breaker = "&") {
    mtcars2 <- mtcars
    colnames(mtcars2)[1] <- paste0("mp", breaker, breaker, "g")
    mtcars2[1,1] <- paste0("mp", breaker, breaker, "g")
    withr::with_tempfile("mtcars_file", fileext = ".xml", code = {
        expect_error(x <- rio::export(mtcars2, mtcars_file), NA)
        temp_df <- rio::import(mtcars_file)
        expect_equal(colnames(temp_df)[1], paste0("mp..g"))
        expect_equal(temp_df[1,1], paste0("mp", breaker, breaker, "g"))
    })
}

test_that("Export to and import from XML", {
    withr::with_tempfile("iris_file", fileext = ".xml", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
    })
})

test_that("Export to XML with &, >, ', \", >, <",{
    ## test all
    useless <- lapply(c("&", "\"", "'", "<", ">"), test_xml)
})

test_that("Export matrix to and import from CSV", {
    withr::with_tempfile("temp_files", fileext = c(".csv", ".csv"), code = {
        export(warpbreaks, temp_files[1])
        export(as.matrix(warpbreaks), temp_files[2])
        expect_true(file.exists(temp_files[1]))
        expect_true(file.exists(temp_files[2]))
        expect_true(identical(import(temp_files[1], colClasses = rep("character", 3)),
                              import(temp_files[2], colClasses = rep("character", 3))))
    })
})

test_that("Weka (.arff) imports/exports", {
    withr::with_tempfile("iris_file", fileext = ".arff", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
    })
})

skip_if_not_installed("yaml")

test_that("Export to and import from YAML", {
    withr::with_tempfile("iris_file", fileext = ".yaml", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
        expect_identical(import(iris_file)[, 1:4], iris[, 1:4])
        expect_identical(import(iris_file)$Species, as.character(iris$Species))
    })
})

test_that("utf-8", {
    skip_if(getRversion() <= "4.2")
    withr::with_tempfile("tempyaml", fileext = ".yaml", code = {
        content <- c("\"", "\u010d", "\u0161", "\u00c4", "\u5b57", "\u30a2", "\u30a2\u30e0\u30ed")
        x <- data.frame(col = content)
        y <- import(export(x, tempyaml))
        testthat::expect_equal(content, y$col)
    })
})

skip_if_not_installed("rmatio")

test_that("Export to and import from matlab", {
    skip("failing mysteriously")
    withr::with_tempfile("iris_file", fileext = ".matlab", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
        expect_true(identical(dim(import(iris_file)), dim(iris)))

    })
})

test_that("Export to Stata", {
    withr::with_tempfile("mtcars_file", fileext = ".dta", code = {
        export(mtcars, mtcars_file)
        expect_true(file.exists(mtcars_file))
        mtcars3 <- mtcars
        names(mtcars3)[1] <- "foo.bar"
        expect_error(export(mtcars3, mtcars_file), label = "Export fails on invalid Stata names")
    })
})

test_that("Import from Stata (read_dta)", {
    withr::with_tempfile("mtcars_file", fileext = ".dta", code = {
        export(mtcars, mtcars_file)
        expect_true(is.data.frame(import(mtcars_file)))
        ## arguments ignored
        expect_error(import(mtcars_file, extraneous.argument = TRUE), NA)
        expect_silent(import(mtcars_file,extraneous.argument = TRUE))
    })
})

test_that("Import from Stata with extended Haven features (read_dta)", {
    withr::with_tempfile("mtcars_file", fileext = ".dta", code = {
        export(mtcars, mtcars_file)

        expect_true(is.data.frame(mtcars_wtam <- import(mtcars_file,
                                                        col_select = c("wt", "am"),
                                                        n_max = 10
                                                        )))
        expect_identical(c(10L, 2L), dim(mtcars_wtam))
        expect_identical(c("wt", "am"), names(mtcars_wtam))
    })
})

test_that("haven is deprecated", {
    withr::with_tempfile("mtcars_file", fileext = ".dta", code = {
        export(mtcars, mtcars_file)
        lifecycle::expect_deprecated(import(mtcars_file, haven = TRUE))
        lifecycle::expect_deprecated(import(mtcars_file, haven = FALSE))
    })
})

test_that("Deprecation of untrusted dump", {
    withr::with_tempfile("iris_file", fileext = ".dump", code = {
      export(iris, iris_file)
      ## expect deprecation to work
      expect_warning(import(iris_file), regexp = "set to FALSE by default")
      ## expect false to error
      expect_error(import(iris_file, trust = FALSE))
  })
})

test_that("Deprecation of untrusted Rdata", {
    withr::with_tempfile("iris_file", fileext = ".Rdata", code = {
        export(iris, iris_file)
        ## expect deprecation to work
        expect_warning(import(iris_file), regexp = "set to FALSE by default")
        ## expect false to error
        expect_error(import(iris_file, trust = FALSE))
    })
})

test_that("Deprecation of untrusted rds", {
    withr::with_tempfile("iris_file", fileext = ".rds", code = {
      export(iris, iris_file)
      ## expect deprecation to work
      expect_warning(import(iris_file), regexp = "set to FALSE by default")
      ## expect false to error
      expect_error(import(iris_file, trust = FALSE))
  })
})

test_that("No deprecation warning if `trust` is explicit", {
    withr::with_tempfile("iris_file", fileext = ".rds", code = {
        export(iris, iris_file)
        expect_silent(import(iris_file, trust = TRUE))
        expect_error(import(iris_file, trust = FALSE)) ## no warning right?
    })
})

test_that("Undocumented feature, options", {
    withr::with_options(list(rio.import.trust = TRUE), {
        withr::with_tempfile("iris_file", fileext = ".rds", code = {
            export(iris, iris_file)
            expect_silent(import(iris_file))
            expect_error(import(iris_file), NA)
        })
    })
    withr::with_options(list(rio.import.trust = FALSE), {
        withr::with_tempfile("iris_file", fileext = ".rds", code = {
            export(iris, iris_file)
            expect_error(import(iris_file))
        })
    })
})

test_that("`trust` wont cause problems for other import methods", {
    withr::with_tempfile("iris_file", fileext = ".xlsx", code = {
        export(iris, iris_file)
        expect_silent(import(iris_file, trust = TRUE))
        expect_error(import(iris_file, trust = FALSE), NA)
    })
})

test_that("`trust` for import_list()", {
    withr::with_tempfile("iris_file", fileext = ".rdata", code = {
        export(iris, iris_file)
        expect_warning(import_list(iris_file), regexp = "set to FALSE by default")
        expect_silent(import_list(iris_file, trust = TRUE))
        expect_error(import_list(iris_file, trust = FALSE))

    })
})

test_that("`trust` wont cause problems for other formats in import_list", {
    withr::with_tempfile("data_file", fileext = ".xlsx", code = {
        export(list(a = mtcars, b = iris), data_file)
        expect_silent(import(data_file))
        expect_silent(import(data_file, trust = TRUE))
        expect_error(import(data_file, trust = FALSE), NA)
    })
})

test_that("Export to and import from rds", {
    withr::with_tempfile("iris_file", fileext = ".rds", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, trust = TRUE)))
    })
})

test_that("Export to rds (non-data frame)", {
    withr::with_tempfile("list_file", fileext = ".rds", code = {
        export(list(1:10, letters), list_file)
        expect_true(file.exists(list_file))
        expect_true(inherits(import(list_file, trust = TRUE), "list"))
        expect_true(length(import(list_file, trust = TRUE)) == 2L)
    })
})

test_that("Export to SPSS (.sav)", {
    mtcars2 <- mtcars
    ## label and value labels
    mtcars2[["cyl"]] <- factor(mtcars2[["cyl"]], c(4, 6, 8), c("four", "six", "eight"))
    attr(mtcars2[["cyl"]], "label") <- "cylinders"
    ## value labels only
    mtcars2[["am"]] <- factor(mtcars2[["am"]], c(0, 1), c("automatic", "manual"))
    ## variable label only
    attr(mtcars2[["mpg"]], "label") <- "miles per gallon"
    withr::with_tempfile("mtcars_file", fileext = ".sav", code = {
        export(mtcars2, mtcars_file)
        expect_true(file.exists(mtcars_file))
        expect_true(d <- is.data.frame(import(mtcars_file)))
        expect_true(!"labelled" %in% unlist(lapply(d, class)))
        ##Variable label and value labels preserved on SPSS (.sav) roundtrip
        d <- import(mtcars_file)
        a_cyl <- attributes(d[["cyl"]])
        expect_true("label" %in% names(a_cyl))
        expect_true("labels" %in% names(a_cyl))
        expect_true(identical(a_cyl[["label"]], "cylinders"))
        expect_true(identical(a_cyl[["labels"]], stats::setNames(c(1.0, 2.0, 3.0), c("four", "six", "eight"))))
        a_am <- attributes(d[["am"]])
        expect_true("labels" %in% names(a_am))
        expect_true(identical(a_am[["labels"]], stats::setNames(c(1.0, 2.0), c("automatic", "manual"))))

        a_mpg <- attributes(d[["mpg"]])
        expect_true("label" %in% names(a_mpg))
        expect_true(identical(a_mpg[["label"]], "miles per gallon"))
        ##haven is deprecated"
        lifecycle::expect_deprecated(import(mtcars_file, haven = TRUE))
        lifecycle::expect_deprecated(import(mtcars_file, haven = FALSE))
    })
})

test_that("Export to SPSS compressed (.zsav)", {
    mtcars2 <- mtcars
    ## label and value labels
    mtcars2[["cyl"]] <- factor(mtcars2[["cyl"]], c(4, 6, 8), c("four", "six", "eight"))
    attr(mtcars2[["cyl"]], "label") <- "cylinders"
    ## value labels only
    mtcars2[["am"]] <- factor(mtcars2[["am"]], c(0, 1), c("automatic", "manual"))
    ## variable label only
    attr(mtcars2[["mpg"]], "label") <- "miles per gallon"
    withr::with_tempfile("mtcars_file", fileext = ".zsav", code = {
        export(mtcars2, mtcars_file)
        expect_true(file.exists(mtcars_file))
        expect_true(d <- is.data.frame(import(mtcars_file)))
        expect_true(!"labelled" %in% unlist(lapply(d, class)))
        d <- import(mtcars_file)
        a_cyl <- attributes(d[["cyl"]])
        expect_true("label" %in% names(a_cyl))
        expect_true("labels" %in% names(a_cyl))
        expect_true(identical(a_cyl[["label"]], "cylinders"))
        expect_true(identical(a_cyl[["labels"]], stats::setNames(c(1.0, 2.0, 3.0), c("four", "six", "eight"))))

        a_am <- attributes(d[["am"]])
        expect_true("labels" %in% names(a_am))
        expect_true(identical(a_am[["labels"]], stats::setNames(c(1.0, 2.0), c("automatic", "manual"))))

        a_mpg <- attributes(d[["mpg"]])
        expect_true("label" %in% names(a_mpg))
        expect_true(identical(a_mpg[["label"]], "miles per gallon"))
    })
})

test_that("Export to and import from XBASE (.dbf)", {
    skip_if_not_installed("foreign")
    withr::with_tempfile("iris_file", fileext = ".dbf", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        d <- import(iris_file)
        expect_true(is.data.frame(d))
        expect_true(!"factor" %in% vapply(d, class, character(1)))
    })
})

skip_if_not_installed("qs2")

test_that("Export to and import from qs2", {
    withr::with_tempfile("iris_file", fileext = ".qs2", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
    })
})

test_that("mapping; both base and tidy conventions work", {
    withr::with_tempfile("tempxlsx", fileext = ".xlsx", code = {
        export(list("mtcars" = mtcars, "iris" = iris), tempxlsx)
        expect_error(y <- import(tempxlsx, n_max = 42, sheet = "iris"), NA)
        expect_equal(nrow(y), 42)
        expect_error(y2 <- import(tempxlsx, n_max = 42, which = "iris"), NA)
        expect_equal(nrow(y2), 42)
        expect_equal(y, y2)
        expect_error(y <- import(tempxlsx, n_max = 42, col_names = FALSE, which = 2), NA)
        expect_equal(nrow(y), 42)
        expect_error(y2 <- import(tempxlsx, n_max = 42, header = FALSE, which = 2), NA)
        expect_equal(y, y2)
    })
})

test_that("Unused arguments are by default ignored silently", {
    withr::with_tempfile("tempxlsx", fileext = ".xlsx", code = {
        export(list("mtcars" = mtcars, "iris" = iris), tempxlsx)
        expect_error(y <- import(tempxlsx, n_max = 42, whatever = TRUE, sheet = 2), NA)
    })
})

test_that("Unused arguments with option", {
    withr::with_tempfile("tempxlsx", fileext = ".xlsx", code = {
        export(list("mtcars" = mtcars, "iris" = iris), tempxlsx)
        expect_error(R.utils::withOptions({
            y <- import(tempxlsx, n_max = 42, whatever = TRUE)
        }, rio.ignoreunusedargs = FALSE))
        expect_error(R.utils::withOptions({
            y <- import(tempxlsx, n_max = 42, sheet = 2, whatever = TRUE)
        }, rio.ignoreunusedargs = FALSE), "whatever") ## not sheet
    })
})

skip_if_not_installed("nanoparquet")
skip_on_os("mac") ## apache/arrow#40991

test_that("Export to and import from parquet", {
    withr::with_tempfile("iris_path", fileext = ".parquet", code = {
        export(iris, iris_path)
        expect_true(file.exists(iris_path))
        expect_true(is.data.frame(import(iris_path)))
    })
})

test_that("Data identical (text formats)", {
    withr::with_tempdir(code = {
        expect_equivalent(import(export(mtcars, "mtcars.txt")), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.csv")), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.tsv")), mtcars)
    })
})

test_that("Data identical (R formats)", {
    withr::with_tempdir(code = {
        expect_equivalent(import(export(mtcars, "mtcars.rds"), trust = TRUE), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.R"), trust = TRUE), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.RData"), trust = TRUE), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.R", format = "dump"), trust = TRUE), mtcars)
    })
})

test_that("Data identical (R formats), feather", {
    skip_if_not_installed("arrow")
    withr::with_tempdir(code = {
        expect_equivalent(import(export(mtcars, "mtcars.feather")), mtcars)
    })
})

test_that("Data identical (haven formats)", {
    withr::with_tempdir(code = {
        expect_equivalent(import(export(mtcars, "mtcars.dta")), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.sav")), mtcars)
    })
})

test_that("Data identical (Excel formats)", {
    withr::with_tempdir(code = {
        expect_equivalent(import(export(mtcars, "mtcars.xlsx")), mtcars)
    })
})

test_that("Data identical (other formats)", {
    skip_if_not_installed("xml2")
    skip_if_not_installed("jsonlite")
    withr::with_tempdir(code = {
        expect_equivalent(import(export(mtcars, "mtcars.dbf")), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.json")), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.arff")), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.xml")), mtcars)
    })
})

test_that("Data identical (optional arguments)", {
    withr::with_tempdir(code = {
        ##expect_equivalent(import(export(mtcars, "mtcars.csv", format = "csv2"), format = "csv2"), mtcars)
        expect_equivalent(import(export(mtcars, "mtcars.csv"), nrows = 4), mtcars[1:4,])
        expect_equivalent(import(export(mtcars, "mtcars.csv", format = "tsv"), format = "tsv"), mtcars)
        expect_true(all.equal(import(export(mtcars, "mtcars", format = "csv"), format = "csv"), mtcars, check.attributes = FALSE))
    })
})

test_that("Export to Excel (.xlsx)", {
    withr::with_tempfile("iris_file", fileext = ".xlsx", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
        expect_true(is.data.frame(import(iris_file, sheet = 1)))
        expect_true(is.data.frame(import(iris_file, which = 1)))
        expect_true(nrow(import(iris_file, n_max = 42)) == 42)
    })
})

test_that("Expert to Excel (.xlsx) a list", {
    withr::with_tempfile("tempxlsx", fileext = ".xlsx", code = {
        export(list(
            mtcars3 = mtcars[1:10, ],
            mtcars2 = mtcars[11:20, ],
            mtcars1 = mtcars[21:32, ]
        ), tempxlsx)
        expect_equal(readxl::excel_sheets(tempxlsx), c("mtcars3", "mtcars2", "mtcars1"))
    })
})

test_that("Is `sheet` passed?", {
    withr::with_tempfile("tempxlsx", fileext = ".xlsx", code = {
        export(list(
            mtcars3 = mtcars[1:10, ],
            mtcars2 = mtcars[11:20, ],
            mtcars1 = mtcars[21:32, ]
        ), tempxlsx)
        expect_equal(readxl::excel_sheets(tempxlsx), c("mtcars3", "mtcars2", "mtcars1"))
        content <- import(tempxlsx, sheet = "mtcars2")
        expect_equal(content$mpg, mtcars[11:20, ]$mpg)
        content <- import(tempxlsx, which = 2)
        expect_equal(content$mpg, mtcars[11:20, ]$mpg)
    })
})


test_that("readxl is deprecated", {
    withr::with_tempfile("iris_file", fileext = ".xlsx", code = {
        export(iris, iris_file)
        lifecycle::expect_deprecated(import(iris_file, readxl = TRUE))
        lifecycle::expect_deprecated(import(iris_file, readxl = FALSE))
    })
})

test_that("Import from Excel (.xls)", {
    expect_true(is.data.frame(import("../testdata/iris.xls")))
    expect_true(is.data.frame(import("../testdata/iris.xls", sheet = 1)))
    expect_true(is.data.frame(import("../testdata/iris.xls", which = 1)))
})

test_that("Import from Excel (.xlsm)", {
    expect_true(is.data.frame(import("../testdata/example.xlsm", sheet = 1, format = "xlsx")))
    expect_true(is.data.frame(import("../testdata/example.xlsm", which = 1, format = "xlsx")))
    expect_true(is.data.frame(import("../testdata/example.xlsm")))

    expect_equal(
        import_list("../testdata/example.xlsm", format = "xlsx"),
        list(
            Sheet1 = data.frame(A = 1),
            Sheet2 = data.frame(B = 2)
        )
    )
    expect_equal(
        import_list("../testdata/example.xlsm"),
        list(
            Sheet1 = data.frame(A = 1),
            Sheet2 = data.frame(B = 2)
        )
    )
})

test_that("Exporting to xlsm is not supported #469", {
    withr::with_tempfile("iris_file", fileext = ".xlsm", code = {
        expect_error(export(iris, iris_file), "Format not supported")
    })
})

test_that("Export to and import from CSVY", {
    withr::with_tempfile("iris_file", fileext = ".csvy", code = {
        suppressWarnings(export(iris, iris_file))
        expect_true(file.exists(export(iris, iris_file)))
        suppressWarnings(d <- import(iris_file))
        expect_true(is.data.frame(d))
    })
})

test_that("Export to and import from Rdata", {
    withr::with_tempfile("iris_file", fileext = ".Rdata", code = {
        ## data frame
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, trust = TRUE)))
        expect_true(is.data.frame(import(iris_file, which = 1, trust = TRUE)))
    })
    withr::with_tempfile("iris_file", fileext = ".Rdata", code = {
        ## environment
        e <- new.env()
        e$iris <- iris
        export(e, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, trust = TRUE)))
        expect_true(is.data.frame(import(iris_file, which = 1, trust = TRUE)))
    })
    withr::with_tempfile("iris_file", fileext = ".Rdata", code = {
        ## character
        export("iris", iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, trust = TRUE)))
        expect_true(is.data.frame(import(iris_file, which = 1, trust = TRUE)))
    })
    withr::with_tempfile("iris_file", fileext = ".Rdata", code = {
        ## expect error otherwise
        expect_error(export(iris$Species, iris_file))
    })
})

test_that("Export to and import from rda", {
    withr::with_tempfile("iris_file", fileext = ".rda", code = {
        ## data frame
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, trust = TRUE)))
        expect_true(is.data.frame(import(iris_file, which = 1, trust = TRUE)))
    })
    withr::with_tempfile("iris_file", fileext = ".rda", code = {
        ## environment
        e <- new.env()
        e$iris <- iris
        export(e, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, trust = TRUE)))
        expect_true(is.data.frame(import(iris_file, which = 1, trust = TRUE)))
    })
    withr::with_tempfile("iris_file", fileext = ".rda", code = {
        ## character
        export("iris", iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, trust = TRUE)))
        expect_true(is.data.frame(import(iris_file, which = 1, trust = TRUE)))
    })
    withr::with_tempfile("iris_file", fileext = ".rda", code = {
        ## expect error otherwise
        expect_error(export(iris$Species, iris_file))
    })
})

test_that("export to nonexisting directory #347", {
    withr::with_tempdir({
        output_dir <- file.path(".", "this_doesnt_exist")
        expect_false(dir.exists(output_dir))
        expect_error(export(iris, file.path(output_dir, "iris.csv")), NA)
        expect_true(file.exists(file.path(output_dir, "iris.csv")))
    })
})

test_that("export to existing directory (contra previous one) #347", {
    withr::with_tempdir({
        output_dir <- file.path(".", "this_surely_exist1")
        dir.create(output_dir, recursive = TRUE) ## right?
        expect_true(dir.exists(output_dir))
        expect_error(export(iris, file.path(output_dir, "iris.csv")), NA)
        expect_true(file.exists(file.path(output_dir, "iris.csv")))
    })
})

test_that("export to nonexisting directory also for compressed file  #347", {
    withr::with_tempdir({
        output_dir <- file.path(".", "this_doesnt_exist")
        expect_false(dir.exists(output_dir))
        expect_error(export(iris, file.path(output_dir, "iris.csv.gz")), NA)
        expect_true(file.exists(file.path(output_dir, "iris.csv.gz")))
    })
})

require("datasets")

test_that("Export to CSV", {
    withr::with_tempfile("iris_file", fileext = ".csv", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
    })
})

test_that("Export (Append) to CSV", {
    withr::with_tempfile("iris_file", fileext = ".csv", code = {
        export(iris, iris_file)
        nlines <- length(readLines(iris_file))
        export(iris, iris_file, append = FALSE)
        expect_true(identical(length(readLines(iris_file)), nlines))
        export(iris, iris_file, append = TRUE)
        expect_true(identical(length(readLines(iris_file)), (2L * nlines) - 1L))
    })
})

test_that("Import from CSV", {
    noheadercsv <- import("../testdata/noheader.csv", header = FALSE)
    expect_that(colnames(noheadercsv)[1], equals("V1"), label = "Header is correctly specified")
})

test_that("Import from (European-style) CSV with semicolon separator", {
    withr::with_tempfile("iris_file", fileext = ".csv", code = {
        write.table(iris, iris_file, dec = ",", sep = ";", row.names = FALSE)
        expect_true(file.exists(iris_file))
        ## import works (even if column classes are incorrect)
        expect_true(is.data.frame(import(iris_file, header = TRUE)))
        iris_imported <- import(iris_file, format = ";", header = TRUE)
        ## import works with correct, numeric column classes
        expect_true(is.data.frame(iris_imported))
        expect_true(is.numeric(iris_imported[["Sepal.Length"]]))
    })
})

test_that("Export to and Import from CSV2", {
    withr::with_tempfile("iris_file", fileext = ".csv", code = {
        export(iris, iris_file, format = "csv2")
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file, format = "csv2")))
    })
})

test_that("Export to and Import from TSV with CSV extension", {
    withr::with_tempfile("iris_file", fileext = ".csv", code = {
        export(iris, iris_file, format = "tsv")
        expect_true(file.exists(iris_file))
        expect_true(ncol(import(iris_file)) == 5L)
        expect_true(ncol(import(iris_file, format = "tsv")) == 5L)
        expect_true(ncol(import(iris_file, format = "tsv", sep = "\t")) == 5L)
        expect_true(ncol(import(iris_file, sep = ",")) == 5L) # use `data.table::fread(sep = "auto")` even if `sep` set explicitly to ","
        expect_true(ncol(import(iris_file, format = "csv")) == 5L)
        expect_true(ncol(import(iris_file, sep = "auto")) == 5L)
    })
})

test_that("fread is deprecated", {
    withr::with_tempfile("iris_file", fileext = ".csv", code = {
        export(iris, iris_file)
        lifecycle::expect_deprecated(import(iris_file, fread = TRUE))
        lifecycle::expect_deprecated(import(iris_file, fread = FALSE))
    })
})

test_that("dat (ambiguous file format) will be attempted #430", {
    withr::with_tempfile("iris_file", fileext = ".dat", code = {
        readr::write_delim(iris, file = iris_file)
        expect_error(suppressMessages(import(iris_file)), NA)
        expect_message(z <- import(iris_file))
        ## export is not supported though
        expect_error(export(iris, iris_file))
    })
})

skip_if_not_installed("readODS")

test_that("Import from ODS", {
    ods0 <- import("../testdata/mtcars.ods")
    expect_error(ods <- import("../testdata/mtcars.ods",
                               sheet = 1, col_names = TRUE,
                               invalid_argument = 42), NA)
    expect_identical(ods0, ods, label = "ODS import ignored arguments don't affect output")
    expect_true(is.data.frame(ods), label = "ODS import returns data.frame")
    expect_true(identical(names(mtcars), names(ods)), label = "ODS import returns correct names")
    expect_true(identical(dim(mtcars), dim(ods)), label = "ODS import returns correct dimensions")
    expect_equivalent(ods, mtcars, label = "ODS import returns correct values")
})

test_that("Export to and import from ODS", {
    withr::with_tempfile("iris_file", fileext = ".ods", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
    })
})

test_that("... correctly passed #318", {
    withr::with_tempfile("mtcars_file", fileext = ".ods", code = {
        export(iris, mtcars_file, sheet = "mtcars")
        expect_equal(readODS::list_ods_sheets(mtcars_file), "mtcars")
    })
})

test_that("Export and Import FODS", {
    withr::with_tempfile("fods_file", fileext = ".fods", code = {
        export(iris, fods_file)
        expect_true(file.exists(fods_file))
        expect_true(is.data.frame(import(fods_file)))
        export(iris, fods_file, sheet = "mtcars")
        expect_equal(readODS::list_fods_sheets(fods_file), "mtcars")
        expect_error(y <- import(fods_file), NA)
        expect_true(is.data.frame(y))

    })
})

test_that("Export list of data frames", {
    withr::with_tempfile("ods_files", fileext = c(".ods", ".fods"), code = {
        dfs <- list("cars" = mtcars, "flowers" = iris)
        expect_error(export(dfs, ods_files[1]), NA)
        expect_error(export(dfs, ods_files[2]), NA)
        expect_equal(import(ods_files[1], which = "flowers"), import(ods_files[2], which = "flowers"))
    })
})

skip_if_not_installed("xml2")

test_html <- function(breaker = "&") {
    mtcars2 <- mtcars
    colnames(mtcars2)[1] <- paste0("mp", breaker, breaker, "g")
    mtcars2[1,1] <- paste0("mp", breaker, breaker, "g")
    withr::with_tempfile("mtcars_file", fileext = ".html", code = {
        expect_error(x <- rio::export(mtcars2, mtcars_file), NA)
        temp_df <- rio::import(x)
        expect_equal(colnames(temp_df)[1], paste0("mp", breaker, breaker, "g"))
        expect_equal(temp_df[1,1], paste0("mp", breaker, breaker, "g"))
    })
}

test_that("Export to and import from HTML", {
    withr::with_tempfile("iris_file", fileext = ".html", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)), label = "import from single-table html works")
    })
})

test_that("Export to HTML with ampersands",{
    withr::with_tempfile("iris_file", fileext = ".html", code = {

        iris$`R & D` <- paste(sample(letters,nrow(iris), replace = TRUE),
                              "&",
                              sample(LETTERS,nrow(iris), replace = TRUE))
        export(iris, iris_file)
        expect_true(file.exists(iris_file),
                    label = "export to html with ampersands works")
    })
})

test_that("Import from HTML", {
    f <- "../testdata/twotables.html"
    expect_true(all(dim(import(f, which = 1)) == c(32, 11)), label = "import from two-table html works (which = 1)")
    expect_true(all(dim(import(f, which = 2)) == c(150, 5)), label = "import from two-table html works (which = 2)")
})

test_that("Import from HTML with multiple tbody elements", {
    expect_true(is.data.frame(import("../testdata/two-tbody.html")), label="import with two tbody elements in a single html table works")
    expect_true(is.data.frame(import("../testdata/br-in-header.html")), label="import with an empty header cell in an html table works")
    expect_true(is.data.frame(import("../testdata/br-in-td.html")), label="import with an empty data cell in a single html table works")
    expect_true(is.data.frame(import("../testdata/th-as-row-element.html")), label="import with th instead of td in a non-header row in a single html table works")
})

test_that("html with &, >, ', \", >, <", {
    ## test all
    useless <- lapply(c("&", "\"", "'", "<", ">"), test_html)
})

library("datasets")
skip_on_cran()

test_that("S3 extension mechanism works for imports", {
    withr::with_tempdir({
        write.csv(iris, "iris.custom")
        expect_error(import("iris.custom"), "Format not supported")
        .import.rio_custom <- function(file, ...) {
            read.csv(file, ...)
        }
        expect_error(.import.rio_custom("iris.custom"), NA)
        assign(".import.rio_custom", .import.rio_custom, envir = .GlobalEnv)
        expect_true(is.data.frame(import("iris.custom")))
        rm(list = ".import.rio_custom", envir = .GlobalEnv)
    })
})

test_that("S3 extension mechanism works for exports", {
    withr::with_tempdir({
        expect_error(export(iris, "iris.custom"))
        .export.rio_custom <- function(file, x, ...) {
            write.csv(x, file, ...)
            invisible(file)
        }
        expect_error(.export.rio_custom("iris.custom", iris), NA)
        assign(".export.rio_custom", .export.rio_custom, envir = .GlobalEnv)
        expect_true(is.character(export(iris, "iris.custom")))
        rm(list = ".export.rio_custom", envir = .GlobalEnv)
    })
})

test_that("Export to and import from PSV", {
    withr::with_tempfile("iris_file", fileext = ".psv", code = {
        export(iris, iris_file)
        expect_true(file.exists(iris_file))
        expect_true(is.data.frame(import(iris_file)))
    })
})


test_that("fread is deprecated", {
    withr::with_tempfile("iris_file", fileext = ".psv", code = {
        export(iris, iris_file)
        lifecycle::expect_deprecated(import(iris_file, fread = TRUE))
        lifecycle::expect_deprecated(import(iris_file, fread = FALSE))
    })

})

mtcars_tibble <- tibble::as_tibble(mtcars)
mtcars_datatable <- data.table::as.data.table(mtcars)

test_that("Set object class", {
    mtcars_tibble <- tibble::as_tibble(mtcars)
    mtcars_datatable <- data.table::as.data.table(mtcars)
    expect_true(inherits(set_class(mtcars), "data.frame"))
    expect_true(inherits(set_class(mtcars_tibble), "data.frame"))
    expect_true(inherits(set_class(mtcars_datatable), "data.frame"))
    expect_true(inherits(set_class(mtcars, class = "fakeclass"), "data.frame"))
    expect_true(!"fakeclass" %in% class(set_class(mtcars, class = "fakeclass")))
})

test_that("Set object class as tibble", {
    mtcars_tibble <- tibble::as_tibble(mtcars)
    mtcars_datatable <- data.table::as.data.table(mtcars)
    expect_true(inherits(set_class(mtcars, class = "tbl_df"), "tbl_df"))
    expect_true(inherits(set_class(mtcars, class = "tibble"), "tbl_df"))
    expect_true(inherits(set_class(mtcars_tibble, class = "tibble"), "tbl_df"))
})

test_that("Set object class as data.table", {
    expect_true(inherits(set_class(mtcars, class = "data.table"), "data.table"))
    withr::with_tempfile("data_file", fileext = ".csv", code = {
        export(mtcars, data_file)
        expect_true(inherits(import(data_file, setclass = "data.table"), "data.table"))
        expect_true(inherits(import(data_file, data.table = TRUE, setclass = "data.table"), "data.table"))
    })
})

test_that("Set object class as arrow table", {
    skip_if(getRversion() <= "4.2")
    skip_if_not_installed("arrow")
    mtcars_arrow <- arrow::arrow_table(mtcars)
    expect_false(inherits(set_class(mtcars_arrow), "data.frame")) ## arrow table is not data.frame
    expect_true(inherits(set_class(mtcars, class = "arrow"), "ArrowTabular"))
    expect_true(inherits(set_class(mtcars, class = "arrow_table"), "ArrowTabular"))
    withr::with_tempfile("data_file", fileext = ".csv", code = {
        export(mtcars, data_file)
        expect_true(inherits(import(data_file, setclass = "arrow"), "ArrowTabular"))
        expect_true(inherits(import(data_file, data.table = TRUE, setclass = "arrow"), "ArrowTabular"))
    })
})

test_that("ArrowTabular can be exported", {
    skip_if(getRversion() <= "4.2")
    skip_if_not_installed("arrow")
    mtcars_arrow <- arrow::arrow_table(mtcars)
    withr::with_tempfile("data_file", fileext = ".csv", code = {
        expect_error(export(mtcars_arrow, data_file), NA) ## no concept of rownames
        expect_true(inherits(import(data_file), "data.frame"))
    })
})

test_that("Simulate arrow is not installed, #376", {
    ## although this is pretty meaningless
    withr::with_tempfile("data_file", fileext = ".csv", code = {
        with_mocked_bindings({
            export(mtcars, data_file)
            expect_error(import(data_file, setclass = "arrow"), "Suggested package")
        }, .check_pkg_availability = function(pkg, lib.loc = NULL) {
            stop("Suggested package `", pkg, "` is not available. Please install it individually or use `install_formats()`", call. = FALSE)
        })
    })
})

library("datasets")

test_that("Basic file conversion", {
    withr::with_tempfile("mtcars_files", fileext = c(".dta", ".csv"), code = {
        export(mtcars, mtcars_files[1])
        convert(mtcars_files[1], mtcars_files[2])
        convert(mtcars_files[2], mtcars_files[1])
        x <- import(mtcars_files[1])
        expect_true(identical(names(mtcars), names(x)))
        expect_true(identical(dim(mtcars), dim(x)))
    })
})

test_that("File conversion with arguments", {
    withr::with_tempfile("mtcars_files", fileext = c(".csv", ".tsv"), code = {
        export(mtcars, mtcars_files[1], format = "tsv")
        convert(mtcars_files[1], mtcars_files[1], in_opts = list(format = "tsv"))
        expect_true(file.exists(mtcars_files[1]))
        expect_true(!file.exists(mtcars_files[2]))
        convert(mtcars_files[1], mtcars_files[2],
                in_opts = list(format = "tsv"), out_opts = list(format = "csv"))
        expect_true(file.exists(mtcars_files[2]))
    })
})

test_that("File conversion w/o out_file errors", {
    withr::with_tempfile("mtcars_file", fileext = ".dta", {
        export(mtcars, mtcars_file)
        expect_error(convert(mtcars_file))
    })
})

#' @rdname import
#' @title Import
#' @description Read in a data.frame from a file. Exceptions to this rule are Rdata, RDS, and JSON input file formats, which return the originally saved object without changing its class.
#' @param file A character string naming a file, URL, or single-file (can be Gzip or Bzip2 compressed), .zip or .tar archive.
#' @param format An optional character string code of file format, which can be used to override the format inferred from `file`. Shortcuts include: \dQuote{,} (for comma-separated values), \dQuote{;} (for semicolon-separated values), and \dQuote{|} (for pipe-separated values).
#' @param setclass An optional character vector specifying one or more classes
#' to set on the import. By default, the return object is always a
#' \dQuote{data.frame}. Allowed values include \dQuote{tbl_df}, \dQuote{tbl}, or
#' \dQuote{tibble} (if using tibble), \dQuote{arrow}, \dQuote{arrow_table} (if using arrow table; the suggested package `arrow` must be installed) or \dQuote{data.table} (if using
#' data.table). Other values are ignored, such that a data.frame is returned.
#' The parameter takes precedents over parameters in \dots which set a different class.
#' @param which This argument is used to control import from multi-object files; as a rule `import` only ever returns a single data frame (use [import_list()] to import multiple data frames from a multi-object file). If `file` is an archive format (zip and tar), `which` can be either a character string specifying a filename or an integer specifying which file (in locale sort order) to extract from the compressed directory. But please see the section `which` below. For Excel spreadsheets, this can be used to specify a sheet name or number. For .Rdata files, this can be an object name. For HTML files, it identifies which table to extract (from document order). Ignored otherwise. A character string value will be used as a regular expression, such that the extracted file is the first match of the regular expression against the file names in the archive.
#' @param \dots Additional arguments passed to the underlying import functions. For example, this can control column classes for delimited file types, or control the use of haven for Stata and SPSS or readxl for Excel (.xlsx) format. See details below.
#' @return A data frame. If `setclass` is used, this data frame may have additional class attribute values, such as \dQuote{tibble} or \dQuote{data.table}.
#' @details This function imports a data frame or matrix from a data file with the file format based on the file extension (or the manually specified format, if `format` is specified).
#'
#' `import` supports the following file formats:
#'
#' \itemize{
#'     \item Comma-separated data (.csv), using [data.table::fread()]
#'     \item Pipe-separated data (.psv), using [data.table::fread()]
#'     \item Tab-separated data (.tsv), using [data.table::fread()]
#'     \item SAS (.sas7bdat), using [haven::read_sas()]
#'     \item SAS XPORT (.xpt), using [haven::read_xpt()]
#'     \item SPSS (.sav), using [haven::read_sav()]
#'     \item SPSS compressed (.zsav), using [haven::read_sav()].
#'     \item Stata (.dta), using [haven::read_dta()]
#'     \item SPSS Portable Files (.por), using [haven::read_por()].
#'     \item Excel (.xls and .xlsx), using [readxl::read_xlsx()] or [readxl::read_xls()]. Use `which` to specify a sheet number.
#'     \item R syntax object (.R), using [base::dget()], see `trust` below.
#'     \item Saved R objects (.RData,.rda), using [base::load()] for single-object .Rdata files. Use `which` to specify an object name for multi-object .Rdata files. This can be any R object (not just a data frame), see `trust` below.
#'     \item Serialized R objects (.rds), using [base::readRDS()]. This can be any R object (not just a data frame), see `trust` below.
#'     \item Serialized R objects (.qs2), using [qs2::qs_read()].
#'     \item Epiinfo (.rec), using [foreign::read.epiinfo()]
#'     \item Minitab (.mtp), using [foreign::read.mtp()]
#'     \item Systat (.syd), using [foreign::read.systat()]
#'     \item "XBASE" database files (.dbf), using [foreign::read.dbf()]
#'     \item Weka Attribute-Relation File Format (.arff), using [foreign::read.arff()]
#'     \item Data Interchange Format (.dif), using [utils::read.DIF()]
#'     \item Fortran data (no recognized extension), using [utils::read.fortran()]
#'     \item Fixed-width format data (.fwf), using a faster version of [utils::read.fwf()] that requires a `widths` argument and by default in rio has `stringsAsFactors = FALSE`
#'     \item [CSVY](https://github.com/csvy) (CSV with a YAML metadata header) using [data.table::fread()].
#'     \item Apache Arrow Parquet (.parquet), using [nanoparquet::read_parquet()]
#'     \item Feather R/Python interchange format (.feather), using [arrow::read_feather()]
#'     \item Fast storage (.fst), using [fst::read.fst()]
#'     \item JSON (.json), using [jsonlite::fromJSON()]
#'     \item Matlab (.mat), using [rmatio::read.mat()]
#'     \item EViews (.wf1), using [hexView::readEViews()]
#'     \item OpenDocument Spreadsheet (.ods, .fods), using [readODS::read_ods()] or [readODS::read_fods()].  Use `which` to specify a sheet number.
#'     \item Single-table HTML documents (.html), using [xml2::read_html()]. There is no standard HTML table and we have only tested this with HTML tables exported with this package. HTML tables will only be read correctly if the HTML file can be converted to a list via [xml2::as_list()]. This import feature is not robust, especially for HTML tables in the wild. Please use a proper web scraping framework, e.g. `rvest`.
#'     \item Shallow XML documents (.xml), using [xml2::read_xml()]. The data structure will only be read correctly if the XML file can be converted to a list via [xml2::as_list()].
#'     \item YAML (.yml), using [yaml::yaml.load()]
#'     \item Clipboard import, using [utils::read.table()] with `row.names = FALSE`
#'     \item Google Sheets, as Comma-separated data (.csv)
#'     \item GraphPad Prism (.pzfx) using [pzfx::read_pzfx()]
#' }
#'
#' `import` attempts to standardize the return value from the various import functions to the extent possible, thus providing a uniform data structure regardless of what import package or function is used. It achieves this by storing any optional variable-related attributes at the variable level (i.e., an attribute for `mtcars$mpg` is stored in `attributes(mtcars$mpg)` rather than `attributes(mtcars)`). If you would prefer these attributes to be stored at the data.frame-level (i.e., in `attributes(mtcars)`), see [gather_attrs()].
#'
#' After importing metadata-rich file formats (e.g., from Stata or SPSS), it may be helpful to recode labelled variables to character or factor using [characterize()] or [factorize()] respectively.
#'
#' # Trust
#' For serialization formats (.R, .RDS, and .RData), please note that you should only load these files from trusted sources. It is because these formats are not necessarily for storing rectangular data and can also be used to store many things, e.g. code. Importing these files could lead to arbitary code execution. Please read the security principles by the R Project (Plummer, 2024). When importing these files via `rio`, you should affirm that you trust these files, i.e. `trust = TRUE`. See example below. If this affirmation is missing, the current version assumes `trust` to be true for backward compatibility and a deprecation notice will be printed. In the next major release (2.0.0), you must explicitly affirm your trust when importing these files.
#'
#' # Which
#' For compressed archives (zip and tar, where a compressed file can contain multiple files), it is possible to come to a situation where the parameter `which` is used twice to indicate two different concepts. For example, it is unclear for `.xlsx.zip`whether `which` refers to the selection of an exact file in the archive or the selection of an exact sheet in the decompressed Excel file. In these cases, `rio` assumes that `which` is only used for the selection of file. After the selection of file with `which`, `rio` will return the first item, e.g. the first sheet.
#'
#' Please note, however, `.gz` and `.bz2` (e.g. `.xlsx.gz`) are compressed, but not archive format. In those cases, `which` is used the same way as the non-compressed format, e.g. selection of sheet for Excel.
#'
#' @note For csv and txt files with row names exported from [export()], it may be helpful to specify `row.names` as the column of the table which contain row names. See example below.
#' @references
#' Plummer, M (2024). Statement on CVE-2024-27322. [https://blog.r-project.org/2024/05/10/statement-on-cve-2024-27322/](https://blog.r-project.org/2024/05/10/statement-on-cve-2024-27322/)
#' @examples
#' ## For demo, a temp. file path is created with the file extension .csv
#' csv_file <- tempfile(fileext = ".csv")
#' ## .xlsx
#' xlsx_file <- tempfile(fileext = ".xlsx")
#' ## create CSV to import
#' export(iris, csv_file)
#' ## specify `format` to override default format: see export()
#' export(iris, xlsx_file, format = "csv")
#'
#' ## basic
#' import(csv_file)
#'
#' ## You can certainly import your data with the file name, which is not a variable:
#' ## import("starwars.csv"); import("mtcars.xlsx")
#'
#' ## Override the default format
#' ## import(xlsx_file) # Error, it is actually not an Excel file
#' import(xlsx_file, format = "csv")
#'
#' ## import CSV as a `data.table`
#' import(csv_file, setclass = "data.table")
#'
#' ## import CSV as a tibble (or "tbl_df")
#' import(csv_file, setclass = "tbl_df")
#'
#' ## pass arguments to underlying import function
#' ## data.table::fread is the underlying import function and `nrows` is its argument
#' import(csv_file, nrows = 20)
#'
#' ## data.table::fread has an argument `data.table` to set the class explicitely to data.table. The
#' ## argument setclass, however, takes precedents over such undocumented features.
#' class(import(csv_file, setclass = "tibble", data.table = TRUE))
#'
#' ## the default import class can be set with options(rio.import.class = "data.table")
#' ## options(rio.import.class = "tibble"), or options(rio.import.class = "arrow")
#'
#' ## Security
#' rds_file <- tempfile(fileext = ".rds")
#' export(iris, rds_file)
#'
#' ## You should only import serialized formats from trusted sources
#' ## In this case, you can trust it because it's generated by you.
#' import(rds_file, trust = TRUE)
#' @seealso [import_list()], [characterize()], [gather_attrs()], [export()], [convert()]
#' @export
import <- function(file, format, setclass = getOption("rio.import.class", "data.frame"), which, ...) {
    if (setclass %in% c("arrow", "arrow_table")) {
        .check_pkg_availability("arrow")
    }
    .check_file(file, single_only = TRUE)
    if (R.utils::isUrl(file)) {
        file <- remote_to_local(file, format = format)
    }
    if ((file != "clipboard") && !file.exists(file)) {
        stop("No such file: ", file, call. = FALSE)
    }
    ## compressed file, f is a pretty bad name; but export() uses it.
    f <- find_compress(file)
    if (!is.na(f$compress)) {
        cfile <- file
        file <- f$file
        which <- ifelse(missing(which), 1, which)
        file <- parse_archive(cfile, which = which, file_type = f$compress)
        format <- .get_compressed_format(cfile, file, f$compress, format)
        ## reset which if `file` is zip or tar. #412
        which <- .reset_which(file_type = f$compress, which = which)
    }
    if (missing(format)) {
        format <- get_info(file)$format
    } else {
        ## format such as "|"
        format <- .standardize_format(format)
    }
    class(file) <- c(paste0("rio_", format), class(file))
    if (missing(which)) {
        x <- .import(file = file, ...)
    } else {
        x <- .import(file = file, which = which, ...)
    }

    # if R serialized object, just return it without setting object class
    if (inherits(file, c("rio_rdata", "rio_rds", "rio_json", "rio_qs", "rio_qs2")) && !inherits(x, "data.frame")) {
        return(x)
    }
    # otherwise, make sure it's a data frame (or requested class)
    return(set_class(x, class = setclass))
}

test_that("External read functions without an import method", {
    extensions <- c("bib", "bmp", "gexf", "gnumeric", "jpeg", "jpg", "npy", "png", "sss", "sdmx", "tiff")
    for (ext in extensions) {
        withr::with_tempfile("dummyfile", fileext = paste0(".", ext), code = {
            file.create(dummyfile)
            expect_error(import(dummyfile))
        })
    }
})

test_that("import method exported by an external package", {
    extensions <- c("bean", "beancount", "ledger", "hledger")
    for (ext in extensions) {
        withr::with_tempfile("dummyfile", fileext = paste0(".", ext), code = {
            file.create(dummyfile)
            expect_error(import(dummyfile))
        })
    }
})

test_that("Import from Fortran", {
    ## no fixed file extension
    withr::with_tempfile("fortran_file", code = {
        cat(file = fortran_file, "123456", "987654", sep = "\n")
        expect_true(is.data.frame(import(fortran_file, format = "fortran", style = c("F2.1","F2.0","I2"))))
    })
})

#' @title Install rio's \sQuote{Suggests} Dependencies
#' @description Not all suggested packages are installed by default. These packages are not installed or loaded by default in order to create a slimmer and faster package build, install, and load. Use `show_unsupported_formats()` to check all unsupported formats. `install_formats()` installs all missing \sQuote{Suggests} dependencies for rio that expand its support to the full range of support import and export formats.
#' @param \dots Additional arguments passed to [utils::install.packages()].
#' @return For `show_unsupported_formats()`, if there is any missing unsupported formats, it return TRUE invisibly; otherwise FALSE. For `install_formats()` it returns TRUE invisibly if the installation is succuessful; otherwise errors.
#' @examples
#' \donttest{
#' if (interactive()) {
#'     install_formats()
#' }
#' }
#' @export
install_formats <- function(...) {
    to_install <- uninstalled_formats()
    utils::install.packages(to_install, ...)
    invisible(TRUE)
}

uninstalled_formats <- function() {
    suggested_packages <- attr(rio_formats, "suggested_packages")
    ## which are not installed
    suggested_packages[!R.utils::isPackageInstalled(suggested_packages)]
}

#' @rdname install_formats
#' @export
show_unsupported_formats <- function() {
    ## default_formats <- sort(unique(rio_formats$format[rio_formats$type == "import"]))
    suggested_formats <- rio_formats[rio_formats$type == "suggest",]
    suggested_formats$pkg <- vapply(strsplit(suggested_formats$import_function, "::"), FUN = `[`, FUN.VALUE = character(1), 1)
    missing_pkgs <- uninstalled_formats()
    suggested_formats$installed <- vapply(suggested_formats$pkg, function(x) x %in% missing_pkgs, logical(1), USE.NAMES = FALSE)
    unsupported_formats <- suggested_formats[suggested_formats$installed,]
    if (nrow(unsupported_formats) == 0) {
        message("All default and optional formats are supported.")
        return(invisible(FALSE))
    }
    temp_display <- unsupported_formats[,c("input", "format", "pkg")]
    colnames(temp_display)[3] <- "Suggested package"
    print(temp_display)
    message("These formats are not supported. If you need to use these formats, please either install the suggested packages individually, or install_formats() to install them all.")
    return(invisible(TRUE))
}

test_that("uninstalled_formats()", {
    skip_on_cran()
    formats <- uninstalled_formats()
    if (is.null(formats)) {
        expect_true(install_formats())
    } else {
        expect_type(formats, "character")
    }
})


test_that("show_unsupported_formats (in the fully supported environment) on CI", {
    skip_on_cran()
    for (pkg in attr(rio_formats, "suggested_packages")) {
        skip_if_not_installed(pkg)
    }
    expect_false(show_unsupported_formats())
    expect_message(show_unsupported_formats(), "All default")
})

test_that("show_unsupported_formats (in the partial supported environment) on CI", {
    skip_on_cran()
    fake_uninstalled_formats <- function() {
        return(c("readODS"))
    }
    with_mocked_bindings(code = {
        expect_true(show_unsupported_formats())
        expect_message(show_unsupported_formats(), "These formats are not supported")
    }, `uninstalled_formats` = fake_uninstalled_formats)
})

#' @rdname characterize
#' @title Character conversion of labelled data
#' @description Convert labelled variables to character or factor
#' @param x A vector or data frame.
#' @param coerce_character A logical indicating whether to additionally coerce character columns to factor (in `factorize`). Default `FALSE`.
#' @param \dots additional arguments passed to methods
#' @details `characterize` converts a vector with a `labels` attribute of named levels into a character vector. `factorize` does the same but to factors. This can be useful at two stages of a data workflow: (1) importing labelled data from metadata-rich file formats (e.g., Stata or SPSS), and (2) exporting such data to plain text files (e.g., CSV) in a way that preserves information.
#' @return a character vector (for `characterize`) or factor vector (for `factorize`)
#' @examples
#' ## vector method
#' x <- structure(1:4, labels = c("A" = 1, "B" = 2, "C" = 3))
#' characterize(x)
#' factorize(x)
#'
#' ## data frame method
#' x <- data.frame(v1 = structure(1:4, labels = c("A" = 1, "B" = 2, "C" = 3)),
#'                 v2 = structure(c(1,0,0,1), labels = c("foo" = 0, "bar" = 1)))
#' str(factorize(x))
#' str(characterize(x))
#'
#' ## Application
#' csv_file <- tempfile(fileext = ".csv")
#' ## comparison of exported file contents
#' import(export(x, csv_file))
#' import(export(factorize(x), csv_file))
#' @seealso [gather_attrs()]
#' @export
characterize <- function(x, ...) {
    UseMethod("characterize")
}

#' @rdname characterize
#' @export
factorize <- function(x, ...) {
    UseMethod("factorize")
}

#' @rdname characterize
#' @export
characterize.default <- function(x, ...) {
    # retain variable label, if present
    if (!is.null(attributes(x)[["label"]])) {
        varlab <- attributes(x)[["label"]]
    } else {
        varlab <- NULL
    }

    if (!is.null(attributes(x)[["labels"]])) {
        x <- as.character(factorize(x, ...))
        if (!is.null(varlab)) {
            attr(x, "label") <- varlab
        }
    }

    return(x)
}

#' @rdname characterize
#' @export
characterize.data.frame <- function(x, ...) {
    x[] <- lapply(x, characterize, ...)
    x
}

#' @rdname characterize
#' @export
factorize.default <- function(x, coerce_character=FALSE, ...) {
    # retain variable label, if present
    if (!is.null(attributes(x)[["label"]])) {
        varlab <- attributes(x)[["label"]]
    } else {
        varlab <- NULL
    }

    if (!is.null(attributes(x)[["labels"]])) {
        x <- factor(x, attributes(x)[["labels"]], names(attributes(x)[["labels"]]), ...)
    } else if (is.character(x) && isTRUE(coerce_character)) {
        levs <- sort(unique(x))
        x <- factor(x, levs)
    }

    if (!is.null(varlab)) {
        attr(x, "label") <- varlab
    }

    return(x)
}

#' @rdname characterize
#' @export
factorize.data.frame <- function(x, ...) {
    x[] <- lapply(x, factorize, ...)
    x
}

context("characterize()/factorize()")
x <- structure(1:4, labels = c("A" = 1, "B" = 2, "C" = 3))
xdf <- data.frame(v1 = structure(1:4, labels = c("A" = 1, "B" = 2, "C" = 3), label = "variable 1"),
                  v2 = structure(c(1,0,0,1), labels = c("foo" = 0, "bar" = 1)),
                  v3 = 4:1, label = "variable 2")


test_that("test characterize.default()", {
    expect_true(identical(characterize(x), c(LETTERS[1:3], NA)))
})

test_that("test characterize.default()", {
    expect_true(identical(characterize(xdf), {xdf[] <- lapply(xdf, characterize); xdf}))
})

test_that("test factorize.data.frame()", {
    expect_true(identical(factorize(x), factor(x, attributes(x)$labels, names(attributes(x)$labels))))
})

test_that("test factorize.data.frame()", {
    expect_true(identical(factorize(xdf), {xdf[] <- lapply(xdf, factorize); xdf}))
})

test_that("test factorize coerce_character", {
    expect_true(identical(letters[1:3], factorize(letters[1:3])))
    expect_true(
        identical(
            factorize(letters[3:1], coerce_character = TRUE),
            factor(letters[3:1], levels = letters[1:3])
        )
    )
})

#' @title Import list of data frames
#' @description Use [import()] to import a list of data frames from a vector of file names or from a multi-object file (Excel workbook, .Rdata file, compressed directory in a zip file or tar archive, or HTML file)
#' @param file A character string containing a single file name for a multi-object file (e.g., Excel workbook, zip file, tar archive, or HTML file), or a vector of file paths for multiple files to be imported.
#' @param which If `file` is a single file path, this specifies which objects should be extracted (passed to [import()]'s `which` argument). Ignored otherwise.
#' @param rbind A logical indicating whether to pass the import list of data frames through [data.table::rbindlist()].
#' @param rbind_label If `rbind = TRUE`, a character string specifying the name of a column to add to the data frame indicating its source file.
#' @param rbind_fill If `rbind = TRUE`, a logical indicating whether to set the `fill = TRUE` (and fill missing columns with `NA`).
#' @param \dots Additional arguments passed to [import()]. Behavior may be unexpected if files are of different formats.
#' @inheritParams import
#' @inheritSection import Trust
#' @inheritSection import Which
#' @inherit import references
#' @return If `rbind=FALSE` (the default), a list of a data frames. Otherwise, that list is passed to [data.table::rbindlist()] with `fill = TRUE` and returns a data frame object of class set by the `setclass` argument; if this operation fails, the list is returned.
#' @details When file is a vector of file paths and any files are missing, those files are ignored (with warnings) and this function will not raise any error. For compressed files, the file name must also contain information about the file format of all compressed files, e.g. `files.csv.zip` for this function to work.
#' @examples
#' ## For demo, a temp. file path is created with the file extension .xlsx
#' xlsx_file <- tempfile(fileext = ".xlsx")
#' export(
#'     list(
#'         mtcars1 = mtcars[1:10, ],
#'         mtcars2 = mtcars[11:20, ],
#'         mtcars3 = mtcars[21:32, ]
#'     ),
#'     xlsx_file
#' )
#'
#' # import a single file from multi-object workbook
#' import(xlsx_file, sheet = "mtcars1")
#' # import all worksheets, the return value is a list
#' import_list(xlsx_file)
#'
#' # import and rbind all worksheets, the return value is a data frame
#' import_list(xlsx_file, rbind = TRUE)
#' @seealso [import()], [export_list()], [export()]
#' @export
import_list <- function(file, setclass = getOption("rio.import.class", "data.frame"), which, rbind = FALSE,
                        rbind_label = "_file", rbind_fill = TRUE, ...) {
    .check_file(file, single_only = FALSE)
    ## special cases
    if (length(file) == 1) {
        x <- .read_file_as_list(file = file, which = which, setclass = setclass, rbind = rbind, rbind_label = rbind_label, ...)
    } else {
        ## note the plural
        x <- .read_multiple_files_as_list(files = file, setclass = setclass, rbind = rbind, rbind_label = rbind_label, ...)
    }
    ## optionally rbind
    if (isTRUE(rbind)) {
        if (length(x) == 1) {
            x <- x[[1L]]
        } else {
            x2 <- try(data.table::rbindlist(x, fill = rbind_fill), silent = TRUE)
            if (inherits(x2, "try-error")) {
                warning("Attempt to rbindlist() the data did not succeed. List returned instead.", call. = FALSE)
                return(x)
            }
            x <- x2
        }
        x <- set_class(x, class = setclass)
    }
    return(x)
}

.strip_exts <- function(file) {
    vapply(file, function(x) tools::file_path_sans_ext(basename(x)), character(1))
}

.read_multiple_files_as_list <- function(files, setclass, rbind, rbind_label, ...) {
    names(files) <- .strip_exts(files)
    x <- lapply(files, function(thisfile) {
        out <- try(import(thisfile, setclass = setclass, ...), silent = TRUE)
        if (inherits(out, "try-error")) {
            warning(sprintf("Import failed for %s", thisfile), call. = FALSE)
            ##out <- NULL
            return(NULL)
        } else if (isTRUE(rbind)) {
            out[[rbind_label]] <- thisfile
        }
        structure(out, filename = thisfile)
    })
    names(x) <- names(files)
    return(x)
}

.read_file_as_list <- function(file, which, setclass, rbind, rbind_label, ...) {
    ## respect the `format` argument, if given
    if ("format" %in% ...names()) {
      format <- list(...)[["format"]]
    } else {
      format <- get_info(file)$format
    }
    if (R.utils::isUrl(file)) {
        file <- remote_to_local(file)
    }
    if (format == "rdata") {
        return(.import.rio_rdata(file = file, .return_everything = TRUE, ...))
    }
    archive_format <- find_compress(file)
    if (!format %in% c("html", "xlsx", "xls", "xlsm", "ods", "fods") && !archive_format$compress %in% c("zip", "tar", "tar.gz", "tar.bz2")) {
        which <- 1
        whichnames <- NULL
    }
    ## getting list of `whichnames`
    if (format == "html") {
        .check_pkg_availability("xml2")
        tables <- xml2::xml_find_all(xml2::read_html(unclass(file)), ".//table")
        if (missing(which)) {
            which <- seq_along(tables)
        }
        whichnames <- vapply(xml2::xml_attrs(tables[which]),
            function(x) if ("class" %in% names(x)) x["class"] else "",
            FUN.VALUE = character(1)
        )
        names(which) <- whichnames
    }
    if (format %in% c("xls", "xlsx", "xlsm","ods", "fods")) {
        ## .check_pkg_availability("readxl")
        sheet_func <- readxl::excel_sheets
        if (format %in% c("ods", "fods")) {
            .check_pkg_availability("readODS")
            sheet_func <- readODS::list_ods_sheets
        }
        whichnames <- sheet_func(path = file)
        if (missing(which)) {
            which <- seq_along(whichnames)
            names(which) <- whichnames
        } else if (is.character(which)) {
            whichnames <- which
        } else {
            whichnames <- whichnames[which]
        }
    }
    if (archive_format$compress %in% c("zip", "tar", "tar.gz", "tar.bz2")) {
        whichnames <- .list_archive(file, archive_format$compress)
        if (missing(which)) {
            which <- seq_along(whichnames)
            names(which) <- .strip_exts(whichnames)
        } else if (is.character(which)) {
            whichnames <- whichnames[whichnames %in% which]
        } else {
            names(which) <- .strip_exts(whichnames)
        }
    }
    ## reading all `whichnames`
    x <- lapply(which, function(thiswhich) {
        out <- try(import(file, setclass = setclass, which = thiswhich, ...), silent = TRUE)
        if (inherits(out, "try-error")) {
            warning(sprintf("Import failed for %s from %s", thiswhich, file))
            out <- NULL
        } else if (isTRUE(rbind) && length(which) > 1) {
            out[[rbind_label]] <- thiswhich
        }
        out
    })
    names(x) <- whichnames
    return(x)
}

#' @title Get File Info
#' @description A utility function to retrieve the file information of a filename, path, or URL.
#' @param file A character string containing a filename, file path, or URL.
#' @return For [get_info()], a list is return with the following slots
#' \itemize{
#'     \item `input` file extension or information used to identify the possible file format
#'     \item `format` file format, see `format` argument of [import()]
#'     \item `type` "import" (supported by default); "suggest" (supported by suggested packages, see [install_formats()]); "enhance" and "known " are not directly supported; `NA` is unsupported
#'     \item `format_name` name of the format
#'     \item `import_function` What function is used to import this file
#'     \item `export_function` What function is used to export this file
#'     \item `file` `file`
#' }
#' For [get_ext()], just `input` (usually file extension) is returned; retained for backward compatibility.
#' @examples
#' get_info("starwars.xlsx")
#' get_info("starwars.ods")
#' get_info("https://github.com/ropensci/readODS/raw/v2.1/starwars.ods")
#' get_info("~/duran_duran_rio.mp3")
#' get_ext("clipboard") ## "clipboard"
#' get_ext("https://github.com/ropensci/readODS/raw/v2.1/starwars.ods")
#' @export
get_info <- function(file) {
    .check_file(file, single_only = TRUE)
    if (tolower(file) == "clipboard") {
        return(.query_format(input = "clipboard", file = "clipboard"))
    }
    if (isFALSE(R.utils::isUrl(file))) {
        ext <- tolower(tools::file_ext(file))
    } else {
        parsed <- strsplit(strsplit(file, "?", fixed = TRUE)[[1]][1], "/", fixed = TRUE)[[1]]
        url_file <- parsed[length(parsed)]
        ext <- tolower(tools::file_ext(url_file))
    }
    if (ext == "") {
        stop("'file' has no extension", call. = FALSE)
    }
    return(.query_format(input = ext, file = file))
}

#' @export
#' @rdname get_info
get_ext <- function(file) {
    get_info(file)$input
}

.query_format <- function(input, file) {
    unique_rio_formats <- unique(rio_formats[, colnames(rio_formats) != "note"])
    if (file == "clipboard") {
        output <- as.list(unique_rio_formats[unique_rio_formats$format == "clipboard", ])
        output$file <- file
        return(output)
    }
    ## TODO google sheets
    matched_formats <- unique_rio_formats[unique_rio_formats$input == input, ]
    if (nrow(matched_formats) == 0) {
        return(list(input = input, format = input, type = "unknown", format_name = NA, import_function = NA, export_function = NA, file = file))
    }
    output <- as.list(matched_formats)
    output$file <- file
    return(output)
}

.standardize_format <- function(input) {
    .query_format(input, "")$format
}

twrap <- function(value, tag) {
    paste0("<", tag, ">", value, "</", tag, ">")
}


escape_xml <- function(x, replacement = c("&amp;", "&quot;", "&lt;", "&gt;", "&apos;")) {
    stringi::stri_replace_all_fixed(
        str = stringi::stri_enc_toutf8(x), pattern = c("&", "\"", "<", ">", "'"),
        replacement = replacement, vectorize_all = FALSE
    )
}

.check_pkg_availability <- function(pkg, lib.loc = NULL) {
    if (identical(find.package(pkg, quiet = TRUE, lib.loc = lib.loc), character(0))) {
        stop("Suggested package `", pkg, "` is not available. Please install it individually or use `install_formats()`", call. = FALSE)
    }
    return(invisible(NULL))
}

.write_as_utf8 <- function(text, file, sep = "") {
    writeLines(enc2utf8(text), con = file, sep = sep, useBytes = TRUE)
}

.check_file <- function(file, single_only = TRUE) {
    ## check the `file` argument
    if (isTRUE(missing(file))) { ## for the case of export(iris, format = "csv")
        return(invisible(NULL))
    }
    if (isFALSE(inherits(file, "character"))) {
        stop("Invalid `file` argument: must be character", call. = FALSE)
    }
    if (isFALSE(length(file) == 1) && single_only) {
        stop("Invalid `file` argument: `file` must be single", call. = FALSE)
    }
    if (any(is.na(file))) {
        stop("Invalid `file` argument: `file` must not be NA", call. = FALSE)
    }
    invisible(NULL)
}

.create_directory_if_not_exists <- function(file) {
    R.utils::mkdirs(dirname(normalizePath(file, mustWork = FALSE)))
    invisible(NULL)
}

.create_outfiles <- function(file, x) {
    names_x <- names(x)
    if (length(file) == 1L) {
        if (!grepl("%s", file, fixed = TRUE)) {
            stop("'file' must have a %s placeholder")
        }
        if (is.null(names_x)) {
            return(sprintf(file, seq_along(x)))
        }
        if (!all(nzchar(names_x))) {
            stop("All elements of 'x' must be named or all must be unnamed")
        }
        if (anyDuplicated(names_x)) {
            stop("Names of elements in 'x' are not unique")
        }
        return(sprintf(file, names_x))
    }
    if (length(x) != length(file)) {
        stop("'file' must be same length as 'x', or a single pattern with a %s placeholder")
    }
    if (anyDuplicated(file)) {
        stop("File names are not unique")
    }
    return(file)
}

.check_trust <- function(trust, format) {
    if (is.null(trust)) {
        warning("Missing `trust` will be set to FALSE by default for ", format, " in 2.0.0.", call. = FALSE)
        trust <- TRUE ## Change this for version 2.0.0
    }
    if (isFALSE(trust)) {
        stop(format, " files may execute arbitary code. Only load ", format, " files that you personally generated or can trust the origin.", call. = FALSE)
    }
    NULL
}

.reset_which <- function(file_type, which) {
    ## see #412
    if (file_type %in% c("zip", "tar", "tar.gz", "tar.bz2")) {
        return(1)
    }
    return(which)
}

test_that("File extension converted correctly", {
    expect_that(get_ext("hello.csv"), equals("csv"))
    expect_that(get_ext("hello.CSV"), equals("csv"))
    expect_that(get_ext("hello.sav.CSV"), equals("csv"))
    expect_that(get_ext("clipboard"), equals("clipboard"))
    expect_error(get_ext(1L))
})

test_that("Format converted correctly", {
    expect_that(.standardize_format(","), equals("csv"))
    expect_that(.standardize_format(";"), equals("csv2"))
    expect_that(.standardize_format("|"), equals("psv"))
    expect_that(.standardize_format("\t"), equals("tsv"))
    expect_that(.standardize_format("excel"), equals("xlsx"))
    expect_that(.standardize_format("stata"), equals("dta"))
    expect_that(.standardize_format("spss"), equals("sav"))
    expect_that(.standardize_format("sas"), equals("sas7bdat"))
})

test_that("Export without file specified", {
    withr::with_tempdir(code = {
        project_path <- getwd()
        export(iris, format = "csv")
        expect_true(file.exists(file.path(project_path, "iris.csv")))
    })
})

test_that(".check_pkg_availability", {
    expect_error(.check_pkg_availability("nonexistingpkg1233222"), "Suggested package `nonexisting")
    expect_error(.check_pkg_availability("rio"), NA)
})

library("datasets")

test_that("Function suggestions for unsupported export", {
    expect_error(export(data.frame(1), "test.jpg"),
                 "jpg format not supported. Consider using the 'jpeg::writeJPEG()' function",
                 fixed = TRUE)
})

test_that("Error for unsupported file types", {
    withr::with_tempfile("test_file", fileext = ".faketype", code = {
        writeLines("123", con = test_file)
        expect_error(import(test_file), "Format not supported")
        expect_error(export(mtcars, test_file), "Format not supported")
    })
    expect_equal(.standardize_format("faketype"), "faketype")
    expect_error(get_ext("noextension"), "'file' has no extension")
})

test_that("Error for mixed support file types", {
    expect_error(import("test.por"), "No such file")
    withr::with_tempfile("mtcars_file", fileext = ".por", code = {
        expect_error(export(mtcars, mtcars_file), "Format not supported")
    })
})

test_that("Only export data.frame or matrix", {
    withr::with_tempfile("test_file", fileext = ".csv", code = {
        expect_error(export(1, test_file), "'x' is not a data.frame or matrix")
    })
})

test_that("Column widths printed for fixed-width format", {
    withr::with_tempfile("test_file", fileext = ".txt", code = {
        expect_true(is.character(export(data.frame(1), test_file, format = "fwf", verbose = FALSE)))
        expect_message(export(data.frame(1), test_file, format = "fwf", verbose = TRUE))
    })
})

test_that("Warning for import_list() with missing file", {
    expect_warning(import_list("fake_file.csv")) ## no error
})

test_that(".create_outfiles works", {
    x <- list(
        a = data.frame(),
        b = data.frame(),
        c = data.frame()
    )
    y <- list(
        data.frame(),
        data.frame(),
        data.frame()
    )
    expect_identical(.create_outfiles("d_%s.csv", x), c("d_a.csv", "d_b.csv", "d_c.csv"))
    expect_identical(.create_outfiles("d_%s.csv", y), c("d_1.csv", "d_2.csv", "d_3.csv"))
    expect_identical(.create_outfiles(c("a.csv", "b.csv", "c.csv"), x), c("a.csv", "b.csv", "c.csv"))
})

test_that(".create_outfiles errors", {
    x <- list(
        a = data.frame(),
        a = data.frame(),
        c = data.frame()
    )
    y <- list(
        a = data.frame(),
        b = data.frame(),
        data.frame()
    )
    expect_error(.create_outfiles("d_%s.csv", x))
    expect_error(.create_outfiles(c("a.csv", "a.csv", "c.csv"), x))
    expect_error(.create_outfiles(c("a.csv", "b.csv"), x))
    expect_error(.create_outfiles(c("a.csv"), x))
    expect_error(.create_outfiles(c("d_%s.csv"), y))
})

set_class <- function(x, class = NULL) {
    if (is.null(class)) {
        return(x)
    }

    if ("data.table" %in% class) {
        return(.ensure_data_table(x))
    }

    if (any(c("tibble", "tbl_df", "tbl") %in% class)) {
        return(.ensure_tibble(x))
    }

    if (any(c("arrow", "arrow_table") %in% class)) {
        ## because setclass can be used without import, must check again
        .check_pkg_availability("arrow")
        return(.ensure_arrow(x))
    }
    return(.ensure_data_frame(x))
}

.ensure_arrow <- function(x) {
    if (inherits(x, "ArrowTabular")) {
        return(x)
    }
    return(arrow::arrow_table(x))
}

.ensure_data_table <- function(x) {
    if (inherits(x, "data.table")) {
        return(x)
    }
    return(data.table::as.data.table(x))
}

.ensure_tibble <- function(x) {
    if (inherits(x, "tbl")) {
        return(x)
    }
    return(tibble::as_tibble(x))
}

.ensure_data_frame <- function(x) {
    out <- structure(x, class = "data.frame")
    if (nrow(out) == 0) {
        return(out)
    }
    if (!length(rownames(out))) {
        rownames(out) <- as.character(seq_len(length(out[, 1L, drop = TRUE])))
    }
    return(out)
}

find_compress <- function(f) {
    if (endsWith(f, ".zip")) {
        return(list(file = sub("\\.zip$", "", f), compress = "zip"))
    }
    if (endsWith(f, ".tar.gz")) {
        return(list(file = sub("\\.tar\\.gz$", "", f), compress = "tar.gz"))
    }
    if (endsWith(f, ".tgz")) {
        return(list(file = sub("\\.tgz$", "", f), compress = "tar.gz"))
    }
    if (endsWith(f, ".tar.bz2")) {
        return(list(file = sub("\\.tar\\.bz2$", "", f), compress = "tar.bz2"))
    }
    if (endsWith(f, ".tbz2")) {
        return(list(file = sub("\\.tbz2$", "", f), compress = "tar.bz2"))
    }
    if (endsWith(f, ".tar")) {
        return(list(file = sub("\\.tar$", "", f), compress = "tar"))
    }
    if (endsWith(f, ".gzip")) {
        ## weird
        return(list(file = sub("\\.gzip$", "", f), compress = "gzip"))
    }
    if (endsWith(f, ".gz")) {
        return(list(file = sub("\\.gz$", "", f), compress = "gzip"))
    }
    if (endsWith(f, ".bz2")) {
        return(list(file = sub("\\.bz2$", "", f), compress = "bzip2"))
    }
    if (endsWith(f, ".bzip2")) {
        ## weird
        return(list(file = sub("\\.bzip2$", "", f), compress = "bzip2"))
    }
    return(list(file = f, compress = NA_character_))
}

compress_out <- function(cfile, filename, type = c("zip", "tar", "tar.gz", "tar.bz2", "gzip", "bzip2")) {
    type <- ext <- match.arg(type)
    cfile2 <- basename(cfile)
    filename <- normalizePath(filename)
    if (type %in% c("gzip", "bzip2")) {
        return(.compress_rutils(filename, cfile, ext = ext))
    }
    tmp <- tempfile()
    dir.create(tmp)
    on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
    file.copy(from = filename, to = file.path(tmp, basename(filename)), overwrite = TRUE)
    wd <- getwd()
    on.exit(setwd(wd), add = TRUE) ## for security, see #438 and #319
    setwd(tmp)
    if (type == "zip") {
        o <- utils::zip(cfile2, files = basename(filename))
    }
    if (type == "tar") {
        o <- utils::tar(cfile2, files = basename(filename), compression = "none")
    }
    if (type == "tar.gz") {
        o <- utils::tar(cfile2, files = basename(filename), compression = "gzip")
    }
    if (type == "tar.bz2") {
        o <- utils::tar(cfile2, files = basename(filename), compression = "bzip2")
    }
    setwd(wd) ## see #438
    if (o != 0) {
        stop(sprintf("File compression failed for %s!", cfile))
    }
    file.copy(from = file.path(tmp, cfile2), to = cfile, overwrite = TRUE)
    return(cfile)
}

parse_archive <- function(file, which, file_type, ...) {
    if (file_type %in% c("gzip", "bzip2")) {
        ## it doesn't have the same interface as unzip
        return(.parse_rutils(filename = file, file_type = file_type))
    }
    if (file_type == "zip") {
        extract_func <- utils::unzip
    }
    if (file_type %in% c("tar", "tar.gz", "tar.bz2")) {
        extract_func <- utils::untar
    }
    file_list <- .list_archive(file, file_type)
    d <- tempfile()
    dir.create(d)

    if (is.numeric(which)) {
        extract_func(file, files = file_list[which], exdir = d)
        return(file.path(d, file_list[which]))
    }
    if (substring(which, 1, 1) != "^") {
        which2 <- paste0("^", which)
    }
    extract_func(file, files = file_list[grep(which2, file_list)[1]], exdir = d)
    return(file.path(d, which))
}

.list_archive <- function(file, file_type = c("zip", "tar", "tar.gz", "tar.bz2")) {
    ## just a simple wrapper to unify the interface of utils::unzip and utils::untar
    file_type <- match.arg(file_type)
    if (file_type == "zip") {
        file_list <- utils::unzip(file, list = TRUE)$Name
    }
    if (file_type %in% c("tar", "tar.gz", "tar.bz2")) {
        file_list <- utils::untar(file, list = TRUE)
    }
    return(file_list)
}

.compress_rutils <- function(filename, cfile, ext, remove = TRUE, FUN = gzfile) {
    ## Caution: Please note that remove = TRUE by default, it will delete `filename`!
    if (ext == "bzip2") {
        FUN <- bzfile
    }
    tmp_cfile <- R.utils::compressFile(filename = filename, destname = tempfile(), ext = ext, FUN = FUN, overwrite = TRUE, remove = remove)
    file.copy(from = tmp_cfile, to = cfile, overwrite = TRUE)
    unlink(tmp_cfile)
    return(cfile)
}

.parse_rutils <- function(filename, file_type) {
    if (file_type == "gzip") {
        decompression_fun <- gzfile
    }
    if (file_type == "bzip2") {
        decompression_fun <- bzfile
    }
    destname <- tempfile()
    R.utils::decompressFile(filename = filename, destname = destname, temporary = TRUE, remove = FALSE, overwrite = TRUE, FUN = decompression_fun, ext = file_type)
}

.check_tar_support <- function(file_type, rversion) {
    if (file_type %in% c("tar", "tar.gz", "tar.bz2") && rversion < "4.0.3") {
        stop("Exporting to tar formats is not supported for this version of R.", call. = FALSE)
    }
    NULL
}

.get_compressed_format <- function(cfile, file, file_type, format) {
    if (file_type %in% c("gzip", "bzip2")) {
        return(ifelse(isFALSE(missing(format)), tolower(format), get_info(find_compress(cfile)$file)$input))
    }
    ## zip or tar formats, use the decompressed file path
    return(ifelse(isFALSE(missing(format)), tolower(format), get_info(file)$input))
}

context("Compressed files")

test_that("Recognize compressed file types", {
    expect_true(rio:::find_compress("file.zip")$compress == "zip")
    expect_true(rio:::find_compress("file.tar")$compress == "tar")
    expect_true(rio:::find_compress("file.tar.gz")$compress == "tar.gz")
    expect_true(rio:::find_compress("file.gzip")$compress == "gzip")
    expect_true(rio:::find_compress("file.gz")$compress == "gzip")
    expect_true(is.na(rio:::find_compress("file.notcompressed")$compress))
})

test_that("Export to compressed (zip, tar) / import", {
    skip_if(getRversion() <= "4.0")
    formats <- c("zip", "tar", "tar.gz", "tgz", "tar.bz2", "tbz2")
    for (format in formats) {
        withr::with_tempfile("iris_path", fileext = paste0(".csv.", format), code = {
            e1 <- export(iris, iris_path)
            expect_true(file.exists(iris_path))
            expect_true(is.data.frame(import(iris_path)))
            expect_true(is.data.frame(import(iris_path, which = 1)))
            base_file_name <- gsub(paste0("\\.", format), "", basename(iris_path))
            expect_true(is.data.frame(import(iris_path, which = base_file_name)))
            if (format %in% c("tar.gz", "tgz")) {
                expect_true(R.utils::isGzipped(iris_path, method = "content"))
            }
            if (format %in% c("tar.bzip2", "tbz2")) {
                expect_true(R.utils::isBzipped(iris_path, method = "content"))
            }
        })
    }
})

test_that("Multi-item zip", {
    formats <- c("zip")
    for (format in formats) {
        withr::with_tempfile("iris_path", fileext = paste0(".csv.", format), code = {
            list_of_df <- list(mtcars1 = mtcars[1:10, ], mtcars2 = mtcars[11:20, ], mtcars3 = mtcars[21:32, ])
            export_list(list_of_df, file = "%s.csv", archive = iris_path)
            y <- import(iris_path)
            expect_true(is.data.frame(y))
            expect_silent(z1 <- import(iris_path, which = 1))
            expect_silent(z2 <- import(iris_path, which = 2))
            expect_true(identical(y, z1))
            expect_false(identical(y, z2))
        })
    }
})

test_that("Export to compressed (gz, bz2) / import", {
    formats <- c("gz", "gzip", "bz2", "bzip2")
    for (format in formats) {
        withr::with_tempfile("iris_path", fileext = paste0(".csv.", format), code = {
            e1 <- export(iris, iris_path)
            expect_true(file.exists(iris_path))
            expect_true(is.data.frame(import(iris_path)))
            if (format %in% c("gzip", "gz")) {
                expect_true(R.utils::isGzipped(iris_path, method = "content"))
            }
            if (format %in% c("bzip2", "bz2")) {
                expect_true(R.utils::isBzipped(iris_path, method = "content"))
            }
        })
    }
})

test_that("Prevent the reuse of `which` for zip and tar", {
    skip_if(getRversion() <= "4.0")
    formats <- c("zip", "tar", "tar.gz", "tgz", "tar.bz2", "tbz2")
    for (format in formats) {
        withr::with_tempfile("data_path", fileext = paste0(".xlsx.", format), code = {
            rio::export(list(some_iris = head(iris)), data_path)
            expect_error(import(data_path), NA)
            raw_file <- .list_archive(data_path, find_compress(data_path)$compress)[1]
            expect_error(import(data_path, which = raw_file), NA)
            expect_error(suppressWarnings(import(data_path, which = "some_iris")))
        })
    }
    ## but not for non-archive format, e.g. .xlsx.gz
    formats <- c("gz", "bz2")
    for (format in formats) {
        withr::with_tempfile("data_path", fileext = paste0(".xlsx.", format), code = {
            rio::export(list(some_iris = head(iris)), data_path)
            expect_error(import(data_path), NA)
            expect_error(import(data_path, which = "some_iris"), NA)
        })
    }
})

test_that(".check_tar_support ref #421", {
    expect_error(.check_tar_support("tar", R_system_version("4.0.2")), "^Exporting")
    expect_error(.check_tar_support("tar.gz", R_system_version("4.0.2")), "^Exporting")
    expect_error(.check_tar_support("tar.bz2", R_system_version("4.0.2")), "^Exporting")
    expect_error(.check_tar_support("tar", R_system_version("4.0.3")), NA)
    expect_error(.check_tar_support("tar.gz", R_system_version("4.0.3")), NA)
    expect_error(.check_tar_support("tar.bz2", R_system_version("4.0.3")), NA)
    expect_error(.check_tar_support("zip", R_system_version("4.0.2")), NA)
    expect_error(.check_tar_support(NA, R_system_version("4.0.2")), NA)
})

test_that("tar export error for R < 4.0.3", {
    skip_if(getRversion() >= "4.0.3")
    withr::with_tempfile("iris_path", fileext = paste0(".csv.", "tar"), code = {
        expect_error(export(iris, iris_path), "^Exporting")
    })
})

test_that("Wild zip and tar ref $425", {
    skip_if(getRversion() <= "4.0")
    ##skip_on_os("windows")
    withr::with_tempfile("test_files", fileext = c(".csv", ".zip"), code = {
        filename <- test_files[1]
        zip_file <- test_files[2]
        write.csv(1, filename)
        zip(zip_file, filename)
        expect_error(rio::import(zip_file), NA)
    })
    for (tar_format in c("tar", "tar.gz", "tar.bz2")) {
        withr::with_tempfile("test_files", fileext = c(".csv", paste0(".", tar_format)), code = {
            filename <- test_files[1]
            tar_file <- test_files[2]
            write.csv(1, filename)
            compress_out(tar_file, filename, type = tar_format)
            expect_error(rio::import(tar_file), NA)
        })
    }
})

test_that("Relative path #438", {
    ## see test_export_list.R
    withr::with_tempdir({
        mylist <- list(
            mtcars3 = mtcars[1:10, ],
            mtcars2 = mtcars[11:20, ],
            mtcars1 = mtcars[21:32, ]
        )
        rio::export_list(mylist, file = paste0("file_", 1:3, ".csv"), archive = "arch/archive.zip")
        expect_true(file.exists("arch/archive.zip"))
    })
})

test_that("Integration test #435", {
    ## filter out all NA cases so that #437 works
    withr::with_tempfile("iris_path", fileext = "csv.zip", code = {
        expect_error(rio::export(iris, NA), "must be character") ## strange to be honest
        expect_error(rio::export(iris, NA_character_), "must not be NA")
    })
})

#' @title Convert from one file format to another
#' @description This function constructs a data frame from a data file using [import()] and uses [export()] to write the data to disk in the format indicated by the file extension.
#' @param in_file A character string naming an input file.
#' @param out_file A character string naming an output file.
#' @param in_opts A named list of options to be passed to [import()].
#' @param out_opts A named list of options to be passed to [export()].
#' @return A character string containing the name of the output file (invisibly).
#' @examples
#' ## For demo, a temp. file path is created with the file extension .dta (Stata)
#' dta_file <- tempfile(fileext = ".dta")
#' ## .csv
#' csv_file <- tempfile(fileext = ".csv")
#' ## .xlsx
#' xlsx_file <- tempfile(fileext = ".xlsx")
#'
#'
#' ## Create a Stata data file
#' export(mtcars, dta_file)
#'
#' ## convert Stata to CSV and open converted file
#' convert(dta_file, csv_file)
#' import(csv_file)
#'
#' ## correct an erroneous file format
#' export(mtcars, xlsx_file, format = "tsv") ## DON'T DO THIS
#' ## import(xlsx_file) ## ERROR
#' ## convert the file by specifying `in_opts`
#' convert(xlsx_file, xlsx_file, in_opts = list(format = "tsv"))
#' import(xlsx_file)
#'
#' ## convert from the command line:
#' ## Rscript -e "rio::convert('mtcars.dta', 'mtcars.csv')"
#' @seealso [Luca Braglia](https://lbraglia.github.io/) has created a Shiny app called [rioweb](https://github.com/lbraglia/rioweb) that provides access to the file conversion features of rio through a web browser.
#' @export
convert <- function(in_file, out_file, in_opts=list(), out_opts=list()) {
    if (missing(out_file)) {
        stop("'outfile' is missing with no default")
    }
    invisible(do.call("export", c(list(file = out_file, x = do.call("import", c(list(file=in_file), in_opts))), out_opts)))
}

standardize_attributes <- function(dat) {
    out <- dat
    a <- attributes(out)
    if ("variable.labels" %in% names(a)) {
        names(a)[names(a) == "variable.labels"] <- "var.labels"
        a$var.labels <- unname(a$var.labels)
    }
    # cleanup import
    attr(out, "var.labels") <- NULL      # Stata
    attr(out, "variable.labels") <- NULL # SPSS
    attr(out, "formats") <- NULL
    attr(out, "types") <- NULL
    attr(out, "label.table") <- NULL
    for (i in seq_along(out)) {
        if ("value.labels" %in% names(attributes(out[[i]]))) {
            attr(out[[i]], "labels") <- attr(out[[i]], "value.labels", exact = TRUE)
            attr(out[[i]], "value.labels") <- NULL
        }
        if (any(grepl("haven_labelled", class(out[[i]]), fixed = TRUE))) {
            out[[i]] <- unclass(out[[i]])
        }
        if ("var.labels" %in% names(a)) {
            attr(out[[i]], "label") <- a$var.labels[i]
        }
        if (any(grepl("$format", names(a), fixed = TRUE))) {
            attr(out[[i]], "format") <- a[[grep("$format", names(a))[1L]]][i]
        }
        if ("types" %in% names(a)) {
            attr(out[[i]], "type") <- a$types[i]
        }
        if ("val.labels" %in% names(a) && (a$val.labels[i] != "")) {
            attr(out[[i]], "labels") <- a$label.table[[a$val.labels[i]]]
        }
    }
    out
}

restore_labelled <- function(x) {
    # restore labelled variable classes
    x[] <- lapply(x, function(v) {
        if (is.factor(v)) {
            haven::labelled(
                x = as.numeric(v),
                labels = stats::setNames(seq_along(levels(v)), levels(v)),
                label = attr(v, "label", exact = TRUE)
            )
        } else if (!is.null(attr(v, "labels", exact = TRUE)) || !is.null(attr(v, "label", exact = TRUE))) {
            haven::labelled(
                x = v,
                labels = attr(v, "labels", exact = TRUE),
                label = attr(v, "label", exact = TRUE)
            )
        } else {
            v
        }
    })
    x
}

.docall <- function(.fcn, ..., args = NULL, alwaysArgs = NULL, .functions = list(.fcn),
                    .ignoreUnusedArgs = TRUE) {
    ## the same as R.utils::doCall, only with default option
    R.utils::doCall(.fcn = .fcn, ..., args = args, alwaysArgs = alwaysArgs, .functions = .functions,
                    .ignoreUnusedArgs = getOption("rio.ignoreunusedargs", default = TRUE))
}

.remap_tidy_convention <- function(func, file, which, header, ...) {
    dots <- list(...)
    if ("path" %in% names(dots)) {
        dots[["path"]] <- NULL
    }
    if ("sheet" %in% names(dots)) {
        which <- dots[["sheet"]]
        dots[["sheet"]] <- NULL
    }
    if ("col_names" %in% names(dots)) {
        header <- dots[["col_names"]]
        dots[["col_names"]] <- NULL
    }
    .docall(func, args = c(dots, list(path = file, sheet = which, col_names = header)))
}

import_delim <- function(file, which = 1, sep = "auto", header = "auto", stringsAsFactors = FALSE, data.table = FALSE, ...) {
    .docall(data.table::fread, ..., args = list(input = file, sep = sep, header = header,
                  stringsAsFactors = stringsAsFactors,
                  data.table = data.table))
}

#' @export
.import.rio_dat <- function(file, which = 1, ...) {
    message(sprintf("Ambiguous file format ('.dat'), but attempting 'data.table::fread(\"%s\")'\n", file))
    import_delim(file = file, ...)
}

#' @export
.import.rio_tsv <- function(file, sep = "auto", which = 1, fread = lifecycle::deprecated(), dec = if (sep %in% c("\t", "auto")) "." else ",", ...) {
    if (lifecycle::is_present(fread)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "import(fread)", details = "tsv will always be read by `data.table:fread()`. The parameter `fread` will be dropped in v2.0.0.")
    }
    import_delim(file = file, sep = sep, dec = dec, ...)
}

#' @export
.import.rio_txt <- function(file, sep = "auto", which = 1, fread = lifecycle::deprecated(), dec = if (sep %in% c(",", "auto")) "." else ",", ...) {
    if (lifecycle::is_present(fread)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "import(fread)", details = "txt will always be read by `data.table:fread()`. The parameter `fread` will be dropped in v2.0.0.")
    }
    import_delim(file = file, sep = sep, dec = dec, ...)
}

#' @export
.import.rio_csv <- function(file, sep = ",", which = 1, fread = lifecycle::deprecated(), dec = if (sep %in% c(",", "auto")) "." else ",", ...) {
    if (lifecycle::is_present(fread)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "import(fread)", details = "csv will always be read by `data.table:fread()`. The parameter `fread` will be dropped in v2.0.0.")
    }
    import_delim(file = file, sep = if (sep == ",") "auto" else sep, dec = dec, ...)
}

#' @export
.import.rio_csv2 <- function(file, sep = ";", which = 1, fread = lifecycle::deprecated(), dec = if (sep %in% c(";", "auto")) "," else ".", ...) {
    if (lifecycle::is_present(fread)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "import(fread)", details = "csv2 will always be read by `data.table:fread()`. The parameter `fread` will be dropped in v2.0.0.")
    }
    import_delim(file = file, sep = if (sep == ";") "auto" else sep, dec = dec, ...)
}

#' @export
.import.rio_csvy <- function(file, sep = ",", which = 1, fread = lifecycle::deprecated(), dec = if (sep %in% c(",", "auto")) "." else ",", yaml = TRUE, ...) {
    if (lifecycle::is_present(fread)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "import(fread)", details = "csvy will always be read by `data.table:fread()`. The parameter `fread` will be dropped in v2.0.0.")
    }
    import_delim(file = file, sep = if (sep == ",") "auto" else sep, dec = dec, yaml = yaml, ...)
}

#' @export
.import.rio_psv <- function(file, sep = "|", which = 1, fread = lifecycle::deprecated(), dec = if (sep %in% c("|", "auto")) "." else ",", ...) {
    if (lifecycle::is_present(fread)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "import(fread)", details = "psv will always be read by `data.table:fread()`. The parameter `fread` will be dropped in v2.0.0.")
    }
    import_delim(file = file, sep = if (sep == "|") "auto" else sep, dec = dec, ...)
}

.get_col_position <- function(file, widths, col_names, col_position) {
    if (missing(col_names)) {
        col_names <- NULL
    }
    if (!is.null(col_position)) {
        return(col_position)
    }
    if (is.list(widths) && isFALSE(c("begin", "end") %in% names(widths))) {
        return(readr::fwf_widths(widths, col_names = col_names))
    }
    if (isTRUE(is.numeric(widths))) {
        return(readr::fwf_widths(abs(widths), col_names = col_names))
    }
    return(readr::fwf_empty(file = file, col_names = col_names))
}

.fix_col_types <- function(col_types, widths) {
    if (isFALSE(is.numeric(widths))) {
        return(col_types)
    }
    col_types <- rep("?", length(widths))
    col_types[widths < 0] <- "?"
    col_types <- paste0(col_types, collapse = "")
    return(col_types)
}

#' @export
.import.rio_fwf <- function(file, which = 1, widths = NULL, col_names, col_types = NULL, col_positions, col.names, col_position = NULL, comment = "#", ...) {
    if (!is.null(widths)) {
        lifecycle::deprecate_warn(when = "1.0.2", what = "import(widths)", details = "`widths` is kept for backward compatibility. Please use `col_positions` or unset `widths` to allow automatic guessing, see `?readr::read_fwf`. The parameter `widths` will be dropped in v2.0.0.")
    }
    if (!missing(col.names)) {
        lifecycle::deprecate_warn(when = "1.0.2", what = "import(widths)", details = "`col.names` is kept for backward compatibility. Please use `col_names`. The parameter `col.names` will be dropped in v2.0.0.")
        col_names <- col.names
    }
    col_positions <- .get_col_position(file = file, widths = widths, col_names = col_names, col_position = col_position)
    if (!is.null(widths) && !is.null(col_types)) {
        col_types <- .fix_col_types(col_types, widths)
    }
    .docall(readr::read_fwf, ..., args = list(file = file, col_positions = col_positions, col_types = col_types, comment = comment))
}

#' @export
.import.rio_r <- function(file, which = 1, trust = getOption("rio.import.trust", default = NULL), ...) {
    .check_trust(trust, format = ".R")
    .docall(dget, ..., args = list(file = file))
}

#' @export
.import.rio_dump <- function(file, which = 1, envir = new.env(), trust = getOption("rio.import.trust", default = NULL), ...) {
    .check_trust(trust, format = "dump")
    source(file = file, local = envir)
    if (missing(which)) {
        if (length(ls(envir)) > 1) {
            warning("Dump file contains multiple objects. Returning first object.")
        }
        which <- 1
    }
    if (is.numeric(which)) {
        get(ls(envir)[which], envir)
    } else {
        get(ls(envir)[grep(which, ls(envir))[1]], envir)
    }
}

#' @export
.import.rio_rds <- function(file, which = 1, trust = getOption("rio.import.trust", default = NULL), ...) {
    .check_trust(trust, format = "RDS")
    readRDS(file = file)
}

#' @export
.import.rio_rdata <- function(file, which = 1, envir = new.env(), trust = getOption("rio.import.trust", default = NULL), .return_everything = FALSE, ...) {
    .check_trust(trust, format = "RData")
    load(file = file, envir = envir)
    if (isTRUE(.return_everything)) {
        ## for import_list()
        return(as.list(envir))
    }
    if (missing(which)) {
        if (length(ls(envir)) > 1) {
          warning("Rdata file contains multiple objects. Returning first object.")
        }
      which <- 1
    }
    if (is.numeric(which)) {
      get(ls(envir)[which], envir)
    } else {
      get(ls(envir)[grep(which, ls(envir))[1]], envir)
    }
}

#' @export
.import.rio_rda <- .import.rio_rdata

#' @export
.import.rio_feather <- function(file, which = 1, ...) {
    .check_pkg_availability("arrow")
    .docall(arrow::read_feather, ..., args = list(file = file))
}

#' @export
.import.rio_fst <- function(file, which = 1, ...) {
    .check_pkg_availability("fst")
    .docall(fst::read.fst, ..., args = list(path = file))
}

#' @export
.import.rio_matlab <- function(file, which = 1, ...) {
    .check_pkg_availability("rmatio")
    rmatio::read.mat(filename = file)
}

#' @export
.import.rio_dta <- function(file, haven = lifecycle::deprecated(),
                            convert.factors = lifecycle::deprecated(), which = 1, ...) {
    if (lifecycle::is_present(haven) || lifecycle::is_present(convert.factors)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "import(haven)", details = "dta will always be read by `haven`. The parameter `haven` will be dropped in v2.0.0.")
    }
    standardize_attributes(.docall(haven::read_dta, ..., args = list(file = file)))
}

#' @export
.import.rio_dbf <- function(file, which = 1, as.is = TRUE, ...) {
    .docall(foreign::read.dbf, ..., args = list(file = file, as.is = as.is))
}

#' @export
.import.rio_dif <- function(file, which = 1, ...) {
    .docall(utils::read.DIF, ..., args = list(file = file))
}

#' @export
.import.rio_sav <- function(file, which = 1, haven = lifecycle::deprecated(), to.data.frame = lifecycle::deprecated(), use.value.labels = lifecycle::deprecated(), ...) {
    if (lifecycle::is_present(haven) || lifecycle::is_present(to.data.frame) || lifecycle::is_present(use.value.labels)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "import(haven)", details = "sav will always be read by `haven`. The parameter `haven` will be dropped in v2.0.0.")
    }
    standardize_attributes(.docall(haven::read_sav, ..., args = list(file = file)))
}

#' @export
.import.rio_zsav <- function(file, which = 1, ...) {
    standardize_attributes(.docall(haven::read_sav, ..., args = list(file = file)))
}

#' @export
.import.rio_spss <- function(file, which = 1, ...) {
    standardize_attributes(.docall(haven::read_por, ..., args = list(file = file)))
}

#' @export
.import.rio_sas7bdat <- function(file, which = 1, column.labels = FALSE, ...) {
    standardize_attributes(.docall(haven::read_sas, ..., args = list(data_file = file)))
}

#' @export
.import.rio_xpt <- function(file, which = 1, haven = lifecycle::deprecated(), ...) {
    if (lifecycle::is_present(haven)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "import(haven)", details = "xpt will always be read by `haven`. The parameter `haven` will be dropped in v2.0.0.")
    }
    standardize_attributes(.docall(haven::read_xpt, ..., args = list(file = file)))
}

#' @export
.import.rio_mtp <- function(file, which = 1, ...) {
    .docall(foreign::read.mtp, ..., args = list(file = file))
}

#' @export
.import.rio_syd <- function(file, which = 1, ...) {
    .docall(foreign::read.systat, ..., args = list(file = file, to.data.frame = TRUE))
}

#' @export
.import.rio_json <- function(file, which = 1, ...) {
    .check_pkg_availability("jsonlite")
    .docall(jsonlite::fromJSON, ..., args = list(txt = file))
}

#' @export
.import.rio_rec <- function(file, which = 1, ...) {
    .docall(foreign::read.epiinfo, ..., args = list(file = file))
}

#' @export
.import.rio_arff <- function(file, which = 1, ...) {
    .docall(foreign::read.arff, ..., args = list(file = file))
}

#' @export
.import.rio_xls <- function(file, which = 1, header = TRUE, ...) {
    .remap_tidy_convention(readxl::read_xls, file = file, which = which, header = header, ...)
}

#' @export
.import.rio_xlsx <- function(file, which = 1, header = TRUE, readxl = lifecycle::deprecated(), ...) {
    if (lifecycle::is_present(readxl)) {
        lifecycle::deprecate_warn(
            when = "0.5.31",
            what = "import(readxl)",
            details = "xlsx will always be read by `readxl`. The parameter `readxl` will be dropped in v2.0.0."
        )
    }
    .remap_tidy_convention(readxl::read_xlsx, file = file, which = which, header = header, ...)
}

#' @export
.import.rio_xlsm <- .import.rio_xlsx

#' @export
.import.rio_fortran <- function(file, which = 1, style, ...) {
    if (missing(style)) {
        stop("Import of Fortran format data requires a 'style' argument. See ? utils::read.fortran().")
    }
    .docall(utils::read.fortran, ..., args = list(file = file, format = style))
}

#' @export
.import.rio_ods <- function(file, which = 1, header = TRUE, ...) {
    .check_pkg_availability("readODS")
    .remap_tidy_convention(readODS::read_ods, file = file, which = which, header = header, ...)
}

#' @export
.import.rio_fods <- function(file, which = 1, header = TRUE, ...) {
    .check_pkg_availability("readODS")
    .remap_tidy_convention(readODS::read_fods, file = file, which = which, header = header, ...)
}

#' @export
.import.rio_xml <- function(file, which = 1, stringsAsFactors = FALSE, ...) {
    .check_pkg_availability("xml2")
    x <- xml2::as_list(xml2::read_xml(unclass(file)))[[1L]]
    d <- do.call("rbind", c(lapply(x, unlist)))
    row.names(d) <- seq_len(nrow(d))
    d <- as.data.frame(d, stringsAsFactors = stringsAsFactors)
    tc2 <- function(x) {
        out <- utils::type.convert(x, as.is = FALSE)
        if (is.factor(out)) {
            x
        } else {
            out
        }
    }
    if (isTRUE(stringsAsFactors)) {
        d[] <- lapply(d, utils::type.convert)
    } else {
        d[] <- lapply(d, tc2)
    }
    d
}

# This is a helper function for .import.rio_html
extract_html_row <- function(x, empty_value) {
    ## Both <th> and <td> are valid for table data, and <th> may be used when
    ## there is an accented element (e.g. the first row of the table)
    to_extract <- x[names(x) %in% c("th", "td")]
    ## Insert a value into cells that eventually will become empty cells (or they
    ## will be dropped and the table will not be generated).  Note that this more
    ## complex code for finding the length is required because of html like
    ## <td><br/></td>
    unlist_length <- lengths(lapply(to_extract, unlist))
    to_extract[unlist_length == 0] <- list(empty_value)
    unlist(to_extract)
}

#' @export
.import.rio_html <- function(file, which = 1, stringsAsFactors = FALSE, ..., empty_value = "") {
    # find all tables
    tables <- xml2::xml_find_all(xml2::read_html(unclass(file)), ".//table")
    if (which > length(tables)) {
        stop("Requested table exceeds number of tables found in file (", length(tables), ")!")
    }
    x <- xml2::as_list(tables[[which]])
    if ("tbody" %in% names(x)) {
        # Note that "tbody" may be specified multiple times in a valid html table
        x <- unlist(x[names(x) == "tbody"], recursive = FALSE)
    }
    # loop row-wise over the table and then rbind()
    ## check for table header to use as column names
    col_names <- NULL
    if ("th" %in% names(x[[1]])) {
        col_names <- extract_html_row(x[[1]], empty_value = empty_value)
        # Drop the first row since column names have already been extracted from it.
        x <- x[-1]
    }
    out <- do.call("rbind", lapply(x, extract_html_row, empty_value = empty_value))
    colnames(out) <-
        if (is.null(col_names)) {
            paste0("V", seq_len(ncol(out)))
        } else {
            col_names
        }
    out <- as.data.frame(out, ..., stringsAsFactors = stringsAsFactors)
    # set row names
    rownames(out) <- seq_len(nrow(out))
    # type.convert() to numeric, etc.
    out[] <- lapply(out, utils::type.convert, as.is = TRUE)
    out
}

#' @export
.import.rio_yml <- function(file, which = 1, stringsAsFactors = FALSE, ...) {
    .check_pkg_availability("yaml")
    as.data.frame(.docall(yaml::read_yaml, ..., args = list(file = file)), stringsAsFactors = stringsAsFactors)
}

#' @export
.import.rio_eviews <- function(file, which = 1, ...) {
    .check_pkg_availability("hexView")
    .docall(hexView::readEViews, ..., args = list(filename = file))
}

#' @export
.import.rio_clipboard <- function(file = "clipboard", which = 1, header = TRUE, sep = "\t", ...) {
    .check_pkg_availability("clipr")
    .docall(clipr::read_clip_tbl, ..., args = list(x = clipr::read_clip(), header = header, sep = sep))
}

#' @export
.import.rio_pzfx <- function(file, which = 1, ...) {
    .check_pkg_availability("pzfx")
    dots <- list(...)
    if ("path" %in% names(dots)) {
        dots[["path"]] <- NULL
    }
    if ("table" %in% names(dots)) {
        which <- dots[["table"]]
        dots[["table"]] <- NULL
    }
    .docall(pzfx::read_pzfx, args = c(dots, list(path = file, table = which)))
}

#' @export
.import.rio_parquet <- function(file, which = 1, ...) {
    .check_pkg_availability("nanoparquet")
    .docall(nanoparquet::read_parquet, ..., args = list(file = file, options = nanoparquet::parquet_options(class = "data.frame")))
}

#' @export
.import.rio_qs2 <- function(file, which = 1, ...) {
  .check_pkg_availability("qs2")
  .docall(qs2::qs_read, ..., args = list(file = file))
}

export_delim <- function(file, x, fwrite = lifecycle::deprecated(), sep = "\t", row.names = FALSE,
                         col.names = TRUE, append = FALSE, ...) {
    if (lifecycle::is_present(fwrite)) {
        lifecycle::deprecate_warn(when = "0.5.31", what = "export(fwrite)", details = "plain text files will always be written with `data.table::fwrite`. The parameter `fwrite` will be dropped in v2.0.0.")
    }
    .docall(data.table::fwrite, ...,
            args = list(x = x, file = file, sep = sep, row.names = row.names,
                        col.names = ifelse(append, FALSE, col.names), append = append))
}

#' @export
.export.rio_txt <- function(file, x, ...) {
    export_delim(x = x, file = file, ...)
}

#' @export
.export.rio_tsv <- function(file, x, ...) {
    export_delim(x = x, file = file, ...)
}

#' @export
.export.rio_csv <- function(file, x, sep = ",", dec = ".", ...) {
    export_delim(x = x, file = file, sep = sep, dec = dec, ...)
}

#' @export
.export.rio_csv2 <- function(file, x, sep = ";", dec = ",", ...) {
    export_delim(x = x, file = file, sep = sep, dec = dec, ...)
}

#' @export
.export.rio_csvy <- function(file, x, sep = ",", dec = ".", yaml = TRUE, ...) {
    export_delim(x = x, file = file, sep = sep, dec = dec, yaml = TRUE, ...)
}

#' @export
.export.rio_psv <- function(file, x, ...) {
    export_delim(x = x, file = file, sep = "|", ...)
}

#' @export
.export.rio_fwf <- function(file, x, verbose = getOption("verbose", FALSE), sep = "", row.names = FALSE, quote = FALSE, col.names = FALSE, digits = getOption("digits", 7), ...) {
    dat <- lapply(x, function(col) {
        if (is.character(col)) {
            col <- as.numeric(as.factor(col))
        } else if (is.factor(col)) {
            col <- as.integer(col)
        }
        if (is.integer(col)) {
            return(sprintf("%i", col))
        }
        if (is.numeric(col)) {
            decimals <- strsplit(as.character(col), ".", fixed = TRUE)
            m1 <- max(nchar(unlist(lapply(decimals, `[`, 1))), na.rm = TRUE)
            decimals_2 <- unlist(lapply(decimals, `[`, 2))
            decimals_2_nchar <- nchar(decimals_2[!is.na(decimals_2)])
            if (length(decimals_2_nchar)) {
                m2 <- max(decimals_2_nchar, na.rm = TRUE)
            } else {
                m2 <- 0
            }
            if (!is.finite(m2)) {
                m2 <- digits
            }
            return(formatC(sprintf(fmt = paste0("%0.", m2, "f"), col), width = (m1 + m2 + 1)))
        } else if (is.logical(col)) {
            return(sprintf("%i", col))
        }
    })
    dat <- do.call(cbind, dat)
    n <- nchar(dat[1, ]) + c(rep(nchar(sep), ncol(dat) - 1), 0)
    col_classes <- vapply(x, class, character(1))
    col_classes[col_classes == "factor"] <- "integer"
    dict <- cbind.data.frame(
        variable = names(n),
        class = col_classes,
        width = unname(n),
        columns = paste0(c(1, cumsum(n) + 1)[-length(n)], "-", cumsum(n)),
        stringsAsFactors = FALSE
    )
    if (isTRUE(verbose)) {
        message("Columns:")
        message(paste0(utils::capture.output(dict), collapse = "\n"))
        if (sep == "") {
            message(
                "\nRead in with:\n",
                'import("', file, '",\n',
                "       widths = c(", paste0(n, collapse = ","), "),\n",
                '       col.names = c("', paste0(names(n), collapse = '","'), '"),\n',
                '       colClasses = c("', paste0(col_classes, collapse = '","'), '"))\n',
                domain = NA
            )
        }
    }
    .write_as_utf8(paste0("#", utils::capture.output(utils::write.csv(dict, row.names = FALSE, quote = FALSE))), file = file, sep = "\n")
    .docall(utils::write.table, ...,
            args = list(x = dat, file = file, append = TRUE, row.names = row.names, sep = sep, quote = quote,
                        col.names = col.names))
}

#' @export
.export.rio_r <- function(file, x, ...) {
    .docall(dput, ..., args = list(x = x, file = file))
}

#' @export
.export.rio_dump <- function(file, x, ...) {
    dump(as.character(substitute(x)), file = file)
}

#' @export
.export.rio_rds <- function(file, x, ...) {
    .docall(saveRDS, ..., args = list(object = x, file = file))
}

#' @export
.export.rio_rdata <- function(file, x, ...) {
    if (isFALSE(is.data.frame(x)) && isFALSE(is.list(x)) && isFALSE(is.environment(x)) && isFALSE(is.character(x))) {
        stop("'x' must be a data.frame, list, or environment")
    }
    if (is.data.frame(x)) {
        return(save(x, file = file, ...))
    }
    if (is.list(x)) {
        e <- as.environment(x)
        return(save(list = names(x), file = file, envir = e, ...))
    }
    if (is.environment(x)) {
        return(save(list = ls(x), file = file, envir = x, ...))
    }
    return(save(list = x, file = file, ...)) ## characters, but is this doing what it does?
}

#' @export
.export.rio_rda <- .export.rio_rdata

#' @export
.export.rio_feather <- function(file, x, ...) {
    .check_pkg_availability("arrow")
    .docall(arrow::write_feather, ..., args = list(x = x, sink = file))
}

#' @export
.export.rio_fst <- function(file, x, ...) {
    .check_pkg_availability("fst")
    .docall(fst::write.fst, ..., args = list(x = x, path = file))
}

#' @export
.export.rio_matlab <- function(file, x, ...) {
    .check_pkg_availability("rmatio")
    .docall(rmatio::write.mat, ..., args = list(object = x, filename = file))
}

#' @export
.export.rio_sav <- function(file, x, ...) {
    x <- restore_labelled(x)
    .docall(haven::write_sav, ..., args = list(data = x, path = file))
}

#' @export
.export.rio_zsav <- function(file, x, compress = TRUE, ...) {
    x <- restore_labelled(x)
    .docall(haven::write_sav, ..., args = list(data = x, path = file, compress = compress))
}

#' @export
.export.rio_dta <- function(file, x, ...) {
    x <- restore_labelled(x)
    .docall(haven::write_dta, ..., args = list(data = x, path = file))
}

#' @export
.export.rio_sas7bdat <- function(file, x, ...) {
    x <- restore_labelled(x)
    .docall(haven::write_sas, ..., args = list(data = x, path = file))
}

#' @export
.export.rio_xpt <- function(file, x, ...) {
    x <- restore_labelled(x)
    .docall(haven::write_xpt, ..., args = list(data = x, path = file))
}

#' @export
.export.rio_dbf <- function(file, x, ...) {
    .docall(foreign::write.dbf, ..., args = list(dataframe = x, file = file))
}

#' @export
.export.rio_json <- function(file, x, ...) {
    .check_pkg_availability("jsonlite")
    .write_as_utf8(.docall(jsonlite::toJSON, ..., args = list(x = x)), file = file)
}

#' @export
.export.rio_arff <- function(file, x, ...) {
    .docall(foreign::write.arff, ..., args = list(x = x, file = file))
}

#' @export
.export.rio_xlsx <- function(file, x, ...) {
    .docall(writexl::write_xlsx, ..., args = list(x = x, path = file))
}

#' @export
.export.rio_ods <- function(file, x, ...) {
    .check_pkg_availability("readODS")
    .docall(readODS::write_ods, ..., args = list(x = x, path = file))
}

#' @export
.export.rio_fods <- function(file, x, ...) {
    .check_pkg_availability("readODS")
    .docall(readODS::write_fods, ..., args = list(x = x, path = file))
}

#' @export
.export.rio_html <- function(file, x, ...) {
    .check_pkg_availability("xml2")
    html <- xml2::read_html("<!doctype html><html><head>\n<title>R Exported Data</title>\n</head><body>\n</body>\n</html>")
    bod <- xml2::xml_children(html)[[2]]
    if (is.data.frame(x)) {
        x <- list(x)
    }
    for (i in seq_along(x)) {
        x[[i]][] <- lapply(x[[i]], as.character)

        x[[i]][] <- lapply(x[[i]], escape_xml)
        names(x[[i]]) <- escape_xml(names(x[[i]]))
        tab <- xml2::xml_add_child(bod, "table")
        # add header row
        invisible(xml2::xml_add_child(tab, xml2::read_xml(paste0(twrap(paste0(twrap(names(x[[i]]), "th"), collapse = ""), "tr"), "\n"))))
        # add data
        for (j in seq_len(nrow(x[[i]]))) {
            xml2::xml_add_child(tab, xml2::read_xml(paste0(twrap(paste0(twrap(unlist(x[[i]][j, , drop = TRUE]), "td"), collapse = ""), "tr"), "\n")))
        }
    }
    .docall(xml2::write_xml, ..., args = list(x = html, file = file))
}

#' @export
.export.rio_xml <- function(file, x, ...) {
    .check_pkg_availability("xml2")
    xml <- xml2::read_xml(paste0("<", as.character(substitute(x)), ">\n</", as.character(substitute(x)), ">\n"))
    att <- attributes(x)[!names(attributes(x)) %in% c("names", "row.names", "class")]
    for (a in seq_along(att)) {
        xml2::xml_attr(xml, names(att)[a]) <- att[[a]]
    }
    # remove illegal characters
    row.names(x) <- escape_xml(row.names(x))
    colnames(x) <- escape_xml(colnames(x), ".")
    x[] <- lapply(x, escape_xml)
    # add data
    for (i in seq_len(nrow(x))) {
        thisrow <- xml2::xml_add_child(xml, "Observation")
        xml2::xml_attr(thisrow, "row.name") <- row.names(x)[i]
        for (j in seq_along(x)) {
            xml2::xml_add_child(thisrow, xml2::read_xml(paste0(twrap(x[i, j, drop = TRUE], names(x)[j]), "\n")))
        }
    }

    .docall(xml2::write_xml, ..., args = list(x = xml, file = file))
}

#' @export
.export.rio_yml <- function(file, x, ...) {
    .check_pkg_availability("yaml")
    .docall(yaml::write_yaml, ..., args = list(x = x, file = file))
}

#' @export
.export.rio_clipboard <- function(file, x, row.names = FALSE, col.names = TRUE, sep = "\t", ...) {
    .check_pkg_availability("clipr")
    .docall(clipr::write_clip, ..., args = list(content = x, row.names = row.names, col.names = col.names, sep = sep))
}

#' @export
.export.rio_pzfx <- function(file, x, ..., row_names = FALSE) {
    .check_pkg_availability("pzfx")
    .docall(pzfx::write_pzfx, ..., args = list(x = x, path = file, row_names = row_names))
}

#' @export
.export.rio_parquet <- function(file, x, ...) {
    .check_pkg_availability("nanoparquet")
    .docall(nanoparquet::write_parquet, ..., args = list(x = x, file = file))
}

#' @export
.export.rio_qs2 <- function(file, x, ...) {
    .check_pkg_availability("qs2")
    .docall(qs2::qs_save, ..., args = list(object = x, file = file))
}

.import <- function(file, ...) {
    UseMethod(".import")
}

.import.default <- function(file, ...) {
    fileinfo <- get_info(file) ## S3 can't be dispatched
    if (fileinfo$type == "unknown" || is.na(fileinfo$import_function) || fileinfo$import_function == "") {
        stop("Format not supported", call. = FALSE)
    }
    if (fileinfo$type == "known") {
        stop(sprintf(gettext("%s format not supported. Consider using the '%s()' function"),
                     fileinfo$format, fileinfo$import_function), call. = FALSE)
    }
    if (fileinfo$type == "enhance") {
        pkg <- strsplit(fileinfo$import_function, "::", fixed = TRUE)[[1]][1]
        stop(sprintf(gettext("Import support for the %s format is exported by the %s package. Run 'library(%s)' then try again."),
                     fileinfo$format, pkg, pkg), call. = FALSE)
    }
}

.export <- function(file, x, ...) {
    UseMethod(".export")
}

.export.default <- function(file, x, ...) {
    fileinfo <- get_info(file)
    if (fileinfo$type == "unknown" || is.na(fileinfo$export_function) || fileinfo$export_function == "") {
        stop("Format not supported", call. = FALSE)
    }
    if (fileinfo$type == "known") {
        stop(sprintf(gettext("%s format not supported. Consider using the '%s()' function"),
                     fileinfo$format, fileinfo$export_function), call. = FALSE)
    }
}

#!/bin/bash
#
# Manual repro for OSC 9;4 progress-bar handling (issue #1509).
# Run inside a Rio window. The fix is verified visually if the
# indeterminate bar in phase 2 *moves* across the window — before
# the fix it froze at the left edge because every heartbeat OSC
# yanked the animation phase back to t=0.

osc() { printf '\033]9;4;%s;%s\033\\' "$1" "$2"; }

echo "== phase 1: determinate, climbing 0->100 =="
for p in 0 10 20 30 40 50 60 70 80 90 100; do
    osc 1 "$p"
    sleep 0.15
done

echo "== phase 2: heartbeat indeterminate (issue #1509 repro) =="
echo "(OSC 9;4;3 every 100ms for 6s — bar should slide L<->R, not freeze)"
end=$(( $(date +%s) + 6 ))
while [ "$(date +%s)" -lt "$end" ]; do
    osc 3 0
    sleep 0.1
done

echo "== phase 3: error state at 50% =="
osc 2 50
sleep 1.5

echo "== phase 4: pause at 75% =="
osc 4 75
sleep 1.5

echo "== phase 5: clear =="
osc 0 0

# Retired from https://iterm2.com/utilities/imgcat
# iTerm2 is licensed under GPL-2.0-or-later.

#!/usr/bin/env bash

set -o pipefail

# tmux requires unrecognized OSC sequences to be wrapped with DCS tmux;
# <sequence> ST, and for all ESCs in <sequence> to be replaced with ESC ESC. It
# only accepts ESC backslash for ST. We use TERM instead of TMUX because TERM
# gets passed through ssh.
function print_osc() {
    if [[ $TERM == screen* || $TERM == tmux* ]]; then
        printf "\033Ptmux;\033\033]"
    else
        printf "\033]"
    fi
}

# More of the tmux workaround described above.
function print_st() {
    if [[ $TERM == screen* || $TERM == tmux* ]]; then
        printf "\a\033\\"
    else
        printf "\a"
    fi
}

function load_version() {
    if [ -z ${IMGCAT_BASE64_VERSION+x} ]; then
        IMGCAT_BASE64_VERSION=$(base64 --version 2>&1)
        export IMGCAT_BASE64_VERSION
    fi
}

function b64_encode() {
    load_version
    if [[ $IMGCAT_BASE64_VERSION =~ GNU ]]; then
        # Disable line wrap
        base64 -w0
    else
        base64
    fi
}

function b64_decode() {
    load_version
    if [[ $IMGCAT_BASE64_VERSION =~ fourmilab ]]; then
        BASE64ARG=-d
    elif [[ $IMGCAT_BASE64_VERSION =~ GNU ]]; then
        BASE64ARG=-di
    else
        BASE64ARG=-D
    fi
    base64 $BASE64ARG
}

# print_image filename inline base64contents print_filename width height preserve_aspect_ratio
#   filename: Filename to convey to client
#   inline: 0 or 1, if set to 1, the file will be displayed inline, otherwise, it will be downloaded
#   base64contents: Base64-encoded contents
#   print_filename: 0 or 1, if set to 1, print the filename after outputting the image
#   width: set output width of the image in character cells, pixels or percent
#   height: set output height of the image in character cells, pixels or percent
#   preserve_aspect_ratio: 0 or 1, if set to 1, fill the specified width and height as much as possible without stretching the image
#   file: Empty string or file type like "application/json" or ".js".
function print_image() {
    print_osc
    printf "1337;File=inline=%s" "$2"
    printf ";size=%d" $(printf "%s" "$3" | b64_decode | wc -c)
    [ -n "$1" ] && printf ";name=%s" "$(printf "%s" "$1" | b64_encode)"
    [ -n "$5" ] && printf ";width=%s" "$5"
    [ -n "$6" ] && printf ";height=%s" "$6"
    [ -n "$7" ] && printf ";preserveAspectRatio=%s" "$7"
    [ -n "$8" ] && printf ";type=%s" "$8"
    printf ":%s" "$3"
    print_st
    printf '\n'
    [ "$4" == "1" ] && echo "$1"
    has_image_displayed=t
}

function error() {
    errcho "ERROR: $*"
}

function errcho() {
    echo "$@" >&2
}

function show_help() {
    errcho
    errcho "Usage: imgcat [-p] [-n] [-W width] [-H height] [-r] [-s] [-u] [-t file-type] [-f] filename ..."
    errcho "       cat filename | imgcat [-W width] [-H height] [-r] [-s]"
    errcho
    errcho "Display images inline in the iTerm2 using Inline Images Protocol"
    errcho
    errcho "Options:"
    errcho
    errcho "    -h, --help                      Display help message"
    errcho "    -p, --print                     Enable printing of filename or URL after each image"
    errcho "    -n, --no-print                  Disable printing of filename or URL after each image"
    errcho "    -u, --url                       Interpret following filename arguments as remote URLs"
    errcho "    -f, --file                      Interpret following filename arguments as regular Files"
    errcho "    -t, --type file-type            Provides a type hint"
    errcho "    -r, --preserve-aspect-ratio     When scaling image preserve its original aspect ratio"
    errcho "    -s, --stretch                   Stretch image to specified width and height (this option is opposite to -r)"
    errcho "    -W, --width N                   Set image width to N character cells, pixels or percent (see below)"
    errcho "    -H, --height N                  Set image height to N character cells, pixels or percent (see below)"
    errcho
    errcho "    If you don't specify width or height an appropriate value will be chosen automatically."
    errcho "    The width and height are given as word 'auto' or number N followed by a unit:"
    errcho "        N      character cells"
    errcho "        Npx    pixels"
    errcho "        N%     percent of the session's width or height"
    errcho "        auto   the image's inherent size will be used to determine an appropriate dimension"
    errcho
    errcho "    If a type is provided, it is used as a hint to disambiguate."
    errcho "    The file type can be a mime type like text/markdown, a language name like Java, or a file extension like .c"
    errcho "    The file type can usually be inferred from the extension or its contents. -t is most useful when"
    errcho "    a filename is not available, such as whe input comes from a pipe."
    errcho
    errcho "Examples:"
    errcho
    errcho "    $ imgcat -W 250px -H 250px -s avatar.png"
    errcho "    $ cat graph.png | imgcat -W 100%"
    errcho "    $ imgcat -p -W 500px -u http://host.tld/path/to/image.jpg -W 80 -f image.png"
    errcho "    $ cat url_list.txt | xargs imgcat -p -W 40 -u"
    errcho "    $ imgcat -t application/json config.json"
    errcho
}

function check_dependency() {
    if ! (builtin command -V "$1" >/dev/null 2>&1); then
        error "missing dependency: can't find $1"
        exit 1
    fi
}

# verify that value is in the image sizing unit format: N / Npx / N% / auto
function validate_size_unit() {
    if [[ ! "$1" =~ ^(:?[0-9]+(:?px|%)?|auto)$ ]]; then
        error "Invalid image sizing unit - '$1'"
        show_help
        exit 1
    fi
}

## Main

if [ -t 0 ]; then
    has_stdin=f
else
    has_stdin=t
fi

# Show help if no arguments and no stdin.
if [ $has_stdin = f ] && [ $# -eq 0 ]; then
    show_help
    exit
fi

check_dependency base64
check_dependency wc
file_type=""

# Look for command line flags.
while [ $# -gt 0 ]; do
    case "$1" in
    -h | --h | --help)
        show_help
        exit
        ;;
    -p | --p | --print)
        print_filename=1
        ;;
    -n | --n | --no-print)
        print_filename=0
        ;;
    -W | --W | --width)
        validate_size_unit "$2"
        width="$2"
        shift
        ;;
    -H | --H | --height)
        validate_size_unit "$2"
        height="$2"
        shift
        ;;
    -r | --r | --preserve-aspect-ratio)
        preserve_aspect_ratio=1
        ;;
    -s | --s | --stretch)
        preserve_aspect_ratio=0
        ;;
    -f | --f | --file)
        has_stdin=f
        is_url=f
        ;;
    -u | --u | --url)
        check_dependency curl
        has_stdin=f
        is_url=t
        ;;
    -t | --t | --type)
         file_type="$2"
         shift
         ;;
    -*)
        error "Unknown option flag: $1"
        show_help
        exit 1
        ;;
    *)
        if [ "$is_url" == "t" ]; then
            encoded_image=$(curl -fs "$1" | b64_encode) || {
                error "Could not retrieve image from URL $1, error_code: $?"
                exit 2
            }
        elif [ -r "$1" ]; then
            encoded_image=$(cat "$1" | b64_encode)
        else
            error "imgcat: $1: No such file or directory"
            exit 2
        fi
        has_stdin=f
        print_image "$1" 1 "$encoded_image" "$print_filename" "$width" "$height" "$preserve_aspect_ratio" "$file_type"
        ;;
    esac
    shift
done

# Read and print stdin
if [ $has_stdin = t ]; then
    print_image "" 1 "$(cat | b64_encode)" 0 "$width" "$height" "$preserve_aspect_ratio" "$file_type"
fi

if [ "$has_image_displayed" != "t" ]; then
    error "No image provided. Check command line options."
    show_help
    exit 1
fi

exit 0

# check if stdout is a terminal...
if test -t 1; then

    # see if it supports colors...
    ncolors=$(tput colors)

    if test -n "$ncolors" && test $ncolors -ge 8; then
        bold="$(tput bold)"
        underline="$(tput smul)"
        standout="$(tput smso)"
        foreground="$(tput sgr0)"
        black="$(tput setaf 0)"
        red="$(tput setaf 1)"
        green="$(tput setaf 2)"
        yellow="$(tput setaf 3)"
        blue="$(tput setaf 4)"
        magenta="$(tput setaf 5)"
        cyan="$(tput setaf 6)"
        white="$(tput setaf 7)"
    fi
fi

echo "${foreground}foreground${foreground}"
echo "${black}black${foreground}"
echo "${red}red${foreground}"
echo "${green}green${foreground}"
echo "${yellow}yellow${foreground}"
echo "${blue}blue${foreground}"
echo "${white}white${foreground}"
echo "${magenta}magenta${foreground}"
echo "${cyan}cyan${foreground}"

#!/bin/bash
#
# Show the codepoint ranges Rio renders with built-in "drawable" sprites
# (box-drawing, blocks, braille, and — as they land — sextants, octants,
# powerline, geometric shapes, branch glyphs). Run inside a Rio window to
# eyeball the sprite rasterizer.
#
# With `fonts.use-drawable-chars = true` (the default) these render
# procedurally and should be crisp + seamless in any font; ranges not yet
# implemented fall back to the active font's glyphs. Toggle the setting to
# compare sprite vs font rendering.

set -u

# Emit one UTF-8 character for a decimal codepoint. Pure shell arithmetic
# + octal printf escapes, so it works on the stock macOS bash (3.2) with
# no python/perl dependency.
emit() {
  local cp=$1
  if [ "$cp" -lt 128 ]; then
    printf "$(printf '\\%03o' "$cp")"
  elif [ "$cp" -lt 2048 ]; then
    printf "$(printf '\\%03o\\%03o' \
      $(((cp >> 6) | 0xC0)) $(((cp & 0x3F) | 0x80)))"
  elif [ "$cp" -lt 65536 ]; then
    printf "$(printf '\\%03o\\%03o\\%03o' \
      $(((cp >> 12) | 0xE0)) $(((cp >> 6 & 0x3F) | 0x80)) $(((cp & 0x3F) | 0x80)))"
  else
    printf "$(printf '\\%03o\\%03o\\%03o\\%03o' \
      $(((cp >> 18) | 0xF0)) $(((cp >> 12 & 0x3F) | 0x80)) \
      $(((cp >> 6 & 0x3F) | 0x80)) $(((cp & 0x3F) | 0x80)))"
  fi
}

# grid <hex-start> <hex-end> [cols]
# Print an inclusive codepoint range as a grid, `cols` glyphs per row.
grid() {
  local cp=$((16#$1)) end=$((16#$2)) cols=${3:-16} i=0
  while [ "$cp" -le "$end" ]; do
    emit "$cp"
    printf ' '
    i=$((i + 1))
    [ $((i % cols)) -eq 0 ] && printf '\n'
    cp=$((cp + 1))
  done
  [ $((i % cols)) -ne 0 ] && printf '\n'
  printf '\n'
}

hr() { printf '\n\033[1;36m%s\033[0m\n' "$1"; }

hr "Box Drawing  U+2500-257F"
grid 2500 257F 16

hr "Block Elements  U+2580-259F"
grid 2580 259F 16

hr "Braille  U+2800-28FF"
grid 2800 28FF 32

hr "Sextants (Legacy Computing)  U+1FB00-1FB3B"
grid 1FB00 1FB3B 16

hr "Octants (Legacy Computing Supplement)  U+1CD00-1CDE5"
grid 1CD00 1CDE5 32

hr "Powerline  U+E0B0-E0D4"
grid E0B0 E0D4 16

hr "Geometric Shapes  U+25A0-25FF"
grid 25A0 25FF 16

hr "Branch Drawing  U+F5D0-F60D"
grid F5D0 F60D 16

# Real-world scenes — literal glyphs, so adjacent cells must join cleanly.
hr "Scenes (joins, shades, gauges)"
printf '  \xe2\x95\xad'; printf '\xe2\x94\x80%.0s' $(seq 1 11); printf '\xe2\x95\xae'
printf '   \xe2\x94\x8c'; printf '\xe2\x94\x80%.0s' $(seq 1 11); printf '\xe2\x94\x90'
printf '   \xe2\x95\x94'; printf '\xe2\x95\x90%.0s' $(seq 1 11); printf '\xe2\x95\x97\n'
echo "  │  rounded  │   │  light    │   ║  double   ║"
printf '  \xe2\x95\xb0'; printf '\xe2\x94\x80%.0s' $(seq 1 11); printf '\xe2\x95\xaf'
printf '   \xe2\x94\x94'; printf '\xe2\x94\x80%.0s' $(seq 1 11); printf '\xe2\x94\x98'
printf '   \xe2\x95\x9a'; printf '\xe2\x95\x90%.0s' $(seq 1 11); printf '\xe2\x95\x9d\n'
echo
echo "  shade ramp:  ░▒▓█        eighths:  ▁▂▃▄▅▆▇█"
echo "  spinner:     ⠋ ⠙ ⠹ ⠸ ⠼ ⠴ ⠦ ⠧ ⠇ ⠏"
echo "  powerline:   ▌ main ▶ feature ▶ "
echo

# Script retired from https://github.com/pop-os/cosmic-term/blob/be808b56cf24d03fc99cf44b0885078a81a16523/ansi-colors.sh#L64
# which licensed under GNU 3.0 https://github.com/pop-os/cosmic-term/blob/master/LICENSE

#!/usr/bin/env bash

set -e

CNAMES=("BLK" "RED" "GRN" "YEL" "BLU" "MAG" "CYN" "WHT")

printf "\033[1m" # bold
printf "\nUnderline With FG Colors:\n"

printf "\033[4m" # underline
for foreground in $(seq 0 7)
do
    printf "\033[$((foreground+30))m ${CNAMES[$foreground]} "
done
printf "\x1B[24m\n" # no underline

printf "\nUnderline Styles And Colors:\n"

printf "\nFG:  "
printf "\033[9mStrikeout\033[0m "
printf "\033[4mUnderline\033[0m "
printf "\033[4:2mDoubleUnderline\033[0m "
printf "\033[4:3mCurlyUnderline\033[0m "
printf "\033[4:4mDottedUnderline\033[0m "
printf "\033[4:5mDashedUnderline\033[0m "
printf "\n"

printf "INV: "
printf "\033[7m\033[9mStrikeout\033[0m "
printf "\033[7m\033[4mUngderline\033[0m "
printf "\033[7m\033[4:2mDoubleUnderline\033[0m "
printf "\033[7m\033[4:3mCurlyUnderline\033[0m "
printf "\033[7m\033[4:4mDottedUnderline\033[0m "
printf "\033[7m\033[4:5mDashedUnderline\033[0m "
printf "\n"

for line_color in $(seq 0 7)
do
    printf "${CNAMES[$line_color]}: "
    printf "          "
    printf "\033[58:5:"${line_color}m
    printf "\033[4mUnderline\033[0m "
    printf "\033[58:5:"${line_color}m
    printf "\033[4:2mDoubleUnderline\033[0m "
    printf "\033[58:5:"${line_color}m
    printf "\033[4:3mCurlyUnderline\033[0m "
    printf "\033[58:5:"${line_color}m
    printf "\033[4:4mDottedUnderline\033[0m "
    printf "\033[58:5:"${line_color}m
    printf "\033[4:5mDashedUnderline\033[0m "
    printf "\n"
done

#!/bin/bash

echo -e "regular"
echo -e "\e[1mbold\e[0m"
echo -e "\e[3mitalic\e[0m"
echo -e "\e[4munderline\e[0m"
echo -e "\e[9mstrikethrough\e[0m"

# https://sw.kovidgoyal.net/kitty/underlines/
echo -e ""
echo -e "\e[4:0mno underline\e[0m"
echo -e "\e[4:1mstraight underline\e[0m"
echo -e "\e[4:2mdouble underline\e[0m"
echo -e "\e[4:3mcurly underline\e[0m"
echo -e "\e[4:4mdotted underline\e[0m"
echo -e "\e[4:5mdashed underline\e[0m"

---
title: "Working with labelled data"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Working with labelled data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r}
library(rio)
library(haven)
```

Formats SAS, SPSS, and Stata  use `haven` as import and export functions. And these formats can have the so-called labelled data. For more information, please read `vignette("semantics", "haven")`. Here, we provide a quick guide on how to work with labelled data using `rio`.

You can use `haven::labelled()` to create labelled data.

```{r}
gender <- haven::labelled(
                     c("M", "F", "F", "F", "M"),
                     c(Male = "M", Female = "F"))
```

Or directly using `attrs`

```{r}
rating <- sample(1:5)
attr(rating, "labels") <-  c(c(Good = 1, Bad = 5))
```

```{r}
mydata <- data.frame(gender, rating)
```

Round trip: The data labels are retained. But they are at the variable level.

```{r}
export(mydata, "mydata.sav")
restored_data <- rio::import("mydata.sav")
str(restored_data)
```

`rio::gather_attrs()` converts attributes to the data.frame level

```{r}
g <- rio::gather_attrs(restored_data)
str(g)
attr(g, "labels")
```

```{r include = FALSE}
unlink("mydata.sav")
```

---
title: "Extending rio"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Extending rio}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(rio)
```

rio implements format-specific S3 methods for each type of file that can be imported from or exported to. This happens via internal S3 generics, `.import` and `.export`. It is possible to write new methods like with any S3 generic (e.g., `print`).

As an example, `.import.rio_csv` imports from a comma-separated values file. If you want to produce a method for a new filetype with extension `myfile`, you simply have to create a function called `.import.rio_myfile` that implements a format-specific importing routine and returns a data.frame. rio will automatically recognize new S3 methods, so that you can then import your file using: `import("file.myfile")`.

The way to develop `export` method is same: `.export.rio_csv`. The first two parameters of `.export` are `file` (file name) and `x` (data frame to be exported).

As general guidance, if an import method creates many attributes, these attributes should be stored --- to the extent possible --- in variable-level attributes fields. These can be gathered to the data.frame level by the user via `gather_attrs`.

# Examples

## arff

The following example shows how the arff import and export methods are implemented internally.

```r
.import.rio_arff <- function(file, which = 1, ...) {
    foreign::read.arff(file = file)
}
.export.rio_arff <- function(file, x, ...) {
    foreign::write.arff(x = x, file = file, ...)
}
```

## ledger

This is the example from the `ledger` package (MIT) by Dr Trevor L David .

```r
.import.rio_ledger <- register # nolint
register <- function(file, ..., toolchain = default_toolchain(file), date = NULL) {
    .assert_toolchain(toolchain)
    switch(toolchain,
        "ledger" = register_ledger(file, ..., date = date),
        "hledger" = register_hledger(file, ..., date = date),
        "beancount" = register_beancount(file, ..., date = date),
        "bean-report_ledger" = {
            file <- .bean_report(file, "ledger")
            on.exit(unlink(file))
            register_ledger(file, ..., date = date)
        },
        "bean-report_hledger" = {
            file <- .bean_report(file, "hledger")
            on.exit(unlink(file))
            register_hledger(file, ..., date = date)
        }
    )
}
```

---
title: "Import, Export, and Convert Data Files"
date: "`r Sys.Date()`"
output:
  html_document:
    fig_caption: false
    toc: true
    toc_float:
      collapsed: false
      smooth_scroll: false
    toc_depth: 3
vignette: >
  %\VignetteIndexEntry{Introduction to 'rio'}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Import, Export, and Convert Data Files

The idea behind **rio** is to simplify the process of importing data into R and exporting data from R. This process is, probably unnecessarily, extremely complex for beginning R users. Indeed, R supplies [an entire manual](https://cran.r-project.org/doc/manuals/r-release/R-data.html) describing the process of data import/export. And, despite all of that text, most of the packages described are (to varying degrees) out-of-date. Faster, simpler, packages with fewer dependencies have been created for many of the file types described in that document. **rio** aims to unify data I/O (importing and exporting) into two simple functions: `import()` and `export()` so that beginners (and experienced R users) never have to think twice (or even once) about the best way to read and write R data.

The core advantage of **rio** is that it makes assumptions that the user is probably willing to make. Specifically, **rio** uses the file extension of a file name to determine what kind of file it is. This is the same logic used by Windows OS, for example, in determining what application is associated with a given file type. By taking away the need to manually match a file type (which a beginner may not recognize) to a particular import or export function, **rio** allows almost all common data formats to be read with the same function.

By making import and export easy, it's an obvious next step to also use R as a simple data conversion utility. Transferring data files between various proprietary formats is always a pain and often expensive. The `convert` function therefore combines `import` and `export` to easily convert between file formats (thus providing a FOSS replacement for programs like Stat/Transfer or Sledgehammer.

## Supported file formats

**rio** supports a variety of different file formats for import and export. To keep the package slim, all non-essential formats are supported via "Suggests" packages, which are not installed (or loaded) by default. To ensure rio is fully functional, install these packages the first time you use **rio** via:

```R
install_formats()
```

The full list of supported formats is below:

```{r, include = FALSE}
suppressPackageStartupMessages(library(data.table))
```

```{r featuretable, echo = FALSE}
rf <- data.table(rio:::rio_formats)[!input %in% c(",", ";", "|", "\\t") & type %in% c("import", "suggest", "archive"),]
short_rf <- rf[, paste(input, collapse = " / "), by = format_name]
type_rf <- unique(rf[,c("format_name", "type", "import_function", "export_function", "note")])

feature_table <- short_rf[type_rf, on = .(format_name)]

colnames(feature_table)[2] <- "signature"

setorder(feature_table, "type", "format_name")
feature_table$import_function <- stringi::stri_extract_first(feature_table$import_function, regex = "[a-zA-Z0-9\\.]+")
feature_table$import_function[is.na(feature_table$import_function)] <- ""
feature_table$export_function <- stringi::stri_extract_first(feature_table$export_function, regex = "[a-zA-Z0-9\\.]+")
feature_table$export_function[is.na(feature_table$export_function)] <- ""

feature_table$type <- ifelse(feature_table$type == "suggest", "Suggest", "Default")
feature_table <- feature_table[,c("format_name", "signature", "import_function", "export_function", "type", "note")]

colnames(feature_table) <- c("Name", "Extensions / \"format\"", "Import Package", "Export Package", "Type", "Note")
knitr::kable(feature_table)
```

Additionally, any format that is not supported by **rio** but that has a known R implementation will produce an informative error message pointing to a package and import or export function. Unrecognized formats will yield a simple "Unrecognized file format" error.

## Data Import

**rio** allows you to import files in almost any format using one, typically single-argument, function. `import()` infers the file format from the file's extension and calls the appropriate data import function for you, returning a simple data.frame. This works for any for the formats listed above.

```{r, echo=FALSE, results='hide'}
library("rio")

export(mtcars, "mtcars.csv")
export(mtcars, "mtcars.dta")
export(mtcars, "mtcars_noext", format = "csv")
```

```{r}
library("rio")

x <- import("mtcars.csv")
y <- import("mtcars.dta")

# confirm identical
all.equal(x, y, check.attributes = FALSE)
```

If for some reason a file does not have an extension, or has a file extension that does not match its actual type, you can manually specify a file format to override the format inference step. For example, we can read in a CSV file that does not have a file extension by specifying `csv`:

```{r}
head(import("mtcars_noext", format = "csv"))
```


```{r, echo=FALSE, results='hide'}
unlink("mtcars.csv")
unlink("mtcars.dta")
unlink("mtcars_noext")
```

### Importing Data Lists

Sometimes you may have multiple data files that you want to import. `import()` only ever returns a single data frame, but `import_list()` can be used to import a vector of file names into R. This works even if the files are different formats:

```r
str(import_list(dir()), 1)
```

Similarly, some single-file formats (e.g. Excel Workbooks, Zip directories, HTML files, etc.) can contain multiple data sets. Because `import()` is type safe, always returning a data frame, importing from these formats requires specifying a `which` argument to `import()` to dictate which data set (worksheet, file, table, etc.) to import (the default being `which = 1`). But `import_list()` can be used to import all (or only a specified subset, again via `which`) of data objects from these types of files.

## Data Export

The export capabilities of **rio** are somewhat more limited than the import capabilities, given the availability of different functions in various R packages and because import functions are often written to make use of data from other applications and it never seems to be a development priority to have functions to export to the formats used by other applications. That said, **rio** currently supports the following formats:


```{r}
library("rio")

export(mtcars, "mtcars.csv")
export(mtcars, "mtcars.dta")
```

It is also easy to use `export()` as part of an R pipeline (from magrittr or dplyr). For example, the following code uses `export()` to save the results of a simple data transformation:

```{r}
library("magrittr")
mtcars %>%
  subset(hp > 100) %>%
  aggregate(. ~ cyl + am, data = ., FUN = mean) %>%
  export(file = "mtcars2.dta")
```

Some file formats (e.g., Excel workbooks, Rdata files) can support multiple data objects in a single file. `export()` natively supports output of multiple objects to these types of files:

```{r}
# export to sheets of an Excel workbook
export(list(mtcars = mtcars, iris = iris), "multi.xlsx")
```

It is also possible to use the new (as of v0.6.0) function `export_list()` to write a list of data frames to multiple files using either a vector of file names or a file pattern:

```{r}
export_list(list(mtcars = mtcars, iris = iris), "%s.tsv")
```

## File Conversion

The `convert()` function links `import()` and `export()` by constructing a dataframe from the imported file and immediately writing it back to disk. `convert()` invisibly returns the file name of the exported file, so that it can be used to programmatically access the new file.

Because `convert()` is just a thin wrapper for `import()` and `export()`, it is very easy to use. For example, we can convert

```{r}
# create file to convert
export(mtcars, "mtcars.dta")

# convert Stata to SPSS
convert("mtcars.dta", "mtcars.sav")
```

`convert()` also accepts lists of arguments for controlling import (`in_opts`) and export (`out_opts`). This can be useful for passing additional arguments to import or export methods. This could be useful, for example, for reading in a fixed-width format file and converting it to a comma-separated values file:

```{r}
# create an ambiguous file
fwf <- tempfile(fileext = ".fwf")
cat(file = fwf, "123456", "987654", sep = "\n")

# see two ways to read in the file
identical(import(fwf, widths = c(1, 2, 3)), import(fwf, widths = c(1, -2, 3)))

# convert to CSV
convert(fwf, "fwf.csv", in_opts = list(widths = c(1, 2, 3)))
import("fwf.csv") # check conversion
```

```{r, echo=FALSE, results='hide'}
unlink("mtcars.dta")
unlink("mtcars.sav")
unlink("fwf.csv")
unlink(fwf)
```

With metadata-rich file formats (e.g., Stata, SPSS, SAS), it can also be useful to pass imported data through `characterize()` or `factorize()` when converting to an open, text-delimited format: `characterize()` converts a single variable or all variables in a data frame that have "labels" attributes into character vectors based on the mapping of values to value labels (e.g., `export(characterize(import("file.dta")), "file.csv")`). An alternative approach is exporting to CSVY format, which records metadata in a YAML-formatted header at the beginning of a CSV file.

It is also possible to use **rio** on the command-line by calling `Rscript` with the `-e` (expression) argument. For example, to convert a file from Stata (.dta) to comma-separated values (.csv), simply do the following:

```
Rscript -e "rio::convert('mtcars.dta', 'mtcars.csv')"
```

```{r, echo=FALSE, results='hide'}
unlink("mtcars.csv")
unlink("mtcars.dta")
unlink("multi.xlsx")
unlink("mtcars2.dta")
unlink("mtcars.tsv")
unlink("iris.tsv")
```

---
title: "Remapping and Ellipsis"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Remapping and Ellipsis}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Remapping

There are two conventions of arguments among the underlying functions used by `rio`. Let's call them *Base Convention* and *"Tidy" Convention*.

| Convention | file location | selection of sheet | header      | examples                                                     |
|------------|---------------|--------------------|-------------|--------------------------------------------------------------|
| Base       | `file`        | `which`            | `header`    | `clipr::read_clip_tbl`                                       |
| "Tidy"     | `path`        | `sheet`            | `col_names` | `readxl::read_xlsx`, `readxl::read_xls`, `readODS::read_ods` |

`rio` can map Base Convention into "Tidy" Convention (but not vice versa).

```{r map1}
library(rio)
export(list("mtcars" = mtcars, "iris" = iris), "example.xlsx")
import("example.xlsx", which = "mtcars")
```

But you can still use the "Tidy" Convention, if the underlying function supports it.

```{r map2}
import("example.xlsx", sheet = "mtcars")
```

# Ellipsis or "dot dot dot"

Additional parameters are usually passed to the underlying function as ellipsis (`...`).

```{r map3}
## n_max is an argument of readxl::read_xlsx
import("example.xlsx", sheet = "iris", n_max = 10)
```

Parameters that the underlying function do not recognize are silently ignored by default.

```{r map4}
import("example.xlsx", sheet = "iris", n_max = 10, pizza = "pineapple")
```

If you don't like this behavior, please change the option `rio.ignoreunusedargs` to `FALSE`, i.e. `options(rio.ignoreunusedargs = FALSE)`.

```r
options(rio.ignoreunusedargs = FALSE)
import("example.xlsx", sheet = "iris", n_max = 10, pizza = "pineapple")
```

```{r map5, error = TRUE, echo = FALSE, purl = FALSE}
R.utils::withOptions({
    import("example.xlsx", sheet = "iris", n_max = 10, pizza = "pineapple")
}, rio.ignoreunusedargs = FALSE)
```

```{r, echo = FALSE, results = 'hide'}
unlink("example.xlsx")
```

# Glyph Protocol

**Author:** Raphael Amorim
**Year:** 2026
**Last updated:** 2026-05-03

**See also:**
- Blog post introducing the protocol and its rationale:
  <https://rapha.land/introducing-glyph-protocol-for-terminals/>
- Reference implementation: [Rio terminal](https://raphamorim.io/rio)
- Example apps (ratatui, bubbletea v2, ink) registering real Nerd Font
  outlines and colrv1/colrv0 emojis at empty PUA-B slots:
  [glyph-protocol-examples](https://github.com/raphamorim/glyph-protocol-examples)

---

## Abstract

Glyph Protocol is a terminal protocol that lets applications ship
custom vector glyphs to the terminal at runtime — monochrome via
OpenType `glyf`, or full-colour via OpenType `COLR` v0 (flat layered
colour) and `COLR` v1 (paint graph with gradients, transforms, and
composites) — without requiring the user to install a patched font
(Nerd Fonts, Powerline, etc.). Registrations are restricted to the
Unicode Private Use Areas — ranges the user never types and existing
text never contains — so the protocol cannot be used to modify the
appearance of real text.

The protocol is transported over APC (Application Program Command)
sequences. The default payload is the OpenType `glyf` simple-glyph
record for monochrome icons; colour icons ride OpenType `COLR` v0
(layered flat colour) or `COLR` v1 (full paint graph). Four verbs
are defined: support-negotiation (`s`), query (`q`), register
(`r`), and clear (`c`).

## 1. Motivation

Terminal applications today rely on out-of-band font distribution
to render non-ASCII iconography. The dominant workflow is:

1. Application author picks codepoints in the Unicode Private Use
   Area.
2. User installs a multi-megabyte patched font that maps those
   codepoints to glyphs.
3. User switches their terminal's font to the patched font.
4. Application emits the codepoint and hopes the mapping is correct.

This workflow has three structural problems:

- **Distribution cost.** Users carry megabytes of glyphs they
  never see.
- **Coupling.** Adding a new icon requires the entire font
  ecosystem to update. Application authors are locked into a fixed
  PUA allocation.
- **Invisible failure.** An application cannot tell whether a
  given codepoint will render; it can only emit it and accept the
  result.

Glyph Protocol moves glyph ownership from the font file to the
application, and gives applications a way to ask the terminal what
it can render before it renders it.

## 2. Design goals

- **Small surface.** Few verbs, three payload formats (one
  required, two optional), no daemons, no caches, no cross-session
  state.
- **Zero new terminal dependencies.** Every terminal that renders
  OpenType text already links a `glyf` rasterizer; terminals that
  render Apple / Google colour emoji also already parse `COLR` +
  `CPAL`. No new format support is required — the protocol reuses
  the tables the font stack already decodes.
- **Resolution independent.** Glyphs are vector and scale to any
  cell size.
- **Graceful degradation.** Terminals that do not implement the
  protocol ignore the APC message. Applications detect support by
  sending a query and watching for a reply.
- **No override of user text.** Registrations are confined to PUA
  codepoints — ranges no user types and no pre-existing text
  contains. The rendered appearance of `a`, `ssh`, or any URL
  cannot be changed by any program that writes to the terminal.
  See §9.
- **Small on the wire.** Typical icons are 50–400 bytes of
  `glyf`, 2–4× smaller than the equivalent SVG.

## 3. Transport

Glyph Protocol uses APC (Application Program Command,
`ESC _ ... ESC \`). APC is specified for application-defined
commands; terminals that do not implement a particular APC command
are required to ignore it, making APC safer than OSC for
introducing new protocols.

### 3.1 Identifier

Every Glyph Protocol message begins with the Unicode codepoint
**U+25A1** (WHITE SQUARE), written in the message as the
lowercase hex string `25a1`. Terminals MUST ignore any APC
message whose body does not begin with this identifier.

### 3.2 Framing

The general form of a Glyph Protocol message is:

```
ESC _ 25a1 ; <verb> [ ; key=value ]* [ ; <payload> ] ESC \
```

Parameter keys use lowercase ASCII. Values are lowercase hex for
codepoints, decimal for integers, base64 for binary payloads, and
lowercase ASCII names for enums (single value, or comma-separated
for lists). The `status` field's shape is verb-specific: `r` and
`c` return a `u8` (`0` = success, nonzero paired with `reason=`);
`q` returns a comma-separated coverage list (§5.2).

### 3.3 Verbs

| Verb | Meaning |
|------|---------|
| `s`  | Advertise supported payload formats. Doubles as a protocol-detection ping — any reply confirms Glyph Protocol; a timeout means unsupported. |
| `q`  | Query the state of a codepoint. |
| `r`  | Register a glyph for a PUA codepoint. |
| `c`  | Clear one slot or every slot in this session's glossary. |

The `s` verb takes no parameters. The reply lists the terminal's
supported payload formats in the `fmt` key as a comma-separated
list of lowercase ASCII names, with no surrounding whitespace:

```
ESC _ 25a1 ; s ; fmt=glyf,colrv0,colrv1 ESC \
```

Format names defined in v1.8:

| Name     | Format                                                   |
|----------|----------------------------------------------------------|
| `glyf`   | Monochrome simple-glyph (§8).                            |
| `colrv0` | Layered flat colour, OpenType `COLR` v0 + `CPAL` (§8.6). |
| `colrv1` | OpenType paint graph, `COLR` v1 (§8.7).                  |

Order is not significant; clients MUST treat the value as a set.
An empty `fmt=` value means the terminal recognises Glyph Protocol
but currently advertises no payload formats — every `r` will be
rejected. Clients MUST ignore names they do not recognise rather
than failing the reply, so future format names are forward-
compatible. A terminal that advertises only `fmt=glyf` and
receives an `r` with `fmt=colrv0` / `fmt=colrv1` MUST reject the
registration (`reason=malformed_payload` is acceptable for
parser-level rejection). Clients SHOULD check the `s` reply
before emitting colour registrations so they can fall back to a
monochrome `fmt=glyf` without making a doomed round-trip.

## 4. Glossary namespace

Registrations target codepoints the application picks, constrained
to the three Unicode Private Use Areas:

| Range                        | Plane | Common use |
|------------------------------|-------|------------|
| `U+E000`–`U+F8FF`             | BMP   | Basic PUA. Nerd Fonts, Powerline, Font Awesome all live here. |
| `U+F0000`–`U+FFFFD`           | 15    | Supplementary PUA-A. Nerd Fonts v3 Material icons live here. |
| `U+100000`–`U+10FFFD`         | 16    | Supplementary PUA-B. No common convention — clean space for apps that want it. |

Any other codepoint — ASCII, Latin-1, CJK, emoji, control chars —
is rejected by `r` and `c` with `reason=out_of_namespace`.

Each terminal session holds at most **1024 simultaneous
registrations**. One codepoint = one slot regardless of payload
format: a `fmt=colrv1` registration with 500 inner outlines still
consumes exactly one glossary slot. When the glossary is full and
a new `r` arrives, the terminal evicts the oldest registration in
FIFO order; the new registration succeeds. Applications that
cannot tolerate silent eviction SHOULD query their codepoint with
`q` before emitting.

Each terminal session (tab, pane, PTY) owns its own glossary. Two
sessions can independently register `U+E0A0`, each pointing at a
different glyph. Registrations MUST NOT leak between sessions.

## 5. Query (`q`)

### 5.1 Request

```
ESC _ 25a1 ; q ; cp=<hex> ESC \
```

Parameters:

- `cp` — codepoint in hex. Any valid Unicode scalar value (not a
  surrogate). May be inside or outside PUA.

### 5.2 Response

```
ESC _ 25a1 ; q ; cp=<hex> ; status=<list> ESC \
```

`status` is a comma-separated list of coverage names — the set of
sources that can render `cp` in this session. Order is not
significant; clients MUST treat the value as a set.

| `status=` value     | Meaning |
|---------------------|---------|
| (empty)             | No font in the fallback chain renders `cp`, and the glossary has no registration for it. The cell will render as tofu. |
| `system`            | Some font in the fallback chain renders `cp`. No glossary registration (or `cp` is outside PUA). |
| `glossary`          | `cp` is in PUA and has a live registration in this session. No system font covers it. |
| `system,glossary`   | `cp` is in PUA, has a live registration, AND a system font also covers it. The registration shadows the system font at render time. |

For non-PUA codepoints only the empty value and `status=system`
are possible. Clients MUST ignore unknown coverage names so
future sources are forward-compatible.

## 6. Register (`r`)

### 6.1 Request

```
ESC _ 25a1 ; r ; cp=<hex> ; fmt=glyf ; reply=<0|1|2> ; upm=<int> ; aw=<int> ; lh=<int> ; width=<1|2> ; size=<mode> ; align=<h>,<v> ; pad=<t>,<r>,<b>,<l> ; <base64-payload> ESC \
```

Parameters:

- `cp` — target codepoint in hex. MUST be in one of the PUA ranges
  defined in §4. Otherwise the request is rejected with
  `reason=out_of_namespace`.
- `fmt` — payload format. One of `glyf`, `colrv0`, `colrv1`.
  Optional; `glyf` is the default. See §8 for each format's wire
  layout.
- `reply` — reply-level control. Optional; default `1`.
  - `reply=0` — the terminal MUST NOT emit any reply for this
    registration (neither success nor failure). Intended for bulk
    fire-and-forget startup registrations that won't be read back.
  - `reply=1` — the terminal emits both success and failure replies
    (the default; equivalent to omitting the parameter).
  - `reply=2` — the terminal emits failure replies only; success
    registrations are silent. Useful for bulk registrations that
    still want to learn about the broken ones without a success
    ACK for every glyph.
  Unknown values fall back to `reply=1`.
- `upm` — units per em, the coordinate space the outline is
  authored in. Optional; default `1000`.
- `aw` — authored advance width, in upm units. The intended
  horizontal extent of the glyph, NOT the outline's bounding box.
  Optional; default `upm`.
- `lh` — authored line height, in upm units. The intended vertical
  extent (descender-to-ascender), NOT the outline's bounding box.
  Optional; default `upm`.
- `width` — the codepoint's Unicode width, in the `wcwidth` /
  UAX #11 sense. One of `1` (narrow) or `2` (wide). Optional;
  default `1`. Authoritative for all terminal layout decisions
  (cursor advance, wrapping, selection geometry), overriding the
  codepoint's UAX #11 East Asian Width (all PUA ranges are
  Ambiguous by default).
- `size` — scale policy. One of `height`, `advance`, `contain`,
  `cover`, `stretch`. Optional; default `height`. See §8.5.
- `align` — placement of the scaled outline within the render span,
  as a comma-separated pair `<h>,<v>`. `<h>` is one of `start`,
  `center`, `end`; `<v>` is one of `start`, `center`, `end`,
  `baseline`. Optional; default `center,center`. See §8.5.
- `pad` — insets from the render span edges, as comma-separated
  fractions `<top>,<right>,<bottom>,<left>` (each `0.0`–`1.0`;
  top/bottom are fractions of cell height, left/right of render
  span width). Optional; default `0,0,0,0`. See §8.5.
- payload — base64-encoded payload for the declared `fmt`.

### 6.2 Response

Replies are gated by the request's `reply` parameter (§6.1).
For `reply=1` (the default), successful registrations emit:

```
ESC _ 25a1 ; r ; cp=<hex> ; status=0 ESC \
```

`cp` is echoed from the request.

Failures, when not suppressed by `reply=0`, emit:

```
ESC _ 25a1 ; r ; cp=<hex> ; status=<nonzero u8> ; reason=<code> ESC \
```

At `reply=2`, successful registrations are silent; failures still
emit the error reply above. At `reply=0`, neither success nor
failure produces any output — the registration is fire-and-forget.

Defined error codes:

| Code                    | Meaning |
|-------------------------|---------|
| `out_of_namespace`      | `cp` is not in any PUA range. |
| `composite_unsupported` | Payload contains composite glyphs. |
| `hinting_unsupported`   | Payload contains hinting instructions. |
| `malformed_payload`     | Payload failed to parse as `glyf`. |
| `payload_too_large`     | Payload exceeds 64 KiB post-base64-decode. |

### 6.3 Overwrite and eviction

A second `r` on the same `cp` overwrites the first. This is how
applications update a glyph or react to theme changes.

When the glossary already holds 1024 registrations and the new `r`
is for a `cp` that is NOT already registered, the terminal evicts
the oldest registration (FIFO) to make room. Eviction silently
invalidates the evicted codepoint: subsequent emissions fall
through to the system font (or tofu). Applications SHOULD query
before emitting if they cannot tolerate silent eviction.

### 6.4 Lifetime

Registrations live for the duration of the terminal session. A
terminal reset (e.g. `ESC c`) MAY clear the entire glossary.
Registrations MUST NOT persist across terminal restarts.

## 7. Clear (`c`)

### 7.1 Request

```
ESC _ 25a1 ; c [ ; cp=<hex> ] ESC \
```

If `cp` is omitted, every slot in the session's glossary is
cleared. Otherwise the slot corresponding to `cp` is cleared. `cp`
MUST be in a PUA range; otherwise the request is rejected with
`reason=out_of_namespace`.

### 7.2 Response

Success:

```
ESC _ 25a1 ; c ; status=0 ESC \
```

Clearing an empty slot is a no-op and MUST return `status=0`.

Failure:

```
ESC _ 25a1 ; c ; status=1 ; reason=out_of_namespace ESC \
```

### 7.3 Cache invalidation

When a slot is cleared (explicitly, via overwrite, or via
eviction), the terminal MUST invalidate any rasterization cached
for that codepoint. A subsequent `r` that reuses the codepoint
MUST rasterize the new outline fresh, not serve stale pixels.

## 8. Payload format: `glyf`

### 8.1 Scope

Glyph Protocol reuses the OpenType `glyf` table's simple-glyph
record as its wire format. Authoritative references:

- OpenType `glyf` specification (Microsoft Typography).
- Apple TrueType Reference Manual, Chapter 6.

### 8.2 Constraints

Terminals implementing Glyph Protocol MUST accept the following
subset of `glyf` and MAY reject anything else with
`reason=composite_unsupported` or `reason=hinting_unsupported`:

- **Simple glyphs only** (in v1). No composite glyphs, no references
  to other glyphs. Composite support may land in a future version
  once the simple-glyph baseline is settled in the field.
- **Standard flag encoding** as defined by the OpenType spec
  (on-curve, off-curve, x-short, y-short, repeat).
- **No hinting instructions** (in v1). The `instructionLength`
  field MUST be zero. Bytecode hinting may be allowed in a future
  version once the unhinted baseline is settled; for now the
  terminal's own anti-aliasing is what users see.
- **Coordinate space** defined by `upm`. The terminal maps this
  space onto its cell at render time.

### 8.3 Contour semantics

A `glyf` record stores a glyph as a set of closed contours. Each
contour is a sequence of points; each point carries a single
on-curve/off-curve flag bit. Contour walking follows standard
TrueType semantics:

- Two on-curve points in a row → straight line.
- An off-curve point between two on-curve points → quadratic
  Bézier with the off-curve point as the control point.
- Two off-curve points in a row → an implied on-curve point at
  their midpoint.

### 8.4 Color

`glyf` outlines carry no color. Terminals MUST render them in the
current foreground color. For colored icons see the `colrv0` and
`colrv1` formats in §8.6 / §8.7.

### 8.5 Sizing, placement, and coordinate convention

Every registered outline passes through three transforms at render
time, in order: **pad** (compute the effective render span),
**size** (pick scale factors), **align** (position the scaled
outline within the span).

#### 8.5.1 Coordinate convention

Outlines are authored in `upm`-unit space, Y-up, with `y=0` at the
baseline. The authored extent is the rectangle `[0, aw] × [y_min,
y_max]` where `lh = y_max − y_min` (matching OpenType line height:
descender to ascender, with descender ≤ 0). Points outside this
rectangle are allowed — they clip or overflow per the `size`/align`
rules below.

#### 8.5.2 Render span and padding

The render span is the rectangle of pixels the outline is scaled
into. Before scaling, the span is:

```
W = width × cell_width_px      (cell_width_px from the terminal)
H = cell_height_px
```

`pad=<t>,<r>,<b>,<l>` shrinks the span:

```
W' = W × (1 − l − r)
H' = H × (1 − t − b)
```

Padding values are fractions. If `l + r ≥ 1` or `t + b ≥ 1` the
terminal MUST treat the request as if `pad=0,0,0,0` (no padding)
— a degenerate span is not useful and suggests a client bug. The
effective span `W' × H'` is what `size` and `align` operate on.

#### 8.5.3 Size modes

Given the authored extent `aw × lh` and effective span `W' × H'`:

| `size`     | `sx`              | `sy`              | Aspect ratio | Notes |
|------------|-------------------|-------------------|--------------|-------|
| `height`   | `H' / lh`         | `H' / lh`         | Preserved    | Default. Line-height drives. Horizontal extent is whatever `aw` scales to; may overflow the span. |
| `advance`  | `W' / aw`         | `W' / aw`         | Preserved    | Advance drives. Vertical extent is whatever `lh` scales to; may overflow. |
| `contain`  | `min(W'/aw, H'/lh)` | same            | Preserved    | Outline fits entirely inside the span on both axes. May leave empty space on one axis. |
| `cover`    | `max(W'/aw, H'/lh)` | same            | Preserved    | Outline fills the span on both axes. May overflow on one axis. |
| `stretch`  | `W' / aw`         | `H' / lh`         | Not preserved | Each axis independent. Useful for box-drawing. |

`height` is the default because it matches how characters behave
(the terminal's line-height maps to the cell's vertical pixels,
the horizontal footprint is what the glyph's own advance dictates).
For icons that must stay inside the cell regardless of aspect,
prefer `contain`.

#### 8.5.4 Alignment

After scaling, the outline has a scaled authored extent `(aw×sx)
× (lh×sy)` positioned somewhere within the effective span. `align`
picks where.

Horizontal:

- `start` — outline's `x=0` aligns with the span's left edge
  (`pad_left`).
- `center` — outline's horizontal midpoint aligns with the span's
  horizontal midpoint.
- `end` — outline's `x=aw` aligns with the span's right edge
  (`W − pad_right`).

Vertical (Y-up):

- `start` — outline's `y=y_min` aligns with the span's bottom edge
  (`pad_bottom`).
- `center` — outline's vertical midpoint aligns with the span's
  vertical midpoint.
- `end` — outline's `y=y_max` aligns with the span's top edge
  (`H − pad_top`).
- `baseline` — outline's `y=0` aligns with the terminal's text
  baseline within the cell. Preferred for character-like glyphs
  that must sit on the same baseline as surrounding text;
  descenders extend below the baseline naturally.

When `size=stretch`, the scaled extent exactly matches the span on
both axes, so `align` has no visible effect. When `size=cover`,
the scaled extent overflows on one axis; `align` picks which edge
to anchor (and therefore which overflow is visible vs. clipped).
When `size=contain`/`height`/`advance`, the scaled extent may be
smaller than the span on at least one axis; `align` picks where
the empty space goes.

#### 8.5.5 Resolution independence

Applications MUST NOT assume a particular cell size and MUST NOT
re-register glyphs on font size change. All scaling is computed
at render time from the parameters above and the terminal's
current cell metrics.

#### 8.5.6 Coordinated sets (no scale groups)

There is no `group` parameter. A set of glyphs that must visually
align — spinner frames, progress-bar steps, a multi-glyph logo —
aligns automatically if authored with identical `aw`, `lh`, `size`,
`align`, and `pad`, and if their outline geometry is coordinated
(e.g. all spinner frames sized inside a common bounding circle).
Scale-group semantics can be added later if authoring experience
shows they are genuinely needed; for v1 the burden sits with
the author, not the protocol.

### 8.6 Payload format: `colrv0`

`fmt=colrv0` carries a layered flat-colour glyph using the
OpenType `COLR` v0 and `CPAL` tables verbatim. The protocol wraps
those tables in a small container that also ships the simple-glyph
outlines each layer references, so a colour glyph is self-
contained: no external font needed.

**Container layout** (all integers big-endian, post-base64-decode):

```
u16     n_glyphs              # 1..=1024
repeat n_glyphs:
  u16   glyf_len
  glyf_len bytes              # simple-glyph record, §8.2 subset
u16     colr_len              # > 0
colr_len bytes                # OpenType COLR v0 table
u16     cpal_len              # may be 0 (see below)
cpal_len bytes                # OpenType CPAL table (required for v0)
```

`GlyphId` values in the `COLR` table resolve to indices into the
outline array (glyph 0 is the base glyph rendered when the
terminal emits `cp`). `paletteIndex` values in the `COLR` layer
records resolve to entries in the CPAL colour records array, in
standard OpenType order (one record = one BGRA quadruple).
`paletteIndex = 0xFFFF` MUST be rendered as the current foreground
colour, per the OpenType spec.

**Rendering rules.**

- Layers composite in painter order (first layer painted first).
- Per-layer colours come from CPAL; `0xFFFF` means foreground.
- `COLR` v0 defines no transforms or compositing modes beyond
  `src-over`, so terminals MAY implement v0 in one pass with no
  graphics-state stack.

**Validation.** Terminals SHOULD validate the wrapped `COLR` and
`CPAL` tables using an OpenType parser (e.g. `ttf-parser`); a
`COLR` table that fails to parse SHOULD be rejected with
`reason=malformed_payload`. Every carried outline MUST satisfy the
`glyf` simple-glyph subset of §8.2; violations use the same error
codes as `fmt=glyf`.

### 8.7 Payload format: `colrv1`

`fmt=colrv1` shares the container layout of §8.6 but ships an
OpenType `COLR` v1 table, which adds a full paint graph: linear,
radial, and sweep gradients, affine transforms, clip boxes, and
per-layer compositing modes. `CPAL` remains valid but is optional
— v1 paints may carry sRGBA directly — so `cpal_len = 0` is
permitted and means "the COLR references no palette index."

**Paint types.** Terminals implementing `colrv1` SHOULD support
the full OpenType paint-graph vocabulary for maximum interop. A
conforming subset for low-overhead implementations is:

- Solid (direct sRGBA or palette index).
- Linear gradient.
- Radial gradient.
- Affine transforms on paint subtrees.
- `src-over` layer composite.

Terminals MAY render unsupported paint nodes (sweep gradients,
blend modes beyond `src-over`, variations) using a reasonable
fallback — typically the paint subtree's first solid colour —
rather than rejecting the registration.

**Foreground inheritance.** CPAL palette index `0xFFFF` and
v1's `PaintSolid` with the foreground sentinel MUST resolve to
the cell's current foreground colour at rasterisation time.
Terminals that cache rasterised colour glyphs MUST re-rasterise
on foreground change for any glyph whose paint graph references
`0xFFFF`.

**Security.** The colour formats add no new attack surface beyond
§9: `cp` is still PUA-only, the cell buffer is still authoritative
for copy/selection, and registrations are still session-scoped.
A malformed `COLR` is a rendering error, not an injection vector
— the rendered pixels can only affect cells the client itself
emits at a PUA codepoint.

### 8.8 Authoring

Most applications will not hand-author `COLR` bytes either.
Typical flows:

- **From an existing colour font.** Use `fontTools` to extract the
  `COLR`/`CPAL` tables for the glyphs of interest, then pack them
  with the referenced outlines into the container above.
- **From SVG.** The Skia team publishes `nanoemoji` / `maximum-color`,
  which compiles a directory of SVGs into a `COLR` v1 font; feed
  its output into the packer.

## 9. Security considerations

The core property Glyph Protocol must preserve is that **an
application cannot change how existing text looks**. Enforcement
is structural:

- Register accepts a `cp` parameter, but `cp` MUST be in one of the
  three Unicode Private Use Areas (§4). Any non-PUA codepoint is
  rejected with `reason=out_of_namespace`.
- Users never type PUA codepoints. No pre-existing text —
  filenames, URLs, commands, variable names, log lines — contains
  them. A program that registers a glyph can only affect how PUA
  codepoints render, and PUA codepoints only appear in text the
  same application (or another one opting into the same
  convention) has deliberately emitted.
- The cell buffer is authoritative. Selection, copy, search,
  hyperlinks, shell history, and any other text extraction MUST
  return the codepoint the application emitted, never the
  rendered glyph.

Without these properties, a program writing to the PTY could
register a glyph for `a` that looks like `o` and mislead the
reader. With them, the worst a program can do is render a
weird-looking character at a PUA codepoint the user never types
and the cell buffer honestly reports.

Other considerations:

- **Resource bounds.** The 1024-slot cap and 64 KiB per-payload cap
  give a hard upper bound of 64 MiB on the glossary's memory
  footprint per session.
- **No code execution.** The `glyf` subset defined in §8.2
  excludes hinting instructions, which is the only part of
  TrueType that is executable. Glyph Protocol is purely
  declarative.
- **No filesystem access.** Glyph Protocol messages do not
  reference files and MUST NOT be used to load data from disk.
- **Session isolation.** Glossaries MUST NOT leak between terminal
  tabs, windows, multiplexer panes, or PTY sessions.

## 10. Non-goals (v1)

- **No non-PUA codepoints.** Registration is restricted to the
  three PUA ranges — see §4.
- **No ligatures.** Registration applies to a single codepoint.
  Sequence-keyed substitution is out of scope; programming
  ligatures are already handled by OpenType fonts.
- **No persistence across sessions.** Glyphs are shipped fresh on
  each run.
- **No cross-application sharing.** Each terminal session owns its
  glossary. No IPC, no daemon.
- **No bitmap colour glyphs.** Colour is delivered via `colrv0`
  and `colrv1` (§8.6 / §8.7), which are vector. `CBDT`/`sbix`/
  `SVG ` tables are explicitly out of scope so resolution
  independence is preserved.
- **No subpixel positioning control.** The terminal's normal cell
  positioning applies.
- **No bitmap payloads.** Vector only, to preserve resolution
  independence.

## 11. Conformance

A terminal emulator is Glyph Protocol v1 conformant if it:

1. Recognizes the `25a1` identifier in APC sequences.
2. Implements the `s`, `q`, `r`, and `c` verbs with the semantics
   defined in this specification, and lists every accepted payload
   format by name in the `fmt=` value of the `s` reply.
3. Restricts register/clear `cp` to the three PUA ranges; rejects
   anything else with `reason=out_of_namespace`.
4. Holds at most 1024 simultaneous registrations per session and
   evicts in FIFO order when full.
5. Accepts the `glyf` simple-glyph subset defined in §8.2. The
   `colrv0` and `colrv1` formats are OPTIONAL; terminals that
   accept them MUST list the corresponding name in the `s` reply.
6. Renders registered `glyf` glyphs in the current foreground
   color; renders `colrv0`/`colrv1` glyphs using the COLR paint
   graph, resolving palette index `0xFFFF` to the current
   foreground color.
7. Scales and positions glyphs according to `upm`, `aw`, `lh`,
   `width`, `size`, `align`, and `pad` as specified in §8.5, and
   treats the registered codepoint as having the declared `width`
   (`1` or `2`) for every layout decision, overriding the
   codepoint's UAX #11 East Asian Width.
8. Enforces the cell-buffer authority invariant in §9: selection,
   copy, and search return the raw codepoint.
9. Ignores unrecognized parameters rather than failing the
   request.

A client (application) is Glyph Protocol v1 conformant if it:

1. Emits `cp` only from the three PUA ranges.
2. Treats query timeout as "terminal does not implement Glyph
   Protocol."
3. Emits only the `glyf` subset defined in §8.2.
4. Handles all `reason=*` error codes without crashing.

## 12. Reference implementation

The reference implementation ships in Rio terminal. A companion
helper, `svg2glyf`, ships alongside to convert existing SVG assets
to the accepted `glyf` subset at build time.

## Appendix A. Worked example: register an icon in empty PUA

```python
import base64, sys
from fontTools.pens.ttGlyphPen import TTGlyphPen

# A stylised outline in glyf coordinate space (upm=1000, Y-up).
pen = TTGlyphPen(None)
# ... draw commands ...
pen.closePath()

payload = base64.b64encode(pen.glyph().compile(None)).decode("ascii")

# Register at U+100000 — the first codepoint of Supplementary PUA-B.
# No known font covers this range, so the registration is the sole
# source of the rendered glyph and the demo is unambiguous.
sys.stdout.write(
    f"\x1b_25a1;r;cp=100000;upm=1000;{payload}\x1b\\"
)
sys.stdout.flush()

# From now on, U+100000 renders our outline.
sys.stdout.write(f"icon: {chr(0x100000)}\n")
```

## Appendix B. Worked example: query before registering

```python
import sys

def q(cp: int) -> None:
    sys.stdout.write(f"\x1b_25a1;q;cp={cp:x}\x1b\\")
    sys.stdout.flush()

# Does the user already have Nerd Fonts installed?
q(0xE0A0)
# Expected reply parsed from the PTY:
#   status=system → system font covers it; don't register, just emit cp
#   status=       → empty list (nothing covers it); register and emit
```

## Appendix C. Implementation notes

These are not normative but reflect lessons from the first
implementations.

**Response draining.** By default `r` and `c` produce an APC reply
on the PTY. Client applications that register at startup and do not
care about the reply have three options, in order of preference:

1. Use `reply=0` on the register request (§6.1). The terminal emits
   nothing, so there's nothing to drain and nothing to leak. Best
   for bulk startup registrations.
2. Use `reply=2` to keep failure replies but drop success ACKs.
   Retains debuggability (you still learn about malformed payloads)
   without the success-reply noise of a 100-glyph registration.
3. Let the framework's input reader swallow the reply — safe only
   while that reader is alive.

The failure mode to watch for is sending `r` or `c` with `reply=1`
AFTER the framework has torn down its input reader — typically on
exit — at which point the reply arrives in the PTY but nobody reads
it, and the shell that takes over the PTY after the app exits emits
the queued bytes as visible text (`.25a1;c;status=0` or
`.25a1;r;cp=…;status=0`). For exit-time cleanup, prefer skipping
the `c` altogether (registrations expire with the session anyway).

**Practical source of `glyf` data.** Most apps will not hand-author
`glyf` bytes. The typical pipeline is:

1. Open a Nerd Font or similar icon TTF with `fontTools`.
2. For each codepoint of interest, pull the glyph record.
3. If composite, flatten via `fontTools.pens.ttGlyphPen.TTGlyphPen`.
4. Strip hinting instructions (`instructionLength := 0`).
5. Compile to bytes; base64-encode; register at a codepoint of the
   app's choosing.

Because the source codepoint in the font is irrelevant to the
protocol, applications commonly pull outlines from a Nerd Font's
basic-PUA codepoints (`U+E0A0`, `U+F07B`, …) and register them at
Supplementary PUA-B slots (`U+100000`+) — that way the rendered glyph
is unambiguously from the registration, not from a system font that
happens to cover the same codepoint.

**Atlas cache invalidation.** The cache-invalidation rule in §7.3
applies to overwrite and eviction too, not only explicit clear.
Implementations that key their glyph atlas on some stable slot id
(rather than `cp`) must ensure the slot id is either released on
clear/evict or paired with a per-registration invalidation tag, so
that a subsequent register reusing the id rasterizes fresh bytes
rather than serving a stale bitmap.

## Appendix D. Change log

| Date       | Version | Notes |
|------------|---------|-------|
| 2026-04-17 | v1      | Initial release. Register accepts a client-picked `cp` restricted to PUA; 256-slot glossary with FIFO eviction; numeric `status` field; ligatures out of scope. |
| 2026-04-19 | v1.1    | Added `s` verb (support advertisement / protocol ping). |
| 2026-04-19 | v1.2    | Added `fmt=colrv0` and `fmt=colrv1` payload formats wrapping OpenType `COLR` / `CPAL` tables with sidecar `glyf` outlines. Both advertised via bits 1 and 2 of the `s` reply's `fmt=` bitfield. |
| 2026-04-19 | v1.3    | Added `reply=0|1|2` parameter to the `r` verb so bulk registrations can suppress success ACKs (`reply=2`) or go fully fire-and-forget (`reply=0`). Default `reply=1` preserves v1.0 behaviour. |
| 2026-04-19 | v1.4    | Raised the glossary capacity from 256 to 1024 simultaneous registrations per session, and raised the `n_glyphs` cap in `fmt=colrv0`/`colrv1` containers from 256 to 1024 outlines. Both bumps quadruple the worst-case memory footprint; the 64 KiB per-payload cap is unchanged. |
| 2026-04-21 | v1.5    | Added `cp=auto` to the `r` verb: the terminal allocates a free PUA codepoint (SHOULD come from PUA-B) and echoes it in the success reply so the client can emit it. Added `reason=auto_unsupported` and `reason=glossary_exhausted` error codes. `cp=auto` forces a success reply regardless of `reply=0|2` because the allocated codepoint is only carried in the reply. |
| 2026-04-23 | v1.6    | Removed `cp=auto` from the `r` verb (introduced in v1.5). Auto-allocation forced a stateful round-trip reply the client depended on to learn its codepoint, which recording tools like `asciinema` and `tee` cannot capture or replay — making `cp=auto` output impossible to reproduce from a transcript. Clients must pick their own PUA codepoint. The `auto_unsupported` and `glossary_exhausted` error codes are withdrawn. |
| 2026-04-23 | v1.7    | Added a sizing and placement model to the `r` verb: `aw` / `lh` (authored extent in upm units), `width` (Unicode/wcwidth width, `1` or `2`, authoritative), `size` (`height`/`advance`/`contain`/`cover`/`stretch`), `align` (`<h>,<v>` positioning after scale, with `v=baseline` for character-like glyphs), and `pad` (fractional insets from the render span). Pinned the coordinate convention: Y-up, `y=0` at baseline, `lh` measured descender-to-ascender (OpenType). Scale groups are intentionally omitted — coordinated sets align via matching parameters and outline geometry. |
| 2026-05-03 | v1.8    | Replaced the `s` reply's `u8` bitfield with a comma-separated list of format names (e.g. `fmt=glyf,colrv0,colrv1`). Names extend without bit-collision worries and stay readable in transcripts; an empty `fmt=` means the terminal advertises no payload formats. Unknown names MUST be ignored by clients, so future formats are forward-compatible. |
| 2026-05-03 | v1.9    | Replaced the `q` reply's `u8` two-bit `status` field with a comma-separated list of coverage names: `status=system`, `status=glossary`, `status=system,glossary`, or empty for "free". Same motivation as v1.8 for the `s` reply. The `r` and `c` replies still use `status=<u8>` for success/failure since that's a closed boolean, not an extensible set. |

---
title: 'Kitty keyboard protocol'
language: 'en'
---

Rio terminal implements Kitty keyboard protocol. It is enabled by default.

```toml
use-kitty-keyboard-protocol = true
```

### How it works?

> Note: All the information of this page was retired from: [sw.kovidgoyal.net/kitty/keyboard-protocol](https://sw.kovidgoyal.net/kitty/keyboard-protocol/)

There are various problems with the current state of keyboard handling in terminals. They include:

- No way to use modifiers other than ctrl and alt

- No way to reliably use multiple modifier keys, other than, shift+alt and ctrl+alt.

- Many of the existing escape codes used to encode these events are ambiguous with different key presses mapping to the same escape code.

- No way to handle different types of keyboard events, such as press, release or repeat

- No reliable way to distinguish single Esc key presses from the start of a escape sequence. Currently, client programs use fragile timing related hacks for this, leading to bugs, for example: [neovim #2035](https://github.com/neovim/neovim/issues/2035).

To solve these issues and others, kitty has created a new keyboard protocol, that is backward compatible but allows applications to opt-in to support more advanced usages.

You can see this protocol with all enhancements in action by running:

```sh
kitten show_key -m kitty
```

You can also run:

```sh
kitty +kitten show_key -m kitty
```

Below an example of Rio terminal with Kitty keyboard protocol by the following instructions (MacOS keyboard example).

- Press left shift, release shift
- Press right shift, release shift
- Press left command

![Demo kitty keyboard protocol](/assets/features/demo-kitty-keyboard-protocol.png)

# librio

`librio` is an embeddable terminal library extracted from Rio: the PTY and VT
state core, usable from Swift, C, and anything that speaks C, without pulling in
a windowing stack.

It began as a research exercise studying two prior art decompositions, the
libghostty split (a terminal-state library shipped first, renderer as a later
peer) and [libvterm](https://www.leonerd.org.uk/code/libvterm/), then kept the
parts that fit Rio's existing engine and dropped the rest.

`librio` is a single Rust crate that wraps [`rio-vt`](../rio-vt), Rio's terminal
core, and compiles to a C ABI. It is deliberately VT-only: PTY, VT state, and a
host-pulled render-state API, with no renderer.

## Distribution

`librio` is **not published to crates.io**. It lives in the Rio source tree and
exists only to produce static artifacts: a `RioKit.xcframework` (static lib +
curated header + `module.modulemap`, `module RioKit`) shipped in Rio's GitHub
releases, built with `cargo build` then `lipo` and
`xcodebuild -create-xcframework`. The C header is curated (cbindgen seeds it,
never auto-committed).

If you write Rust, depend on `rio-vt` directly; that is the published, idiomatic
crate. `librio` is the C boundary for everyone else, and versions with Rio.

## Design

Owns the PTY (`teletypewriter`), the VT machine (`Machine` / `performer`), and
the grid (`Crosswords`). Knows nothing about pixels: the host pulls a
render-state snapshot with per-row dirty flags (which the host resets) and
draws it however it likes. This pull model is the permanent dogfood test: if
our own renderer can be built on it, so can a third-party CPU renderer, a wasm
canvas, a TUI proxy, or a test harness.

The callback surface is deliberately small: `userdata` + a `wakeup_cb` + a
single enum-tagged `action_cb` + clipboard and close-surface callbacks. Adding
a new action never changes the ABI struct. Input is a per-surface set of entry
points: key (returns a consumed bool), text, preedit (IME), mouse
button/pos/scroll, set_size, set_content_scale, set_focus, set_occlusion.

### Rust API

```rust
pub struct Engine { /* shared config, action fan-out */ }
pub struct Surface { /* one PTY + grid; owns the IO thread */ }
pub struct RenderState { /* snapshot + dirty rows */ }

pub trait SurfaceDelegate: Send + Sync {
    fn wakeup(&self);                                  // any thread; schedule only
    fn action(&self, surface_id: SurfaceId, action: Action);
    fn clipboard_write(&self, kind: ClipboardKind, text: &str);
    fn close_surface(&self, surface_id: SurfaceId);
}

pub enum Action {
    SetTitle(String),
    RingBell,
    ChildExited(i32),
    CursorShape(CursorShape),
    // grows freely
}

impl Surface {
    pub fn new(engine: &Engine, desc: SurfaceDesc) -> Result<Surface>;
    pub fn key(&self, event: KeyEvent) -> bool;        // true = consumed
    pub fn text(&self, s: &str);
    pub fn preedit(&self, s: &str);
    pub fn mouse_button(&self, ...); pub fn mouse_pos(&self, ...);
    pub fn scroll(&self, dx: f64, dy: f64);
    pub fn set_size(&self, cols: u16, rows: u16, px_w: u32, px_h: u32);
    pub fn set_content_scale(&self, scale: f64);
    pub fn set_focus(&self, focused: bool);
    pub fn set_occlusion(&self, occluded: bool);
}

impl RenderState {
    pub fn new(surface: &Surface) -> RenderState;
    pub fn update(&mut self);                          // snapshot under the terminal lock
    pub fn dims(&self) -> (u16, u16);
    pub fn row_dirty(&self, row: u16) -> bool;
    pub fn reset_dirty(&mut self);
    pub fn cell(&self, row: u16, col: u16) -> Cell;    // materialized view
    pub fn row_cells(&self, row: u16) -> &[Cell];
    pub fn cursor(&self) -> Cursor;
    pub fn selection(&self) -> Option<SelectionRange>;
    pub fn palette(&self) -> &Palette;
}
```

### C ABI

```c
typedef struct rio_engine        rio_engine_t;
typedef struct rio_config        rio_config_t;
typedef struct rio_surface       rio_surface_t;
typedef struct rio_render_state  rio_render_state_t;

typedef enum {
  RIO_ACTION_SET_TITLE, RIO_ACTION_RING_BELL,
  RIO_ACTION_CHILD_EXITED, RIO_ACTION_CURSOR_SHAPE,
} rio_action_tag_t;
typedef struct { rio_action_tag_t tag; /* tagged union payload */ } rio_action_s;

typedef struct {
  void *userdata;
  void (*wakeup_cb)(void *ud);                    /* any thread; schedule only */
  void (*action_cb)(void *ud, rio_surface_t*, rio_action_s);
  void (*clipboard_write_cb)(void *ud, int kind, const char *utf8);
  void (*close_surface_cb)(void *ud, rio_surface_t*);
} rio_runtime_config_s;

rio_engine_t  *rio_engine_new(const rio_runtime_config_s*, const rio_config_t*);
void           rio_engine_free(rio_engine_t*);

rio_surface_t *rio_surface_new(rio_engine_t*, const rio_surface_config_s*);
void           rio_surface_free(rio_surface_t*);
bool rio_surface_key(rio_surface_t*, rio_key_event_s);   /* true = consumed */
void rio_surface_text(rio_surface_t*, const char*, uintptr_t);
void rio_surface_preedit(rio_surface_t*, const char*, uintptr_t);
void rio_surface_mouse_button(rio_surface_t*, rio_mouse_event_s);
void rio_surface_mouse_pos(rio_surface_t*, double x, double y);
void rio_surface_scroll(rio_surface_t*, double dx, double dy);
void rio_surface_set_size(rio_surface_t*, uint16_t cols, uint16_t rows,
                          uint32_t px_w, uint32_t px_h);
void rio_surface_set_content_scale(rio_surface_t*, double);
void rio_surface_set_focus(rio_surface_t*, bool);
void rio_surface_set_occlusion(rio_surface_t*, bool);

rio_render_state_t *rio_render_state_new(rio_surface_t*);
void rio_render_state_free(rio_render_state_t*);
void rio_render_state_update(rio_render_state_t*);
bool rio_render_state_row_dirty(const rio_render_state_t*, uint16_t row);
void rio_render_state_reset_dirty(rio_render_state_t*);
void rio_render_state_dims(const rio_render_state_t*, uint16_t *cols, uint16_t *rows);
const rio_cell_s *rio_render_state_row(const rio_render_state_t*, uint16_t row,
                                       uintptr_t *out_len);
rio_cursor_s rio_render_state_cursor(const rio_render_state_t*);
```

Every `extern "C"` entry point wraps its body in `catch_unwind`, a Rust panic
must never unwind into Swift/C (UB).

Threading contract (identical to today's rioterm split): the surface owns one
PTY IO thread; `wakeup_cb`/`action_cb` may fire on it and must only
flag/schedule. All `render_state_*` calls and input calls are host-thread safe
(input posts to the IO channel; render-state update takes the FairMutex).

---
layout: post
title: 'Rio 0.1.x is here'
date: 2024-06-12 12:34
description: 'Font ligatures, emojis'
categories: windows macos linux
draft: true
---

import { Tweet } from 'react-tweet';

Font ligatures, emojis, new macos client.

## Support to Font ligatures

Demo usage of Rio terminal on MacOS using [Fira Code](https://fonts.google.com/specimen/Fira+Code) and font ligatures enabled:

![Font ligatures demo](/assets/posts/0.1.0/demo-rio.png)

## Support to Emojis

TODO

## Support to MacOS open file URLs

![Macos open files](/assets/posts/0.1.0/rio-demo-open-file.gif)

## Terminal integration with VS Code (MacOS and Windows)

![Macos open folders](/assets/posts/0.1.0/rio-demo-open-folder.gif)

---
layout: post
title: 'Rio 0.0.7'
date: 2023-07-07 12:34
description: 'Configuration reload, x11 and wayland stabilization, sugarloaf ported to WebAssembly, addition to NIX OS packages and other updates.'
categories: release macos linux webassembly
---

import { Tweet } from 'react-tweet';

## v0.0.7: Highlights

- Support to custom fonts
- Configuration reload
- x11 and wayland stabilization
- Sugarloaf ported to WebAssembly
- Addition to NIX OS packages (kudos to [@otavio](https://github.com/otavio), PR: [NixOS/nixpkgs/pull/237664](https://github.com/NixOS/nixpkgs/pull/237664))
- Styles properties have been moved to root level

## Overview

Rio release 0.0.7 is finally here, there's a lot of updates to cover so let's get started. I also would like to remind you that Rio is still unstable. This release adds a lot of bug fixes and feature additions in order to make Rio terminal stable.

## Breaking change: Styles properties have been moved to root level

The only breaking change of v0.0.7 is that styles properties such as "font", "theme" and "font-size" moved to the root level.

Before:

```toml
cursor = '|'
[style]
font-size = 18
theme = "lucario"
```

After:

```toml
cursor = '|'
theme = "lucario"
font-size = 18
```

## Reload automatically when the configuration file has changed.

Any file update in the configuration file will trigger a render operation in Rio terminal with the new configuration.

If you have any suggestion of configuration ideas to Rio, please feel free to [open an issue](https://github.com/raphamorim/rio/issues/new).

## width and height is not longer available in the configuration file.

The property width and height doesn't work anymore after the 0.0.7 release.

Before you could set the initial width and height of the screen, this configuration is not longer possible. It was removed due to compatibility with different targets.

## Stabilization of Linux x11 and Wayland

In older versions of Rio terminal it was possible to actually use Rio but it was slow and buggy for many users. Many fixes has been done in the v0.0.7 to bring the experience of Linux users close to what MacOS have right now.

The tweet below have a video demo that shows Rio running with X11:

<Tweet id="1670404595246419968" />

## Custom font rendering

Now you can actually use different installed fonts by changing the property "font".

The example below considers that you have "Fira code" font installed in your machine.

```toml
font = "Fira code"
```

A quick look of how font updates looks like in v0.0.7:

<Tweet id="1670808836213547009" />

## Sugarloaf ported to WebAssembly

The Rio support to WASM has moved one step ahead by introduction of Sugarloaf (Rio renderer) to WebAssembly.

In the image below you can see it in action:

## Changelog of v0.0.7

- Breaking changes for configuration file regarding "Style" property.
- Fix Background color not entirely set on vim [#88](https://github.com/raphamorim/rio/issues/88)
- Scroll now works for x11 and wayland.
- No longer renders to macos and x11 windows that are fully occluded / not directly visible.
- Introduced "window-opacity" config property for WebAssembly and Wayland builds.
- Add permissions instructions to Rio macos builds (Fix [#99](https://github.com/raphamorim/rio/issues/99)).
- Fixes for x11 and wayland rendering (Related: [#98](https://github.com/raphamorim/rio/issues/98) and [#100](https://github.com/raphamorim/rio/issues/100)).
- Performance fixes (Related: [#101](https://github.com/raphamorim/rio/issues/101)).
- Sugarloaf WebAssembly support.
- Fixed resize for all contexts: removed the glitch when resizing and switching between tabs.
- Fixed cursor inconsistencies [#95](https://github.com/raphamorim/rio/issues/95).
- Added command line interface support ("--help", "--version", "-e" and "--command").
- Added a fallback for WPGU request device operation: downlevel limits, which will allow the code to run on all possible hardware.
- Added "padding-x" to configuration.
- Reload automatically when the configuration file is changed ([#69](https://github.com/raphamorim/rio/issues/69)).
- Fix "Ctrl+D" and "exit" command not closing the app ([#87](https://github.com/raphamorim/rio/issues/87)).
- Changed default "light-black" color.

---
layout: post
title: 'Support to tabs, custom key bindings, performance improvements and other updates'
date: 2023-08-02 12:34
description: 'Support to tabs, custom key bindings, performance improvements, color automation for tabs, binary size optimization, bug fixes and other updates.'
categories: windows macos linux webassembly
---

import { Tweet } from 'react-tweet';

This post also includes changes from v0.0.13 and v0.0.12.

## Highlights

- Navigation.
- Custom key bindings.
- Performance updates.
- Color automation for tabs.

## Navigation

Rio allows to choose navigation between the following options:

- `CollapsedTab`
- `BottomTab`
- `TopTab`
- `Breadcrumb`

You can set the navigation mode in the configuration file. An example of configuration:

```toml
[navigation]
mode = "BottomTab"
use-current-path = true
clickable = false
```

Note: BottomTab and TopTab does not support click mode yet.

Below some images with different navigation setups:

- First image: BottomTab navigation x11.
- Second image: TopTav navigation wayland.

![example navigation x11](/assets/posts/0.0.15/demo-navigation-x11.png)

![second example navigation Wayland](/assets/posts/0.0.15/demo-navigation-wayland.png)

- First image: Breadcrumb navigation macos.
- Second image: BottomTab navigation macos.

![example breadcrumb navigation macos](/assets/features/demo-breadcrumb.png)

![example BottomTab navigation macos](/assets/features/demo-bottom-tab.png)

More info about navigation in [docs/navigation](/docs/navigation/).

## Sugarloaf performance improvements

Performance improvements with Sugarloaf de-duplication of input data.

Before an average sugarloaf.stack call would cost around "~253.5µs" and now it does cost an average of "~51.5µs".

The algorithm is effective whenever the terminal has a lot of repetition like empty spaces. The benchmark test below was made rendering 100.000 characters that had repetition:

![second example navigation Wayland](/assets/posts/0.0.15/de-duplication-sugarloaf.jpg)

The de-duplication algorithm in the renderer considerably reduced the initial rendering time from avg **6ms to 2ms** and using Rio in a large screen (>=136 columns) has dropped from **36ms to 8ms**.

Tweet reference:

<Tweet id="1685612156773961728" />

## Color automation for navigation

Rio allows to specify color overwrites for tabs based on program context.

The example below sets `#FFFF00` as color background whenever `nvim` is running.

![example navigation with color automation](/assets/features/demo-colorized-navigation.png)
![second example navigation with color automation](/assets/features/demo-colorized-navigation-2.png)

The configuration would be like:

```toml
[navigation]
color-automation = [
  { program = "nvim", color = "#FFFF00" }
]
```

## Custom key bindings

Now you can configure custom key bindings for Rio terminal.

Below an example of two custom key bindings:

```toml
[bindings]
keys = [
  { key = "q", with = "super", action = "Quit" }
  # Bytes[27, 91, 53, 126] is equivalent to "\x1b[5~"
  { key = "home", with = "super | shift", bytes = [27, 91, 53, 126] }
]
```

For example, in the first binding will tell Rio to close the terminal every time that "q" key is pressed along with "super" (in MacOS the "super" key is "command").

More info about custom key bindings in [docs/default-key-bindings](/docs/default-key-bindings/).

## Changelog of v0.0.15 along with 0.0.13 and 0.0.12

- Introduce configurable navigation with the following options: "CollapsedTab" (default), "Breadcrumb", "TopTab" and "BottomTab".
- Performance improvements with Sugarloaf de-duplication of input data. Before: "~253.5µs" and now "~51.5µs".
- Introduce "navigation.use-current-path" which sets if a tab/breacrumb should be open from the current context path.
- Fix rendering unicode with 1 width glyphs (fix [#160](https://github.com/raphamorim/rio/issues/160)).
- Increased max tabs from 9 to 72.
- Default colors "selection-foreground" and "selection-background" has changed.
- Default colors "tab" and "tab-active" has changed.
- Implementation of custom key bindings ([#117](https://github.com/raphamorim/rio/issues/117)).
- Fix .deb packing in GH Actions.
- Fix key binding for switch tab next (MacOS only).
- Fix scroll when copying text outside of offset.
- Fix copy key bindings.
- Fix Fuzzy Finder issue ([#132](https://github.com/raphamorim/rio/issues/132)).
- Introduce Copa (Alacritty's VTE forked version to introduce new sequences/instructions in next versions).
- Upgraded Winit to 0.29.0-beta.0
- Support for keybindings with dead keys.
- "Back"/"Forward" mouse buttons support in bindings.
- Fix unconditional query of xdg-portal settings on Wayland
- Fix "Maximized" startup mode not filling the screen properly on GNOME Wayland.
- Fix Default Vi key bindings for "Last"/"First" actions not working on X11/Wayland.
- Set "padding-x" to 0 for non-macos.
- Set "app_id"/"WM_CLASS" property on Wayland/X11.
- Strip binary is on for release builds.
- Each paste or key binding that has writing leads to clear selection and scroll bottom.
- Fixed over-rendering when scrolling.
- Fix selection.
- Support to copy using VIM.
- Fix for MacOS deadzone changing cursor to draggable on window buttons.
- Fix for scroll using tmux.

---
title: 'Features'
language: 'en'
---

Rio is available for Microsoft Windows, Linux distros, FreeBSD and Apple MacOS.

Short introduction of Rio terminal features.

- [Vi mode](/docs/features/vi-mode)
- [Hyperlinks](/docs/features/hyperlinks)
- [iTerm2 image protocol](/docs/features/iterm2-image-protocol)
- [Kitty keyboard protocol](/docs/features/kitty-keyboard-protocol)
- [Rio is fast](/docs/features/rio-is-fast)
- [Adaptive theme](/docs/features/adaptive-theme)
- [Multi windows](/docs/features/multi-windows)
- [Navigation](/docs/config/navigation)
- [Color automation for navigation](/docs/features/color-automation-for-navigation)
- [Sixel protocol](/docs/features/sixel-protocol)
- [Spawn or Fork](/docs/features/spawn-or-fork)

Note: Many other features are in development.

---
title: 'Spawn or Fork'
language: 'en'
---

In POSIX-based systems, Rio spawn processes instead of forking due to some compatibility issues between platforms.

However you can also switch from spawn to fork, forking a process is faster than spawning a process.

See how to configure it in the advanced section [here](/docs/config).

---
title: 'Navigation'
language: 'en'
---

Rio support different types of navigation modes:

<p>
<img alt="Demo TopTab" src="/rio/assets/features/demo-top-tab.png" width="48%"/>
<img alt="Demo native tabs" src="/rio/assets/posts/0.0.17/demo-native-tabs.png" width="48%"/>
</p>

<p>
<img alt="Demo Bookmark" src="https://miro.medium.com/v2/resize:fit:1400/format:webp/1*gMLWcZkniSHUT6Cb7L06Gg.png" width="48%" />
<img alt="Demo BottomTab" src="/rio/assets/features/demo-bottom-tab.png" width="48%"/>
</p>

See more about it [here](/docs/config/navigation).

---
title: 'Sixel protocol'
language: 'en'
---

Sixel, short for "six pixels", is a bitmap graphics format supported by terminals and printers from DEC. It consists of a pattern six pixels high and one wide, resulting in 64 possible patterns. Each possible pattern is assigned an ASCII character, making the sixels easy to transmit on 7-bit serial links.

> Demo of sixel on Rio.

![Demo sixel as gif](/assets/features/demo-sixel.gif)

Sixel was first introduced as a way of sending bitmap graphics to DEC dot matrix printers like the LA50. After being put into "sixel mode" the following data was interpreted to directly control six of the pins in the nine-pin print head. A string of sixel characters encodes a single 6-pixel high row of the image.

![Demo sixel with timg](/assets/features/demo-sixel-2.png)

The system was later re-used as a way to send bitmap data to the VT200 series and VT320 terminals when defining custom character sets. A series of sixels are used to transfer the bitmap for each character. This feature is known as soft character sets or dynamically redefinable character sets (DRCS). With the VT240, VT241, VT330, and VT340, the terminals could decode a complete sixel image to the screen, like those previously sent to printers.

![Demo sixel](/assets/features/demo-sixel.png)

For more information on Sixel: [en.wikipedia.org/wiki/Sixel](https://en.wikipedia.org/wiki/Sixel)

---
layout: post
title: 'Release 0.0.6'
date: 2023-06-07 10:34:14 +0200
categories: macos linux release
description: 'Tabs support, underline and beam cursor, text styles and many bug fixes.'
---

import { Tweet } from 'react-tweet';

# Tabs support, underline and beam cursor, text styles and many bug fixes.

Rio release 0.0.6 is finally here, there's a lot of updates to cover so let's get started. I also would like to remind you that Rio is still unstable. This release adds a lot of bug fixes and feature additions in order to make Rio terminal stable.

## Underline and strikethrough style

Support to text styling as such "Underline" and "Strikethrough".

## Tabs support

Tabs has been added to Rio terminal for macos and linux platform.

The shortcuts:

- Create tab: Logo key (Command in macos) + T.
- Close tab: Logo key (Command in macos) + W.
- Switch tab: Control key + Tab key.

Below you can see an example of usage:

<Tweet id="1664585160958922755" />

Note: There's a limit of maximum of 6 tabs for now.

## Support to Beam and Underline cursors

Beam and underline cursor support has been added to Rio terminal. Also, block cursor and IME state allow a character to be visible.

<Tweet id="1664146499398139906" />

## Changelog

- Fix: support to clipboard in linux by [@joseemds](https://github.com/joseemds).
- Font style for custom fonts by [@OlshaMB](https://github.com/OlshaMB) (closed [#80](https://github.com/raphamorim/rio/issues/80) and [#81](https://github.com/raphamorim/rio/issues/81))
- Text styles Underline and Strikethrough (closed [#79](https://github.com/raphamorim/rio/issues/79)).
- Update default colors for tabs/tabs-active.
- Tabs support.
- Fix rendering tab and hidden chars by replacing to space by [@niuez](https://github.com/niuez), (closed [#56](https://github.com/raphamorim/rio/issues/56)).
- Block cursor hover a character and still allow it to be visible.
- Support to caret Beam and Underline cursor [#67](https://github.com/raphamorim/rio/issues/67) by [@niuez](https://github.com/niuez).
- Fix panics if custom font is not found [#68](https://github.com/raphamorim/rio/issues/68).
- MacOs ignore alt key in cntrlseq (same behavior as Terminal.app, Hyper, iTerm and etecetera).

---
title: 'Terminfo'
language: 'en'
---

To make sure Rio works correctly, the "rio" terminfo must be used. The rio terminfo will be picked up automatically if it is installed.

If the following command returns without any errors, the rio terminfo is already installed:

```sh
infocmp rio
```

If it is not present already, you can install it globally with the following command:

When cloned locally, from the root of the repository run `sudo tic -xe rio misc/rio.terminfo`

If the source code has not been cloned locally:

```sh
curl -o rio.terminfo https://raw.githubusercontent.com/raphamorim/rio/main/misc/rio.terminfo
sudo tic -xe rio rio.terminfo
rm rio.terminfo
```

---
layout: post
title: 'Rio 0.0.9'
date: 2023-07-17 12:34
description: 'Multi window support, FreeBSD support, possibility of choose between Spawn and Fork, inverse instruction, multi byte rendering fixes, support to older MacOS versions, select shell and working-dir in configuration file.'
categories: release windows macos linux
---

# v0.0.9: Highlights

- Multi window support.
- Support to older MacOS versions (older or equal to Big Sur).
- Possibility of choose between Spawn and Fork.
- Inverse instruction.
- Introduced "rio" terminfo.
- Multi byte rendering fixes.
- Select shell and working-dir in configuration file.
- FreeBSD support and [added to FreeBSD ports](https://github.com/freebsd/freebsd-ports/commit/8582b8c59459a7dc5112a94a39de45f6cc124c3e).

<br/>

# Overview

Rio release 0.0.9 is finally here, there's a lot of updates to cover so let's get started. I also would like to remind you that Rio is still unstable. This release adds a lot of bug fixes and feature additions in order to make Rio terminal stable.

## Multi window support

Rio terminal has added the support to multi window feature.

<img src="https://miro.medium.com/v2/resize:fit:2914/format:webp/1*KyVD4EJ-wQU8pTmOFTwaQg.png" width="100%" />

Shortcuts:

- MacOS: Command + N
- Windows: Alt + N
- Linux: Control + N

## Support to older MacOS versions (older or equal to Big Sur).

Now it is available for versions below Ventura (Rome).

## Choose between fork or spawn

Decide which one to use in the configuration file by the usage of property "use-fork":

```toml
use-fork = false
```

Defaults for POSIX-based systems:

- MacOS: spawn processes
- Linux/BSD: fork processes

## Inverse instructions

Before the v0.0.9, programs were not correctly rendered in the terminal due to the lack of inverse instruction in Rio.

For example, ranger uses inverse to show the active path.

Before:

![Rio before 0.0.9](https://user-images.githubusercontent.com/3630346/245280487-12af56a1-d9ad-4310-bd4e-3d96d42e8a42.png)

Now:

![Rio now](https://user-images.githubusercontent.com/3630346/251224085-39531c88-8e04-48ad-b09a-8fd12a0e011b.png)

## Introduced "rio" terminfo

Terminfo (formerly Termcap) is a database of terminal capabilities and more. For every (well almost) model of terminal it tells application programs what the terminal is capable of doing.

It tells what escape sequences (or control characters) to send to the terminal in order to do things such as move the cursor to a new location, erase part of the screen, scroll the screen, change modes, change appearance (colors, brightness, blinking, underlining, reverse video etc.).

After about 1980, many terminals supported over a hundred different commands (some of which take numeric parameters).

Rio terminal has now it's own terminfo:

```sh
$ echo $TERM
rio
```

More info [here](/docs/install#terminfo).

## Fixes for multi-byte rendering

Before the v0.0.9, multi-byte characters were not correctly rendered in the terminal.

Before: As reported in the issue [#127](https://github.com/raphamorim/rio/issues/127)

Now:

![Now](https://user-images.githubusercontent.com/3630346/252940293-da345089-3d4c-4d05-8b0f-d0665f1e1347.png)

## Configuration file addition: shell

Now you can set "shell.program" to the path of your favorite shell, e.g. "/bin/fish".

Entries in "shell.args" are passed unmodified as arguments to the shell.

```toml
shell = { program = "/bin/zsh", args = ["--login"] }
```

Defaults:

- MacOS: /bin/zsh --login
- Linux/BSD: user login shell
- Windows: powershell

## Configuration file addition: working_dir

Directory the shell is started in. If this is unset the working
directory of the parent process will be used.

This configuration only has effect if use-fork is disabled

```toml
working_dir = "/Users/raphael/Documents/"
```

## Configuration file removal: Advanced

Breaking changes for configuration file regarding "Advanced". The configuration "Advanced" has moved to root level and "disable-render-when-unfocused" renamed to "disable-unfocused-render".

```toml
theme = "dracula"

[advanced]
disable-render-when-unfocused = true
```

**now**

```toml
theme = "dracula"
disable-unfocused-render = true
```

## Changelog of v0.0.9

- The configuration "Advanced" has moved to root level and "disable-render-when-unfocused" renamed to "disable-unfocused-render".
- Support to **spawn and fork processes**, spawn has became default. Spawn increases Rio compatibility in a broad range, like old MacOS versions (older or equal to Big Sur). However, If you want to use Rio terminal to fork processes instead of spawning processes, enable "use-fork" in the configuration file.
- Introduced "RIO_LOG_LEVEL" variable usage.
- Increased max tabs from 6 to 9.
- Fix Incorrect cursor position when using multi-byte characters (Ref: [#127](https://github.com/raphamorim/rio/issues/127))
- Fix bug ["black screen with nearly zero interactivity"](https://github.com/raphamorim/rio/issues/112) and new tab hanging.
- Fix cursor disappearing after resize.
- Introduction of "shell" and "working_dir" in configuration file.
- Multi window support [#97](https://github.com/raphamorim/rio/issues/97).
- Corrections on select and scroll experience (it was using wrongly font-bound for line calculation).
- Add selection color to the theme config (closed [#125](https://github.com/raphamorim/rio/issues/125)).
- Implemented Inverse (fix [#92](https://github.com/raphamorim/rio/issues/92)).
- Proper choose formats that matches Rio requirements (related [#122](https://github.com/raphamorim/rio/issues/122)).
- Corcovado: Filter windows crate dependency to only Windows targets (related: [#119](https://github.com/raphamorim/rio/issues/119)).
- Teletypewriter: Fixes for musl as target_env (related: [#119](https://github.com/raphamorim/rio/issues/119)).
- FreeBSD support, implementation by [yurivict](https://github.com/yurivict) ([Commit](https://github.com/freebsd/freebsd-ports/commit/8582b8c59459a7dc5112a94a39de45f6cc124c3e), Ref: [#115](https://github.com/raphamorim/rio/issues/115))

---
title: 'Command-line interface'
language: 'en'
---

A command-line interface (CLI) is a means of interacting with a device or computer program with commands from a user or client, and responses from the device or program, in the form of lines of text. Rio terminal has a command-line interface that you can use for different purposes.

```sh
$ rio --help
A hardware-accelerated GPU terminal emulator powered by WebGPU, focusing to run in desktops and browsers

Usage: rio [OPTIONS]

Options:
  -e, --command <COMMAND>...       Command and args to execute (must be last argument)
  -w, --working-dir <WORKING_DIR>  Start the shell in the specified working directory
      --write-config [<PATH>]      Writes the config to a given path or the default location
      --log-file                   Writes the logs to a file inside the config directory
  -h, --help                       Print help
  -V, --version                    Print version
```

The options "-e" and "--command" executes the command and closes the terminal right way after the execution.

```sh
$ rio -e sleep 10
```

You can also `RIO_LOG_LEVEL` environment variable for filter logs on-demand, for example:

```sh
$ RIO_LOG_LEVEL=debug rio -e echo 85
```

---
title: 'theme'
language: 'en'
---

The configuration property `theme` is used for specifying the theme. Rio will look in the `themes` folder for the theme.

You can see common paths for the `themes` directory here:

Note: Remember to replace "YOUR_USERNAME" with your actual user name.

| Platform | Path                                              |
| -------- | ------------------------------------------------- |
| Mac      | `/Users/YOUR_USERNAME/.config/rio/themes`         |
| Linux    | `/home/YOUR_USERNAME/.config/rio/themes`          |
| Windows  | `C:\Users\YOUR_USERNAME\AppData\Local\rio\themes` |

In the example below, we will setup the Dracula theme for Rio. The theme can be downloaded from [github.com/dracula/rio-terminal](https://github.com/dracula/rio-terminal).

After downloading the `dracula.toml` file, move it inside the folder `themes` in the configuration folder.

```toml
#  ~/.config/rio/config.toml
theme = "dracula"
```

It should look like this:

![Dracula theme example](/assets/posts/0.0.5/dracula-nvim.png)

Another option would be to install the [Lucario color scheme for Rio terminal](https://github.com/raphamorim/lucario/#rio-terminal), by moving the downloaded file to `~/.config/rio/themes/lucario.toml` and setting the `theme` property:

```toml
#  ~/.config/rio/config.toml
theme = "lucario"
```

![Lucario theme example](https://github.com/raphamorim/lucario/raw/main/images/rio.png)

You can find more than 250 themes for Rio terminal in this repository: [mbadolato/iTerm2-Color-Schemes/tree/master/rio](https://github.com/mbadolato/iTerm2-Color-Schemes/tree/master/rio).

## Building your own theme

Building your own theme for Rio is very straightforward.

Simply create a new theme file in your configuration themes folder (E.g. `~/.config/rio/themes/foobar.toml`) and choose your preferred colors:

Note: Missing fields will use the default for Rio.

```toml
# ~/.config/rio/themes/foobar.toml

[colors]
background = ""
foreground = ""

# Selection
selection-background = ""
selection-foreground = ""

# Navigation
tabs-active = ""
tabs-active-foreground = ""
tabs-active-highlight = ""
tabs-active-foreground = ""
bar = ""
cursor = ""
vi-cursor = ""

# Search
search-match-background = ""
search-match-foreground = ""
search-focused-match-background = ""
search-focused-match-foreground = ""

# Regular colors
black = ""
blue = ""
cyan = ""
green = ""
magenta = ""
red = ""
tabs = ""
white = ""
yellow = ""

# Dim colors
dim-black = ""
dim-blue = ""
dim-cyan = ""
dim-foreground = ""
dim-green = ""
dim-magenta = ""
dim-red = ""
dim-white = ""
dim-yellow = ""

# Light colors
light-black = ""
light-blue = ""
light-cyan = ""
light-foreground = ""
light-green = ""
light-magenta = ""
light-red = ""
light-white = ""
light-yellow = ""
```

After that all you have to do is set the `theme` property in your configuration file.

```toml
# ~/.config/rio/config.toml
theme = "foobar"
```

Proud of your new theme? Why not share it on the [Rio Discord](https://discord.gg/zRvJjmKGwS)!

---
title: 'Vi mode'
language: 'en'
---

Rio implements the Vi mode feature originally from [Alacritty](https://github.com/alacritty/alacritty/blob/master/docs/features.md#vi-mode).

By default you can launch Vi mode by using `alt` + `shift` + `space`.

![Demo Vi mode](/assets/features/demo-vi-mode.gif)

Below you can see the list of all default key bindings related to Vi mode. If you don't like of any specified key binding you can always turn off or modify (check [key bindings documentation section](/docs/default-key-bindings) for more information).

| Trigger                   | Action                     | Condition            |
| :------------------------ | :------------------------- | :------------------- |
| `alt` + `shift` + `space` | Toggle Vi Mode             | No restriction       |
| `i`                       | Toggle Vi Mode             | Vi mode is activated |
| `control` + `c`           | Toggle Vi Mode             | Vi mode is activated |
| `y`                       | `Copy` or `ClearSelection` | Vi mode is activated |
| `v`                       | Start normal selection     | Vi mode is activated |
| `v` + `shift`             | Start line selection       | Vi mode is activated |
| `v` + `control`           | Start block selection      | Vi mode is activated |
| `v` + `alt`               | Start semantic selection   | Vi mode is activated |
| `z`                       | Center around Vi cursor    | Vi mode is activated |
| `y` + `control`           | Scroll up 1 line           | Vi mode is activated |
| `e` + `control`           | Scroll down 1 line         | Vi mode is activated |
| `b` + `control`           | Scroll page up             | Vi mode is activated |
| `u` + `control`           | Scroll half page up        | Vi mode is activated |
| `d` + `control`           | Scroll half page down      | Vi mode is activated |
| `e` + `control`           | Scroll down 1 line         | Vi mode is activated |
| `k`                       | Move cursor up             | Vi mode is activated |
| `j`                       | Move cursor down           | Vi mode is activated |
| `h`                       | Move cursor left           | Vi mode is activated |
| `l`                       | Move cursor right          | Vi mode is activated |
| Arrow up                  | Move cursor up             | Vi mode is activated |
| Arrow down                | Move cursor down           | Vi mode is activated |
| Arrow left                | Move cursor left           | Vi mode is activated |
| Arrow right               | Move cursor right          | Vi mode is activated |
| Arrow right               | Move cursor right          | Vi mode is activated |
| `0`                       | Move first                 | Vi mode is activated |
| `4` + `shift`             | Move last                  | Vi mode is activated |
| `6` + `shift`             | Move first occupied        | Vi mode is activated |
| `h` + `shift`             | Move high                  | Vi mode is activated |
| `m` + `shift`             | Move middle                | Vi mode is activated |
| `l` + `shift`             | Move low                   | Vi mode is activated |
| `b`                       | Move semantic left         | Vi mode is activated |
| `w`                       | Move semantic right        | Vi mode is activated |
| `e`                       | Move semantic right end    | Vi mode is activated |
| `b` + `shift`             | Move word left             | Vi mode is activated |
| `w` + `shift`             | Move word right            | Vi mode is activated |
| `e` + `shift`             | Move word right end        | Vi mode is activated |
| `5`                       | Move by bracket rule       | Vi mode is activated |

---
title: 'Linux'
language: 'en'
---

Installation options:

- Alpine Linux package: [pkgs.alpinelinux.org/packages?name=rio](https://pkgs.alpinelinux.org/packages?name=rio)
- Arch Linux package: [archlinux.org/packages/extra/x86_64/rio](https://archlinux.org/packages/extra/x86_64/rio)
- Nix package: [NixOS/nixpkgs/rio](https://github.com/NixOS/nixpkgs/blob/nixos-unstable/pkgs/applications/terminal-emulators/rio/default.nix)
- openSUSE package: [openSUSE:Factory/rioterm](https://software.opensuse.org/package/rioterm)
- Void Linux package: https://github.com/void-linux/void-packages/tree/master/srcpkgs/rio

In case your distro doesn't have the package manager option listed above, proceed to [build from the source](/docs/install/build-from-source).

---
title: 'Hyperlinks'
language: 'en'
---

Rio terminal supports opening hyperlinks from the terminal.

## MacOS

To activate hyperlink feature hold `Command` key when hovering a link:

![Demo macos hyperlink](/assets/features/demo-hyperlink-macos.gif)

## Windows / Linux / BSD

To activate hyperlink feature hold `alt` key when hovering a link:

![Demo windows hyperlink](/assets/features/demo-hyperlink-windows.png)

![Demo linux hyperlink](/assets/features/demo-hyperlink-linux.png)

## OSC 8

Rio terminal support OSC 8 for defining hyperlinks.

```bash
OSC 8 ; [params] ; [url] ST
```

The `[params]` consists of zero or more colon-delimited key-value pairs. A key-value pair is formatted as `key=value`. The only currently defined key is id.

If the url is absent then that ends the hyperlink. Typical usage would look like:

```bash
OSC 8 ; ; https://example.com/ ST Link to example website OSC 8 ; ; ST
```

Will work as default rio terminal hyperlinks, by holding command or `alt` key (depending of your platform) and clicking the link.

### OSC 8 Example

```bash
printf '\e]8;;https://raphamorim.io/rio/\e\\This is a link\e]8;;\e\\\n'
```

![Demo hyperlink using OSC 8](/assets/features/demo-hyperlink-osc-8.png)

---
title: 'developer'
language: 'en'
---

This property enables log level filter and file. The default level is "OFF" and the logs are not logged to a file as default.

```toml
[developer]
log-level = "OFF"
enable-log-file = false
```

If you have any suggestion of configuration ideas to Rio, please feel free to [open an issue](https://github.com/raphamorim/rio/issues/new).

---
title: 'Rio is Fast'
language: 'en'
---

Rio is perceived fast, there's few reasons behind the speed. First reason is that Rio is built in Rust ("Speed of Rust vs C" [kornel.ski/rust-c-speed](https://kornel.ski/rust-c-speed)). The terminal is also built over ANSI handler and parser is built from Alacritty terminal's VTE [github.com/alacritty/vte](https://github.com/alacritty/vte/).

The renderer called Sugarloaf has a "sugar" architecture created for minimal and quick interactions in render steps using WebGPU with performance at highest.

<img src="https://miro.medium.com/v2/resize:fit:1400/1*1enyoIVZivAcHY_kfYXUvQ.gif" width="100%" />

---
title: 'navigation'
language: 'en'
---

- `hide-if-single` - Hide navigation UI if there is only one tab. It does not work for `NativeTab`. 
- `clickable` - Enable click on tabs to switch.
- `use-current-path` - Use same path whenever a new tab is created (Note: requires [`use-fork`](/docs/config/use-fork) to be set to false).
- `color-automation` - Set a specific color for the tab whenever a specific program is running, or in a specific directory.

```toml
[navigation]
mode = "Bookmark"
clickable = false
hide-if-single = true
use-current-path = false
color-automation = []
```

Rio has multiple styles of showing navigation/tabs.

### Bookmark

Note: The example below is using the [Dracula](https://github.com/dracula/rio-terminal) color scheme instead of Rio default colors.

`Bookmark` is the default navigation mode.

<img src="https://miro.medium.com/v2/resize:fit:1400/format:webp/1*gMLWcZkniSHUT6Cb7L06Gg.png" width="60%" />

Usage:

```toml
[navigation]
mode = "Bookmark"
```

### NativeTab (MacOS only)

<img alt="Demo NativeTab" src="/rio/assets/posts/0.0.17/demo-native-tabs.png" width="60%"/>

Usage:

```toml
[navigation]
mode = "NativeTab"
```

### BottomTab

Note: `BottomTab` does not support click mode yet.

<img alt="Demo BottomTab" src="/rio/assets/features/demo-bottom-tab.png" width="58%"/>

Usage:

```toml
[colors]
tabs = "#000000"

[navigation]
mode = "BottomTab"
```

### TopTab

Note: `TopTab` does not support click mode yet.

<img alt="Demo TopTab" src="/rio/assets/features/demo-top-tab.png" width="70%"/>

Usage:

```toml
[colors]
tabs = "#000000"

[navigation]
mode = "TopTab"
```

### Plain

Plain navigation mode will simply turn off any tab key binding.

This mode is perfect if you use Rio terminal with tmux or zellij.

Usage:

```toml
[navigation]
mode = "Plain"
```

### Hide if is only one tab

The property `hide-if-single` hides navigation UI if there is only one tab. It does not work for `NativeTab`.

Default is `true`.

```toml
[navigation]
hide-if-single = true
```

### Color automation for navigation

Rio supports specifying the color of tabs using the `program` and `path` options.

Note: `path` is only available for MacOS, BSD and Linux.

```toml
[navigation]
color-automation = [
  # Set tab to red (#FF0000) when NeoVim is open.
  { program = "nvim", color = "#FF0000" },
  # Set tab to green  (#00FF00) when in the projects folder
  { path = "/home/YOUR_USERNAME/projects", color = "#00FF00" },
    # Set tab to blue (#0000FF) when in the Rio folder AND vim is open
  { program = "vim", path = "/home/YOUR_USERNAME/projects/rio", color = "#0000FF" },
]
```

#### Program

The example below sets `#FFFF00` as color background whenever `nvim` is running.

<p>
<img alt="example navigation with program color automation using BottomTab" src="/rio/assets/features/demo-colorized-navigation.png" width="48%"/>

<img alt="example navigation with program color automation using Bookmark" src="/rio/assets/features/demo-colorized-navigation-2.png" width="48%"/>
</p>

The configuration would be like:

```toml
[navigation]
color-automation = [
  { program = "nvim", color = "#FFFF00" }
]
```

#### Path

The example below sets `#FFFF00` as color background when in the `/home/geg/.config/rio` path.

Note: `path` is only available for MacOS, BSD and Linux.

The configuration would be like:

```toml
[navigation]
color-automation = [
  { path = "/home/geg/.config/rio", color = "#FFFF00" }
]
```

<p>
<img alt="example navigation with path color automation using TopTab" src="/rio/assets/features/demo-colorized-navigation-path-1.png" width="48%"/>

<img alt="example navigation with path color automation using Bookmark" src="/rio/assets/features/demo-colorized-navigation-path-2.png" width="48%"/>
</p>

#### Program and path

It is possible to use both `path` and `program` at the same time.

The example below sets `#FFFF00` as color background when in the `/home` path and `nvim` is open.

Note: `path` is only available for MacOS, BSD and Linux.

The configuration would be like:

```toml
[navigation]
color-automation = [
  { program = "nvim", path = "/home", color = "#FFFF00" }
]
```

<p>
<img alt="example navigation with program and path color automation using TopTab" src="/rio/assets/features/demo-colorized-navigation-program-and-path-1.png" width="48%"/>

<img alt="example navigation with program and path color automation using Bookmark" src="/rio/assets/features/demo-colorized-navigation-program-and-path-2.png" width="48%"/>
</p>

---
title: 'Plugins'
language: 'en'
---

## Note: Plugins are not ready yet

Plugins in Rio terminal are powered by WebAssembly. Which means they can be written in any programming language, as long as it is able to be compiled to WebAssembly.

Rio provides hooks and controlling functions to WASM modules that are loaded in initialization time. Let's take a look in a plugin written with JavaScript or Rust.

```rust
#[link(wasm_import_module = "Rio")]
extern "C" {
    fn render() -> bool;
}

#[export_name = "render"]
pub fn render() {

}
```

---
layout: post
title: 'Settings UI, Welcome UI, Adaptive theme, crates.io and more'
date: 2023-09-19 13:53
description: 'Crates.io, Alpine Linux, openSUSE, Nix home manager, adaptive theme, settings UI, welcome UI, blinking cursor, plain navigation mode ...'
categories: windows macos linux webassembly
---

This post also includes changes of v0.0.19.

### Highlight #1: New window API (breaking change)

Configuration properties: window_height, window_width and window_opacity has been moved to a new window/background API:

```toml
# Window configuration
#
# • width - define the initial window width.
#   Default: 600
#
# • height - define the initial window height.
#   Default: 400
#
# • mode - define how the window will be created
#     - "Windowed" (default) is based on width and height
#     - "Maximized" window is created with maximized
#     - "Fullscreen" window is created with fullscreen
#
[window]
width = 600
height = 400
mode = "Windowed"
```

From now and on you can also set image as background.

```toml
# Background configuration
#
# • opacity - changes the background transparency state
#   Default: 1.0
#
# • mode - defines background mode between "Color" and "Image"
#   Default: Color
#
# • image - Set an image as background
#   Default: None
#
[background]
mode = "Image"
opacity = 1.0
[background.image]
path = "/Users/rapha/Desktop/eastward.jpg"
width = 200.0
height = 200.0
x = 0.0
```

Although you can define opacity though the new configuration API it does not work either for the image neither for the window. This something will be fixed in upcoming releases. An example below using the configuration we just saw.

![Example background](https://media.discordapp.net/attachments/1110689410258841662/1152881776348364811/F6JGpamXsAAWvcG.png?width=1048&height=848)

note: The image above is from the [Eastward game](<https://en.wikipedia.org/wiki/Eastward_(video_game)>)

## Highlight #2: crates.io

Rio terminal is now also available in crates.io: https://crates.io/crates/rioterm .

```sh
cargo install rioterm
```

## Highlight #3: Alpine Linux, openSUSE and Nix home manager.

• Rio has landed in Nix home manager: [nix-community/home-manager/pull/4118](https://github.com/nix-community/home-manager/pull/4118) and huge props to [@otaviosalvador](https://twitter.com/otaviosalvador) for making it possible.

• Rio has landed in Alpine Linux: [gitlab.alpinelinux.org](https://gitlab.alpinelinux.org/alpine/aports/-/merge_requests/51369) and huge props to [@androw](https://github.com/androw) for making it possible.

• Rio has landed in openSUSE: [build.opensuse.org](https://build.opensuse.org/package/show/X11:terminals/rioterm) and huge props to [@androw](https://github.com/androw) for making it possible.

## Highlight #4: Welcome screen and default creation of configuration file

Now Rio terminal does create the configuration file per first initialization.

![Welcome message](https://user-images.githubusercontent.com/240594/268500927-788ea316-ede8-4d23-ba1f-9621a6a82908.png)

## Highlight #5: Settings UI

After v0.0.19, Rio will start to provide a custom internal settings UI, where you can manage your configuration without edit the file.

![Settings UI](/assets/posts/0.0.19/demo-settings.png)

Note: Not all the configuration options has been ported, so you would still need to use the config file for specific configuration updates.

## Highlight #6: Error handling

Configuration errors, command errors, font not found and etecetera are not silent anymore. This version introduces an UI for error handling.

![Error handling](/assets/posts/0.0.19/demo-error-handling.png)

## Highlight #7: Adaptive theme

Rio support theme based on the system theme (light and dark). This configuration only works for Web, MacOS and Windows.

```toml
[adaptive-theme]
light = "belafonte-day"
dark = "belafonte-night"
```

Example of usage with MacOS:

![Settings UI](/assets/features/adaptive-theme.gif)

## Highlight #8: Blinking cursor

Rio now supports blinking cursor.

```toml
# Blinking Cursor
#
# Default is false
#
blinking-cursor = true
```

## Highlight #9: Navigation mode called "Plain"

Navigation mode "Plain" is perfect if you use tmux/zellij or anything similar.

Plain navigation mode will simply turn off any platform key binding, will not add any key binding (like creation of tabs, windows, panels and other platform key bindings). Note it also get rid of platform key bingins such as shift+control+c for example, to re-add it you would need to map yourself.

This mode is great if you want configure every single bit of how Rio key bindings will work.

Usage:

```toml
[navigation]
mode = "Plain"
```

## Full changelog of v0.0.19

- Fix for retrieving shell environment variable when running inside of Flatpak sandbox (Ref: https://github.com/raphamorim/rio/issues/198).
- Rio terminal is now also available in crates.io: https://crates.io/crates/rioterm .
- Added "navigation.mode = "Plain"", it basically disables all platform key bindings for tabs, windows and panels creation (Ref https://github.com/raphamorim/rio/issues/213).
- Support for blinking cursor (Ref: https://github.com/raphamorim/rio/issues/137) (this option is not enabled by default).
- Migrated font-kit to a custom font loader.
- Support to MacOS tile window positioning feature (left or right).
- Added support to MacOS display native top bar items.
- Support to adaptive theme (theme selection based on user system theme variant "dark" or "light").
- Implemented "ScrollPageUp", "ScrollPageDown", "ScrollHalfPageUp", "ScrollHalfPageDown", "ScrollToTop", "ScrollToBottom", "ScrollLineUp", "ScrollLineDown" (Ref: https://github.com/raphamorim/rio/issues/206).
- Support to "fonts.family" (it overwrites regular, bold, bold-italic and italic font families).
- Added a welcome screen UI.
- Added a settings UI.
- Exposes "RIO_CONFIG" environment variable that contains the path of the configuration.
- Rio creates a configuration file with all defaults if does not exist.
- Added "OpenConfigEditor" key binding for all platforms.
- Configuration property "editor" was removed.
- Created Assistant, Rio terminal UI for display error (Ref: https://github.com/raphamorim/rio/issues/168).
- Fix 'Backspace' keypress triggers Ctrl+h keybinding in Zellij instead of deleting character. (Ref: https://github.com/raphamorim/rio/issues/197).
- Implemented "TERM_PROGRAM" and "TERM_PROGRAM_VERSION" (Ref: https://github.com/raphamorim/rio/issues/200).
- Whenever native tabs is on disable macos deadzone logic.

---
title: 'Install'
language: 'en'
---

Rio is available for:

- [MacOS](/docs/install/macos)
- [Linux](/docs/install/linux)
- [Windows](/docs/install/windows)
- [FreeBSD](/docs/install/freebsd)
- [WebAssembly](/docs/install/webassembly)

Rio has two type of builds: stable and nightly.

While stable versions are thoroughly tested and takes longer to release, nightly versions are created on a daily basis.

[Github releases for Nightly](https://github.com/raphamorim/rio/releases/tag/nightly)

---
title: 'shell'
language: 'en'
---

You can set `shell.program` to the path of your favorite shell, e.g. `/bin/fish`.

Entries in `shell.args` are passed unmodified as arguments to the shell.

Default:

- (macOS) user login shell
- (Linux/BSD) user login shell
- (Windows) powershell

### Shell Examples

1. MacOS using fish shell from bin path:

```toml
[shell]
program = "/bin/fish"
args = ["--login"]
```

2. Windows using powershell:

```toml
[shell]
program = "pwsh"
args = []
```

3. Windows using powershell with login:

```toml
[shell]
program = "pwsh"
args = ["-l"]
```

4. MacOS with tmux installed by homebrew:

```toml
[shell]
program = "/opt/homebrew/bin/tmux"
args = ["new-session", "-c", "/var/www"]
```

---
title: 'window'
language: 'en'
---

- `width` - define the initial window width.

  - Default: `600`

- `height` - define the initial window height.

  - Default: `400`

- `mode` - define how the window will be created

  - `Windowed` (default) is based on width and height
  - `Maximized` window is created with maximized
  - `Fullscreen` window is created with fullscreen

- `opacity` Set window background opacity.

  - Default: `1.0`.

- `blur` Set blur on the window background. Changing this config requires restarting Rio to take effect.

  - Default: `false`.

- `background-image` Set an image as background.

  - Default: `None`

- `decorations` - Set window decorations
  - `Enabled` (default for Windows/Linux/BSD) enable window decorations.
  - `Disabled` disable all window decorations.
  - `Transparent` (default for MacOS) window decorations with transparency.
  - `Buttonless` remove buttons from window decorations.

Example:

```toml
[window]
width = 600
height = 400
mode = "Windowed"
opacity = 1.0
blur = false
decorations = "Enabled"
```

### Using blur and background opacity:

```toml
[window]
opacity = 0.5
decorations = "enabled"
blur = true
```

![Demo blur and background opacity](/assets/demos/demo-macos-blur.png)

![Demo blur and background opacity 2](/assets/demos/demos-nixos-blur.png)

### Using image as background:

If both properties `width` and `height` are occluded then background image will use the terminal width/height.

```toml
[window.background-image]
path = "/Users/hugoamor/Desktop/musashi.png"
opacity = 0.5
x = 0.0
y = -100.0
```

![Demo image as background](/assets/demos/demo-background-image.png)

If any property `width` or `height` are used then background image will be respected.

```toml
[window.background-image]
path = "/Users/hugoamor/Desktop/harvest-moon.png"
width = 1200
height = 800
opacity = 0.5
x = 0.0
y = 0.0
```

![Demo image as background](/assets/demos/demo-background-image-partial.png)

---
title: 'Color automation for navigation'
language: 'en'
---

Rio allows specifying color for tabs based on program and path contexts, using the `program` and `path` options.

It is possible to combine `program` and `path`.

Note: `path` is only available for MacOS, BSD and Linux.

#### Program

The example below sets `#FFFF00` as color background whenever `nvim` is running.

<p>
<img alt="example navigation with program color automation using TopTab" src="/rio/assets/features/demo-colorized-navigation.png" width="48%"/>

<img alt="example navigation with program color automation using CollapsedTab" src="/rio/assets/features/demo-colorized-navigation-2.png" width="48%"/>
</p>

The configuration would be like:

```toml
[navigation]
color-automation = [
  { program = "nvim", color = "#FFFF00" }
]
```

#### Path

The example below sets `#FFFF00` as color background when in the `/home/geg/.config/rio` path.

Note: `path` is only available for MacOS, BSD and Linux.

The configuration would be like:

```toml
[navigation]
color-automation = [
  { path = "/home/geg/.config/rio", color = "#FFFF00" }
]
```

<p>
<img alt="example navigation with path color automation using TopTab" src="/rio/assets/features/demo-colorized-navigation-path-1.png" width="48%"/>

<img alt="example navigation with path color automation using CollapsedTab" src="/rio/assets/features/demo-colorized-navigation-path-2.png" width="48%"/>
</p>

#### Program and path

It is possible to combine `path` and `program`.

The example below sets `#FFFF00` as color background when in the `/home` path and `nvim` is open.

Note: `path` is only available for MacOS, BSD and Linux.

The configuration would be like:

```toml
[navigation]
color-automation = [
  { program = "nvim", path = "/home", color = "#FFFF00" }
]
```

<p>
<img alt="example navigation with program and path color automation using TopTab" src="/rio/assets/features/demo-colorized-navigation-program-and-path-1.png" width="48%"/>

<img alt="example navigation with program and path color automation using CollapsedTab" src="/rio/assets/features/demo-colorized-navigation-program-and-path-2.png" width="48%"/>
</p>

---
title: 'Windows'
language: 'en'
---

Note: Rio is only available for Windows 10 or later.

Prebuilt binaries for Windows:

- [Download Microsoft installer](https://github.com/raphamorim/rio/releases/download/v0.1.14/Rio-installer.msi)
- [Download Microsoft executable](https://github.com/raphamorim/rio/releases/download/v0.1.14/Rio-portable.exe)
- [Using Chocolatey package manager](https://community.chocolatey.org/packages/rio-terminal)

```sh
choco install rio-terminal
```

There's a few things to note about the installer and the portable version:

- The browser will ask if you want to keep the file, click "Keep" to save the installer/executable on your computer.
- When opening the file, Windows will give you a warning, click "More info" and then "Run anyway" to run the installer/executable.

If you want to change the default shell to the new PowerShell platform, change the following line in your config file (see [Configuration file](/docs/config) for more information):

```toml
shell = { program = "pwsh", args = ["--login"] }
```

You may want to use a specific GPU on your system, specially if you're on a laptop configuration, this can enable hardware acceleration and improve performance of the application.
To make Windows utilize a GPU for a specific application through Windows display settings, follow the instructions:

1. Simultaneously press the Windows key and the letter "i" on your keyboard to open Windows Settings.
2. Select System.
3. Choose the Display option.
4. Click on the Graphics setting link located at the bottom of the page.
5. Select the application from the list or press the Browse button, then select the executable file for the application.
6. Click on the Options button to display the GPU selection window.
7. Choose the GPU you want to prioritize for the selected application.
8. Click on the Save button.

---
title: 'keyboard'
language: 'en'
---

- `use-kitty-keyboard-protocol` - Enable Kitty Keyboard protocol

- `disable-ctlseqs-alt` - Disable ctlseqs with ALT keys
  - Useful for example if you would like Rio to replicate Terminal.app, since it does not deal with ctlseqs with ALT keys

Example:

```toml
[keyboard]
use-kitty-keyboard-protocol = false
disable-ctlseqs-alt = false
```

---
title: 'bindings'
language: 'en'
---

You can see the default [default key bindings](/docs/default-key-bindings)

Rio allows you to add new keybindings and overwrite any default key bindings.

Keybinds are built using the following trigger fields:

| Name          | Description     |
| ------------- | --------------- |
| [key](#key)   | The key pressed |
| [with](#with) | Modifier keys   |
| [mode](#mode) | Terminal mode   |

Whom can be be combined with the following effect fields:

| Name              | Description            |
| ----------------- | ---------------------- |
| [action](#action) | Predefined Rio actions |
| [bytes](#bytes)   | Write byte sequence    |
| [text](#text)     | Write text sequence    |

```toml
[bindings]
keys = [
  { key = "q", with = "super", action = "Quit" },
  # Bytes[27, 91, 53, 126] is equivalent to "\x1b[5~"
  { key = "home", with = "super | shift", bytes = [27, 91, 53, 126] },
  # Remove existing keybind
  { key = "v", with = "control | shift", action = "none" },
]
```

### [Key](#key)

Each value in key binding will specify an identifier of the key pressed:

- `a-z`
- `0-9`
- `F1-F24`
- `tab` `esc`
- `home` `space` `delete` `insert` `pageup` `pagedown` `end` `back`
- `up` `down` `left` `right`
- `@` `colon` `.` `return` `[` `]` `;` `\\` `+` `,` `/` `=` `-` `*`
- `numpadenter` `numpadadd` `numpadcomma` `numpaddivide` `numpadequals` `numpadsubtract` `numpadmultiply`
- `numpad1` `numpad2` `numpad3` `numpad4` `numpad5` `numpad6` `numpad7` `numpad8` `numpad9` `numpad0`

### [Action](#action)

Execute a predefined action in Rio terminal.

#### [Basic Actions](#basic-actions)

| Action           | Description                                                                   |
| :--------------- | :---------------------------------------------------------------------------- |
| None             |                                                                               |
| ReceiveChar      |                                                                               |
| ToggleVIMode     |                                                                               |
| Paste            | Paste command                                                                 |
| Copy             |                                                                               |
| OpenConfigEditor |                                                                               |
| ResetFontSize    |                                                                               |
| IncreaseFontSize |                                                                               |
| DecreaseFontSize |                                                                               |
| Run(string)      | Example: Running command `Run(code)` or `Run(code ~/.config/rio/config.toml)` |
| PasteSelection   |                                                                               |
| ClearSelection   |                                                                               |

#### [Window Actions](#window-actions)

| Action       | Description |
| :----------- | :---------- |
| CreateWindow |             |
| Quit         |             |

#### [Pane Actions](#pane-actions)

| Action            | Description |
| :---------------- | :---------- |
| SplitHorizontally |             |
| SplitVertically   |             |
| ClosePane         |             |

#### [Tab Actions](#tab-actions)

| Action               | Description                                                         |
| :------------------- | :------------------------------------------------------------------ |
| CreateTab            |                                                                     |
| CloseTab             |                                                                     |
| CloseUnfocusedTabs   |                                                                     |
| SelectPrevTab        |                                                                     |
| SelectNextTab        |                                                                     |
| SelectLastTab        |                                                                     |
| SelectTab(tab_index) | Example: Select first tab `SelectTab(0)`, second tab `SelectTab(1)` |

#### [Scroll Actions](#scroll-actions)

| Action             | Description                                                                |
| :----------------- | :------------------------------------------------------------------------- |
| Scroll(int)        | Example: Scroll up 8 lines `Scroll(8)` or scroll down 5 lines `Scroll(-5)` |
| ScrollPageUp       |                                                                            |
| ScrollPageDown     |                                                                            |
| ScrollHalfPageUp   |                                                                            |
| ScrollHalfPageDown |                                                                            |
| ScrollToTop        |                                                                            |
| ScrollToBottom     |                                                                            |

### [Search](#search)

| Action             | Description                                                                |
| :----------------- | :------------------------------------------------------------------------- |
| SearchForward      |  |
| SearchBackward       |                                                                            |
| SearchConfirm     |                                                                            |
| SearchClear   |                                                                            |
| SearchFocusNext |                                                                            |
| SearchFocusPrevious        |                                                                            |
| SearchDeleteWord     |                                                                            |
| SearchHistoryNext     |                                                                            |
| SearchHistoryPrevious     |                                                                            |

### [Bytes](#bytes)

Send a byte sequence to the running application.

The `bytes` field writes the specified string to the terminal. This makes
it possible to pass escape sequences, like `PageUp` ("\x1b[5~"). Note that applications use terminfo to map escape sequences back
to keys. It is therefore required to update the terminfo when changing an escape sequence.

### [With](#with)

Key modifiers to filter binding actions

- `none`
- `control`
- `option`
- `super`
- `shift`
- `alt`

Multiple modifiers can be combined using `|` like this:

```toml
with = "control | shift"
```

<!--
 - `mode`: Indicate a binding for only specific terminal reported modes
    This is mainly used to send applications the correct escape sequences
    when in different modes.
    - AppCursor
    - AppKeypad
    - Alt
    A `~` operator can be used before a mode to apply the binding whenever
    the mode is *not* active, e.g. `~Alt`. -->

### [Mode](#mode)

There is currently four different modes:

- `vi`
- `alt` (Alt screen)
- `appcursor`
- `appkeypad`

`~` can be prefixed to disable the keybind while in that mode.

```toml
[bindings]
keys = [
  # Enable VI mode on escape, when not in VI mode.
  { key = "esc", mode = "~vi", action = "ToggleVIMode" },
]
```

### [Text](#text)

`text` can be used to write specific text on key press:

```toml
[bindings]
keys = [
  # Write `Rio is awesome!` on `Control + r`
  { key = "r", with = "control", text = "Rio is awesome!" },
]
```

### [Overwriting](#overwriting)

Bindings are always filled by default, but will be replaced when a new binding with the same triggers is defined. To unset a default binding, it can be mapped to the `ReceiveChar` action. Alternatively, you can use `None` for a no-op if you do not wish to receive input characters for that binding.

The example below will disable window creation binding in the macos:

```toml
[bindings]
keys = [
  { key = "n", with = "super", action = "ReceiveChar" }
]
```

`ReceiveChar` will treat the binding as non existent and simply receive the input and put the character into the terminal.

Optionally you can ignore/disable completely a binding using `None`. In the example below, whenever you use key "n" along with "super" key nothing will happen.

```toml
[bindings]
keys = [
  { key = "n", with = "super", action = "None" }
]
```

If you are missing a key binding that you believe that should be a default in the platform that you are using, feel free to [open an issue](https://github.com/raphamorim/rio).

---
title: 'Multi windows'
language: 'en'
---

The terminal supports multi window features in the following platforms: Windows, MacOS, FreeBSD and Linux.

<img src="https://miro.medium.com/v2/resize:fit:2914/format:webp/1*KyVD4EJ-wQU8pTmOFTwaQg.png" width="100%" />

---
layout: post
title: 'Rio 0.0.11'
date: 2023-07-19 12:34
description: 'Performance improvements around 80-110%, binary size optimization, usage of pixel scale and others.'
categories: release windows macos linux
---

# v0.0.11: Changelog

Example of Rio v0.0.11 using the new sugarloaf:

v0.0.11 includes v0.0.10 with a bug fix of CachedSugar for font styles.

- Major rewrote of sugarloaf.
- Sugarloaf: Performance improvements around 80-110%.
- Sugarloaf: Introduced CachedSugar.
- Sugarloaf: Usage of PixelScale.
- Sugarloaf: Line-height support.
- Open new tab using the current tab directory.
- Fix some symbols break the horizontal and vertical alignment of lines (ref [#148](https://github.com/raphamorim/rio/issues/148)).
- Fix font size configuration is confusing (ref [#139](https://github.com/raphamorim/rio/issues/139)).
- Fix Glyph not rendered in prompt (ref: [#135](https://github.com/raphamorim/rio/issues/135)).
- Use fork by default in test context.
- Updated terminfo.
- Increased default font size to 18.
- Move to next and prev tab using keybindings.
- Setting editor by keybindings and new property called "editor" in configuration file.
- Rio creates ".deb" packages (canary and release).
- Binary size optimization (ref: [#152](https://github.com/raphamorim/rio/pull/152)) by [@OlshaMB](https://github.com/OlshaMB)
- A huge thanks to [@ralgozino](https://github.com/ralgozino) for all bug reporting and testing through this version development.

---
layout: post
title: 'Release 0.0.5'
date: 2023-05-31 10:34:14 +0200
categories: macos linux release
description: 'Arrival to Linux, themes support and many bug fixes.'
---

# Arrival to Linux, Themes support and many bug fixes.

Rio release 0.0.5 is finally here, there's a lot of updates to cover so let's get started. I also would like to remind you that Rio is still unstable. This release adds a lot of bug fixes and feature additions in order to make Rio terminal stable.

## New configuration path

Configuration path has changed from `{$HOME}/.rio/` to `{$HOME}/.config/rio`.

Changes were applied for macOS and Linux.

## Linux Support

Rio is now available to Linux, build information was added in the [install](/docs/install) page.

![Linux support](/assets/posts/0.0.5/linux.jpeg)

Linux builds are considered less stable than macOs, due to macOs features availability comparison.

## Themes Support

A new property was added to Rio configuration file called "theme". You can set the theme that you want to use and Rio will look in the folder "themes" in the configuration path.

![Themes support](/assets/posts/0.0.5/themes.png)

Dracula theme example:

![Example dracula](/assets/posts/0.0.5/dracula-nvim.png)

## Bold and italic style

Support to text styling as such "bold" and "italic".

![Themes support](/assets/posts/0.0.5/font-macos.png)

## Changelog

- Fix to render specific 24bit colors [#66](https://github.com/raphamorim/rio/issues/#66) by [@niuez](https://github.com/niuez).
- Cross build for arm64 and x86
- Bold and Italic support [#33](https://github.com/raphamorim/rio/issues/#33).
- Add RioEvent::ColorRequest events to write color updates on pty.
- Theme support [#42](https://github.com/raphamorim/rio/issues/42).
- Fix font-size dependency for serialization
- Fix cursor visibility on VI mode and scroll [#51](https://github.com/raphamorim/#51)
- Performance fixes for rendering from teletypewriter updates.
- Fix scale issues for 1.0 scale factor or using monitor with different scale factor. [#50](https://github.com/raphamorim/rio/issues/#50)
- Improved release process to only contain Rio.app file. [#54](https://github.com/raphamorim/rio/issues/#54)

---
title: 'cursor'
language: 'en'
---

## Cursor

By default, the cursor shape is set to `block`. You can also choose from other options like `underline` and `beam`.

Additionally, you can enable or disable cursor blinking, which is set to `false` by default.

### Shape

Options: 'block', 'underline', 'beam'

```toml
[cursor]
shape = 'block'
```

### Blinking

Enable/disable blinking (default: false)

```toml
[cursor]
blinking = false
```

### Blinking-interval

Set cursor blinking interval (default: 800, only configurable from 350ms to 1200ms).

```toml
[cursor]
blinking-interval = 800
```

---
title: 'Build from source'
language: 'en'
---

## Build from the source

Before compiling Rio terminal, you'll have to first clone the source code:

```sh
git clone https://github.com/raphamorim/rio.git
```

Then install the Rust compiler with `rustup` ([rustup.rs](https://rustup.rs/)).

After installation of Rust, ensure you have the correct Rust compiler installed by running:

```sh
rustup override set stable
rustup update stable
```

### Dependencies

These are the minimum dependencies required to build Rio terminal, please note that with some setups additional dependencies might be desired.

If you're running Wayland with an Nvidia GPU and you are planning to use Rio with `OpenGL` as primary renderer backend, you'll likely want the _EGL_ drivers installed too (these are called `libegl1-mesa-dev` on Ubuntu).

#### Debian/Ubuntu

If you'd like to build a local version manually, you need a few extra libraries to build Rio. Here's an apt command that should install all of them. If something is still found to be missing, please open an issue. This has been tested on Debian Bookworm.

```sh
apt install cmake pkg-config libfreetype6-dev libfontconfig1-dev libxcb-xfixes0-dev libxkbcommon-dev python3
```

#### Arch Linux

On Arch Linux, you need a few extra libraries to build Rio. Here's a `pacman` command that should install all of them. If something is still found to be missing, please open an issue.

```sh
pacman -S cmake freetype2 fontconfig pkg-config make libxcb libxkbcommon python
```

#### Fedora

On Fedora, you need a few extra libraries to build Rio. Here's a `dnf` command that should install all of them. If something is still found to be missing, please open an issue.

```sh
dnf install cmake freetype-devel fontconfig-devel libxcb-devel libxkbcommon-devel g++
```

#### Void Linux

On Void Linux, install following packages before compiling Rio:

```sh
xbps-install cmake freetype-devel expat-devel fontconfig-devel libxcb-devel pkg-config python3
```

#### FreeBSD

On FreeBSD, you need a few extra libraries to build Rio. Here's a `pkg` command that should install all of them. If something is still found to be missing, please open an issue.

```sh
pkg install cmake freetype2 fontconfig pkgconf python3
```

### Building

Linux with X11:

```sh
# Build for X11
cargo build -p rioterm --release --no-default-features --features=x11

# Running it
target/release/rio
```

Linux with Wayland:

```sh
# Build for Wayland
cargo build -p rioterm --release --no-default-features --features=wayland

# Running it
target/release/rio
```

MacOS:

```sh
make release-macos
```

After the command execution an application called "Rio.app" will be created inside of a folder "release" (this folder is generated only after the command execution).

Windows:

```sh
cargo build -p rioterm --release
```

After the command execution an executable will be created called Rio.exe inside of "target/release"

Optionally you can also build and run the terminal with "cargo run".

If all goes well, this should place a zip file with Rio application inside at `release` (folder created in rio root path after the command execution).

---
title: 'Default colors'
language: 'en'
---

Default Rio terminal colors.

```toml
[colors]
# Regular colors
background = '#0F0D0E'
black = '#4C4345'
blue = '#006EE6'
cursor = '#F38BA3'
cyan = '#88DAF2'
foreground  = '#F9F4DA'
green = '#0BA95B'
magenta = '#7B5EA7'
red = '#ED203D'
white = '#F1F1F1'
yellow = '#FCBA28'

# Cursor
cursor = '#F712FF'
vi-cursor = '#12d0ff'

# Navigation
tabs = '#12B5E5'
tabs-foreground = '#7d7d7d'
tabs-active = '#303030'
tabs-active-highlight = '#ffa133'
tabs-active-foreground = '#FFFFFF'
bar = '#1b1a1a'

# Search
search-match-background = '#44C9F0'
search-match-foreground = '#FFFFFF'
search-focused-match-background = '#E6A003'
search-focused-match-foreground = '#FFFFFF'

# Selection
selection-foreground = '#0F0D0E'
selection-background = '#44C9F0'

# Dim colors
dim-black = '#1C191A'
dim-blue = '#0E91B7'
dim-cyan = '#93D4E7'
dim-foreground = '#ECDC8A'
dim-green = '#098749'
dim-magenta = '#624A87'
dim-red = '#C7102A'
dim-white = '#C1C1C1'
dim-yellow = '#E6A003'

# Light colors
light-black = '#ADA8A0'
light-blue = '#44C9F0'
light-cyan = '#7BE1FF'
light-foreground = '#F2EFE2'
light-green = '#0ED372'
light-magenta = '#9E88BE'
light-red = '#F25E73'
light-white = '#FFFFFF'
light-yellow = '#FDF170'
```

---
title: 'Adaptive theme'
language: 'en'
---

### Adaptive theme

Rio support theme based on the system theme (light and dark). This configuration only works for Web, MacOS and Windows.

```toml
[adaptive-theme]
light = "belafonte-day"
dark = "belafonte-night"
```

Example of usage with MacOS:

![Adaptive theme](/assets/features/adaptive-theme.gif)

---
layout: post
title: 'Looking ahead to 0.1.0'
date: 2024-04-18 00:34
description: "Let's talk about what's coming to Rio terminal"
categories: windows macos linux
---

import { Tweet } from 'react-tweet';

Hey there folks!

The Rio Terminal is a side project based on my free time (low this year due to being a new parent), when I started to work on Rio my goal was to learn mostly and eventually, I ended up using Rio in my daily life. However, the terminal wasn't designed to fit people's usage more than myself and would be quite hard to scale the source code to fit some of the feature requests, considering I never expected Rio to be used besides myself.

As I open-sourced Rio, many people started to use and eventually requests came in, making me realize that parts of the terminal would have to evolve to comprehend such requests and would eventually mean a rewrite of core parts along with incremental changes in other parts.

I have been working on 0.1.0 release since last December and still have a lot to be done ([Github backlog for 0.1.0](https://github.com/users/raphamorim/projects/6/views/1)). It is hard to give an estimation because there's no regular promised cadence and, I am not in a rush because it will require a lot of further testing and writing a lot of tests as well. Releases are likely to be made in small versions, is very likely that 0.1.0 will start with MacOS with Windows (I hope so at least).

Summary of 0.1.0 goals:

- Split screens.
- Stabilization of different fonts besides the default one.
- New windowing, migration from Winit to a project created for Rio purposes, called WA. WA uses gaming directives and is a mix of windowing architecture from [not-fl3/miniquad](https://github.com/not-fl3/miniquad) and [makepad/makepad](https://github.com/makepad/makepad) along with [winit](https://github.com/rust-windowing/winit).
- Sixel support.
- Font ligatures.
- Auto-Italicize and Auto-Bolding fonts.
- Addition of SIMD support for a few critical/repetitive operations.
- Emoji support.
- Give some love to Linux X11/Wayland to bring the native experience closer.

Note: Finally got a second computer with Windows/Linux, and it will make it insanely easier to work on Rio than I was doing before.

<Tweet id="1774416415073702204" />

---
title: 'editor'
language: 'en'
---

Default editor is `vi`.

Whenever the key binding `OpenConfigEditor` is triggered it will use the value of the editor along with the rio configuration path.

An example, considering you have VS Code installed and you want to use it as your editor:

```toml
[editor]
program = "code"
args = []
```

Whenever `OpenConfigEditor` runs it will trigger `$ code <path-to-rio-configuration-file>`.

---
title: 'startup-directory'
language: 'en'
---

Directory the shell is started in. If this is unset the working directory of the parent process will be used.

This configuration only works if [`use-fork`](#use-fork) is disabled.

```toml
working-dir = "/Users/raphael/Documents/"
```

---
title: 'Frequently Asked Questions'
language: 'en'
---

### I get errors about the terminal being unknown or opening the terminal failing or functional keys like arrow keys don’t work?

The famous case of "'rio' unknown terminal type".

All the issues all have the same root cause: the rio terminfo is not available. The most common way this happens is SSHing into a computer that does not have the rio terminfo files. The simplest fix would be install the terminfo files on the remote machine (following terminfo instruction in install section [install/terminfo](/docs/install/terminfo))

Other alternative is use a different terminfo either in config or per connection with environment variable like `TERM=xterm-256color ssh`.

If you are using a well maintained Linux distribution, it will have a rio-terminfo package that you can simply install to make the rio terminfo files available system-wide. Then the problem will no longer occur. Arch Linux for example [rio-terminfo](https://archlinux.org/packages/extra/x86_64/rio-terminfo/)

### Colors scheme not working as intended with tmux

The reason it happens is because tmux is using 256 colors configuration, you need to enable it or stop using Rio term without tmux (in case want to keep with 256 colors).

1. In case your shell does not report this correctly, add the following to your `~/.${SHELL}rc`:

```sh
export COLORTERM=truecolor
```

2. Please also note you need to add true color override to your `.tmux.conf` file:

```sh
set -g default-terminal "rio"
set-option -ga terminal-overrides ",rio:Tc"
```

3. Optionally you can also use `screen-256color`

```sh
set -g default-terminal "screen-256color"
set-option -ga terminal-overrides ",screen-256color:Tc"
```

---
title: 'padding'
language: 'en'
---

## Padding-x

Define x axis padding (default is 0)

```toml
padding-x = 10
```

## Padding-y

Define y axis padding based on a format `[top, left]`

- Default is `[0, 0]`

```toml
padding-y = [15, 10]
```

---
title: 'renderer'
language: 'en'
---

- `Performance` - Set WGPU rendering performance

  - `High`: Adapter that has the highest performance. This is often a discrete GPU.
  - `Low`: Adapter that uses the least possible power. This is often an integrated GPU.

- `Backend` - Set WGPU rendering backend

  - `Automatic`: Leave Sugarloaf/WGPU to decide
  - `GL`: Supported on Linux/Android, and Windows and macOS/iOS via ANGLE
  - `Vulkan`: Supported on Windows, Linux/Android
  - `DX12`: Supported on Windows 10
  - `Metal`: Supported on macOS/iOS

- `disable-unfocused-render` - This property disable renderer processes while Rio is unfocused.

- `max-fps` - Limits the maximum number of frames per second that rio terminal will attempt to draw. If you set as `0` then this limit will be ignored. The default on MacOS is 120 and all other platforms is 60.

Example:

```toml
[renderer]
performance = "High"
backend = "Automatic"
disable-unfocused-render = false
max-fps = 120
```

---
title: 'Configuration file'
language: 'en'
---

The configuration should be the following paths otherwise Rio will use the default configuration.

MacOS and Linux configuration file path is `~/.config/rio/config.toml`.

Windows configuration file path is `C:\Users\USER\AppData\Local\rio\config.toml` (replace "USER" with your user name).

Updates to the configuration file automatically triggers Rio to render the terminal with the new configuration.

Note that all parameters without a header must be at the beginning of the file, otherwise they will be ignored. Example:

```toml
[editor]
program = "vi"
args = []

theme = "dracula" # ignore it, be under the `editor` header
```

```toml
theme = "dracula" # it works, be without heading

[editor]
program = "vi"
args = []
```

---
layout: post
title: 'Rio 0.0.8'
date: 2023-07-03 12:34
description: 'Support to Microsoft Windows, Homebrew casks and formulas, migration to Corcovado, selection improvements and increase/decrease font-size in a session.'
categories: release windows macos linux
---

import { Tweet } from 'react-tweet';

## v0.0.8: Highlights

- Support to Microsoft Windows.
- Support to Homebrew as cask [formulae.brew.sh/cask/rio](https://formulae.brew.sh/cask/rio).
- Support to Homebrew as formula [formulae.brew.sh/formula/rio](https://formulae.brew.sh/formula/rio).
- Selection improvements and increase/decrease font-size in a session.
- Migration to Corcovado.

<br />

# Overview

Rio release 0.0.8 is finally here, there's a lot of updates to cover so let's get started. I also would like to remind you that Rio is still unstable. This release adds a lot of bug fixes and feature additions in order to make Rio terminal stable.

## Microsoft Windows

Added support to Microsoft Windows.

Windows 10:
![Demo windows in use](/assets/posts/0.0.8/demo-windows-11.png)

Windows 11:
![Demo windows in search](/assets/posts/0.0.8/demo-windows-11.png)

## Homebrew as Cask

Homebrew is a free and open-source software package management system that simplifies the installation of software on Apple's operating system, macOS.

Rio has been added as a cask [formulae.brew.sh/cask/rio](https://formulae.brew.sh/cask/rio) to their package source. To install Rio with homebrew you need to run the command below:

```sh
brew update && brew install --cask rio
```

Github reference: [github.com/Homebrew/homebrew-cask/pull/149824](https://github.com/Homebrew/homebrew-cask/pull/149824)

## Homebrew as Formula

Rio has been added as a formula [formulae.brew.sh/formula/rio](https://formulae.brew.sh/formula/rio) to their package source. To install Rio with homebrew you need to run the command below:

```sh
brew update && brew install rio
```

Github reference: [github.com/Homebrew/homebrew-cask/pull/149824](https://github.com/Homebrew/homebrew-core/pull/134295)

## Selection improvements and increase/decrease font-size in a session

The version v0.0.8 has added the following missing functionalities: Scroll and select, Semantic and line selection and the ability to increase, decrease and reset font size using keyboard shortcut during session coming for Rio terminal.

Below you can see a demo with all those functionalities:

<Tweet id="1673705339336761344" />

## Migration to Corcovado

Rio terminal migrated from Mio to Corcovado. Corcovado is a maintained fork of mio 0.6.x along mio-signal-hook, mio-extras and using Windows API that works in Windows 11. It uses Rust edition 2021 instead of 2018.

Corcovado also uses Rust standard library for net and io instead of Mio 0.6.x.

<br />

## Changelog of v0.0.8

- Added generation of ".msi" and ".exe" files to the release pipeline (stable and canary).
- Support to Microsoft Windows 11.
- Ability to in/decrease font size using keyboard shortcut during session (ref: [#109](https://github.com/raphamorim/rio/issues/109))
- Inverted Canary and Stable icons.
- ANSI mouse reports (e.g: scroll and click working on VIM).
- Scroll and apply selection.
- Semantic and line selection.
- Rio is available in Homebrew casks and formulas.
- Rio stable versions are notarized now.
- Migration of mio, mio-extras, mio-signal-hook to Corcovado.
- Changed default black color to "#4c4345".
- Fix mouse position for when selecting text.

---
layout: docs
class: docs
title: 'Default key bindings'
language: 'en'
---

### MacOS

| Action                 | Key                                                            |
| ---------------------- | -------------------------------------------------------------- |
| Open configuration     | `Command + Comma (,)`                                          |
| Toggle VI Mode         | `Alt + Shift + Space`                                          |
| Copy                   | `Command + C`                                                  |
| Paste                  | `Command + V`                                                  |
| Create new window      | `Command + N`                                                  |
| Create new tab         | `Command + T`                                                  |
| Move to next tab       | `Control + Tab` or `Command + Shift + RightBracket (])`        |
| Move to previous tab   | `Control + Shift + Tab` or `Command + Shift + LeftBracket ([)` |
| Increase font size     | `Command + Plus (+)`                                           |
| Decrease font size     | `Command + Minus (-)`                                          |
| Reset font size        | `Command + 0`                                                  |
| Minimize window        | `Command + M`                                                  |
| Quit                   | `Command + Q`                                                  |
| Close tab              | `Command + W`                                                  |
| Select the first tab   | `Command + 1`                                                  |
| Select the second tab  | `Command + 2`                                                  |
| Select the third tab   | `Command + 3`                                                  |
| Select the fourth tab  | `Command + 4`                                                  |
| Select the fifth tab   | `Command + 5`                                                  |
| Select the sixth tab   | `Command + 6`                                                  |
| Select the seventh tab | `Command + 7`                                                  |
| Select the eighth tab  | `Command + 8`                                                  |
| Select the last tab    | `Command + 9`          |
| Search forward         | `Command + f`          |
| Search backward        | `Command + b`          |

### Windows

| Action               | Key                                                         |
| -------------------- | ----------------------------------------------------------- |
| Open configuration   | `Control + Shift + Comma (,)`                               |
| Toggle VI Mode       | `Control + Shift + Space`                                   |
| Copy                 | `Control + Shift + C`                                       |
| Paste                | `Control + Shift + V`                                       |
| Create new window    | `Control + Shift + N`                                       |
| Create new tab       | `Control + Shift + T`                                       |
| Move to next tab     | `Control + Tab` or `Control + Shift + RightBracket (])`     |
| Move to previous tab | `Control + Shift + Tab` `Control + Shift + LeftBracket ([)` |
| Increase font size   | `Control + Plus (+)`                                        |
| Decrease font size   | `Control + Minus (-)`                                       |
| Reset font size      | `Control + 0`                                               |
| Close tab or quit    | `Control + Shift + W`                                       |
| Select the first tab   | `Control + Shift + 1`                                                  |
| Select the second tab  | `Control + Shift + 2`                                                  |
| Select the third tab   | `Control + Shift + 3`                                                  |
| Select the fourth tab  | `Control + Shift + 4`                                                  |
| Select the fifth tab   | `Control + Shift + 5`                                                  |
| Select the sixth tab   | `Control + Shift + 6`                                                  |
| Select the seventh tab | `Control + Shift + 7`                                                  |
| Select the eighth tab  | `Control + Shift + 8`                                                  |
| Select the last tab    | `Control + Shift + 9`   |
| Search forward         | `Control + Shift + f`  |
| Search backward        | `Control + Shift + b`  |

### Linux and BSD

| Action               | Key                                                            |
| -------------------- | -------------------------------------------------------------- |
| Open configuration   | `Control + Shift + Comma (,)`                                  |
| Toggle VI Mode       | `Alt + Shift + Space`                                          |
| Copy                 | `Control + Shift + C`                                          |
| Paste                | `Control + Shift + V`                                          |
| Create new window    | `Control + Shift + N`                                          |
| Create new tab       | `Control + Shift + T`                                          |
| Move to next tab     | `Control + Tab` or `Control + Shift + RightBracket (])`        |
| Move to previous tab | `Control + Shift + Tab` or `Control + Shift + LeftBracket ([)` |
| Increase font size   | `Control + Plus (+)`                                           |
| Decrease font size   | `Control + Minus (-)`                                          |
| Reset font size      | `Control + 0`                                                  |
| Close tab or quit    | `Control + Shift + W`                                          |
| Select the first tab   | `Control + Shift + 1`                                                  |
| Select the second tab  | `Control + Shift + 2`                                                  |
| Select the third tab   | `Control + Shift + 3`                                                  |
| Select the fourth tab  | `Control + Shift + 4`                                                  |
| Select the fifth tab   | `Control + Shift + 5`                                                  |
| Select the sixth tab   | `Control + Shift + 6`                                                  |
| Select the seventh tab | `Control + Shift + 7`                                                  |
| Select the eighth tab  | `Control + Shift + 8`                                                  |
| Select the last tab    | `Control + Shift + 9`          |
| Search forward          | `Control + Shift + f`  |
| Search backward         | `Control + Shift + b`  |

### Search

| Action                  | Key                                    |
| ----------------------- | -------------------------------------- |
| Quit search             | `Escape` (while search is open)        |
| Quit search             | `Control + c` (while search is open)   |
| Focus next              | `Enter` (while search is open)         |
| Focus previous          | `Shift + Enter` (while search is open) |
| Clear search word       | `Control + w` (while search is open)   |
| Search history previous | `Control + p` (while search is open)   |
| Search history previous | `ArrowUp` (while search is open)       |
| Search history next     | `Control + n` (while search is open)   |
| Search history next     | `ArrowDown` (while search is open)     |

---
title: 'iTerm2 Image Protocol'
language: 'en'
---

Rio implements support for the iTerm2 inline image protocol.

To render an image inline in your terminal, you can use `imgcat` provided as a script by iTerm2 in [iterm2.com/utilities/imgcat](https://iterm2.com/utilities/imgcat) or from other sources like:

- [pypi.org/project/imgcat](https://pypi.org/project/imgcat/)
- [npmjs.com/package/term-img](https://www.npmjs.com/package/term-img)
- [npmjs.com/package/imgcat](https://www.npmjs.com/package/imgcat)
- ... and etecetera.

![Demo iTerm2 image protocol](/assets/features/demo-iterm2-image-protocol.png)

More info regarding [iTerm image protocol](https://iterm2.com/documentation-images.html)

---
title: 'scroll'
language: 'en'
---

You can change how many lines are scrolled each time by setting this option. Scroll calculation for canonical mode will be based on `lines = (accumulated scroll * multiplier / divider)`.

If you want a quicker scroll, keep increasing the multiplier. If you want to reduce scroll speed you will need to increase the divider.

You can combine both properties to find the best scroll for you.

- Multiplier default is `3.0`.
- Divider default is `1.0`.

Example:

```toml
[scroll]
multiplier = 3.0
divider = 1.0
```

---
title: 'fonts'
language: 'en'
---

Configure fonts used by the terminal.

Note: You can set different font families but Rio terminal
will always look for regular font bounds whene

You can also set family on root to overwrite all fonts.

```toml
[fonts]
family = "cascadiacode"
```

## Extra fonts

You can also specify extra fonts to load:

```toml
[fonts]
extras = [{ family = "Microsoft JhengHei" }]
```

## Font features

In case you want to specify any font feature:

```toml
[fonts]
features = ["ss02", "ss03", "ss05", "ss19"]
```

Note: Font features do not have support to live reload on configuration, so to reflect your changes, you will need to close and reopen Rio.

## Default configuration

The font configuration default:

```toml
[fonts]
size = 18
features = []

[fonts.regular]
family = "cascadiacode"
style = "normal"
weight = 400

[fonts.bold]
family = "cascadiacode"
style = "normal"
weight = 800

[fonts.italic]
family = "cascadiacode"
style = "italic"
weight = 400

[fonts.bold-italic]
family = "cascadiacode"
style = "italic"
weight = 800
```

## Emojis

You can also specify which emoji font you would like to use, by default will be loaded a built-in Twemoji color by Mozilla.

In case you would like to change:

```toml
# Apple
# [fonts.emoji]
# family = "Apple Color Emoji"

# In case you have Noto Color Emoji installed
# [fonts.emoji]
# family = "Noto Color Emoji"
```

## User interface

You can specify user interface font on Rio.

Note: `fonts.ui` does not have live reload configuration update, you need to close and open Rio again.

```toml
[fonts.ui]
family = "Departure Mono"
```

---
layout: post
title: 'New font configuration API and native tabs'
date: 2023-08-29 17:53
description: 'Rio terminal release: New font configuration API, native tabs, kitty keyboard protocol and other stuff.'
categories: windows macos linux webassembly
---

This post also includes changes from v0.0.16 and v0.0.17.

## Highlights

- New font configuration API.
- Native Tabs for MacOS.

## Breaking change: New font API

If you don't use Rio default font configuration you will be affected by this change.

A new API has been introduced to select fonts, now you can configure even different font families for "regular", "bold", "italic" and "bold-italic".

```toml
[fonts]
size = 18

[fonts.regular]
family = "cascadiamono"
style = "normal"
weight = 400

[fonts.bold]
family = "cascadiamono"
style = "normal"
weight = 800

[fonts.italic]
family = "cascadiamono"
style = "italic"
weight = 400

[fonts.bold-italic]
family = "cascadiamono"
style = "italic"
weight = 800
```

## Native Tabs (macOs only)

Native tabs have arrived for MacOS users.

![native tab example](/assets/posts/0.0.17/demo-native-tabs.png)

To configure it you will need to update the configuration file and set navigation mode as "NativeTab":

```toml
[navigation]
mode = "NativeTab"
```

## Changelog of v0.0.17 along with v0.0.16

- Configuration "font" does not work anymore, a new configuration API of font selection has been introduced.
- Action "TabSwitchNext" and "TabSwitchPrev" has been renamed to "SelectNextTab" and "SelectPrevTab".
- Support to "NativeTab" (MacOS only).
- Support for kitty's keyboard protocol ("CSI u"). Ref: https://sw.kovidgoyal.net/kitty/keyboard-protocol/
- Added new actions for tab selection: "SelectTab1", "SelectTab2", "SelectTab3", "SelectTab4", "SelectTab5", "SelectTab6", "SelectTab7", "SelectTab8", "SelectTab9", "SelectLastTab".
- Support lowercased action and fix overwrite for actions in custom key bindings.
- Added action "Minimize" for minimize Rio terminal window.
- Added action "ClearHistory" for clear terminal saved history.
- Added action "ReceiveChar" for custom key bindings.
- New default key bindings for Linux and Windows so that conflicts with readline key bindings are removed.
- Winit Version 0.29.1-beta.
- Allow paste with the middle mouse of the button (fixes https://github.com/raphamorim/rio/issues/123).
- Support startup notify protocol to raise initial window on Wayland/X11.
- Fix Double-tap by touchpad on the titlebar doesn't maximize/unmaximize the window in GNOME 44, Wayland.
- Fix tab/breadcrumb bug introduced in 0.0.15
- Introduce new configuration property: "navigation.macos-hide-window-button".

---
title: 'colors'
language: 'en'
---

Defining colors in the configuration file will not have any effect if you're using a theme.

The default configuration is without a theme.

Example:

```toml
[colors]
background = '#0F0D0E'
foreground = '#F9F4DA'
cursor = '#F38BA3'
tabs = '#443d40'
tabs-active = '#F38BA3'
green = '#0BA95B'
red = '#ED203D'
blue = '#12B5E5'
yellow = '#FCBA28'
```

---
title: 'MacOS'
language: 'en'
---

You can download Rio terminal application for macOS platform:

- [Download Rio for MacOS v0.1.14](https://github.com/raphamorim/rio/releases/download/v0.1.14/Rio-v0.1.14.dmg)

Alternatively you can install Rio through [Homebrew](https://brew.sh/)...

```sh
brew install --cask rio
```

...or [MacPorts](https://www.macports.org):

```sh
sudo port install rio
```

**For Homebrew:** remember to run a "brew update" in case Homebrew cannot find a rio cask to install.

**For MacPorts:** more details [here](https://ports.macports.org/port/rio/).

Canary versions for MacOS are not notarized, so if you want to install a canary version you need to download and install the canary app from [github.com/raphamorim/rio/releases](https://github.com/raphamorim/rio/releases) and then follow the steps below:

- Try to run, it will show a window explaining it cannot be opened because "Apple cannot check it for malicious software.", then click Ok.
- Open System Preferences and select "Security & Privacy".
- If the padlock in the bottom left is locked, click it and authenticate to unlock it.
- Next to the message explaining the app "was blocked from use because it is not from an identified developer," click "Open Anyway".
- Close System Preferences and run the app.
- A notice will reiterate the warning about an inability to check if it is malicious, click Open.

---
title: 'Using cargo'
language: 'en'
---

Rio can be installed using Cargo, the package manager for Rust.

Cargo is available for most platforms, and is a good alternative to building Rio from source.

Rust and Cargo can be installed from [rust-lang.org](https://www.rust-lang.org/learn/get-started).

You might need to install additional packages depending on your platform. See the guide to [building from source](/docs/install/build-from-source) for more information.

```sh
cargo install rioterm
```

Then simply run `rio` to open the terminal.

```sh
rio
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/suggestions.R
\name{install_formats}
\alias{install_formats}
\alias{show_unsupported_formats}
\title{Install rio's \sQuote{Suggests} Dependencies}
\usage{
install_formats(...)

show_unsupported_formats()
}
\arguments{
\item{\dots}{Additional arguments passed to \code{\link[utils:install.packages]{utils::install.packages()}}.}
}
\value{
For \code{show_unsupported_formats()}, if there is any missing unsupported formats, it return TRUE invisibly; otherwise FALSE. For \code{install_formats()} it returns TRUE invisibly if the installation is succuessful; otherwise errors.
}
\description{
Not all suggested packages are installed by default. These packages are not installed or loaded by default in order to create a slimmer and faster package build, install, and load. Use \code{show_unsupported_formats()} to check all unsupported formats. \code{install_formats()} installs all missing \sQuote{Suggests} dependencies for rio that expand its support to the full range of support import and export formats.
}
\examples{
\donttest{
if (interactive()) {
    install_formats()
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/export_list.R
\name{export_list}
\alias{export_list}
\title{Export list of data frames to files}
\usage{
export_list(x, file, archive = "", ...)
}
\arguments{
\item{x}{A list of data frames to be written to files.}

\item{file}{A character vector string containing a single file name with a \verb{\\\%s} wildcard placeholder, or a vector of file paths for multiple files to be imported. If \code{x} elements are named, these will be used in place of \verb{\\\%s}, otherwise numbers will be used; all elements must be named for names to be used.}

\item{archive}{character. Either empty string (default) to save files in current
directory, a path to a (new) directory, or a .zip/.tar file to compress all
files into an archive.}

\item{\dots}{Additional arguments passed to \code{\link[=export]{export()}}.}
}
\value{
The name(s) of the output file(s) as a character vector (invisibly).
}
\description{
Use \code{\link[=export]{export()}} to export a list of data frames to a vector of file names or a filename pattern.
}
\details{
\code{\link[=export]{export()}} can export a list of data frames to a single multi-dataset file (e.g., an Rdata or Excel .xlsx file). Use \code{export_list} to export such a list to \emph{multiple} files.
}
\examples{
## For demo, a temp. file path is created with the file extension .xlsx
xlsx_file <- tempfile(fileext = ".xlsx")
export(
    list(
        mtcars1 = mtcars[1:10, ],
        mtcars2 = mtcars[11:20, ],
        mtcars3 = mtcars[21:32, ]
    ),
    xlsx_file
)

# import a single file from multi-object workbook
import(xlsx_file, sheet = "mtcars1")
# import all worksheets, the return value is a list
import_list(xlsx_file)
library('datasets')
export(list(mtcars1 = mtcars[1:10,],
            mtcars2 = mtcars[11:20,],
            mtcars3 = mtcars[21:32,]),
    xlsx_file <- tempfile(fileext = ".xlsx")
)

# import all worksheets
list_of_dfs <- import_list(xlsx_file)

# re-export as separate named files

## export_list(list_of_dfs, file = c("file1.csv", "file2.csv", "file3.csv"))

# re-export as separate files using a name pattern; using the names in the list
## This will be written as "mtcars1.csv", "mtcars2.csv", "mtcars3.csv"

## export_list(list_of_dfs, file = "\%s.csv")
}
\seealso{
\code{\link[=import]{import()}}, \code{\link[=import_list]{import_list()}}, \code{\link[=export]{export()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/export.R
\name{export}
\alias{export}
\title{Export}
\usage{
export(x, file, format, ...)
}
\arguments{
\item{x}{A data frame, matrix or a single-item list of data frame to be written into a file. Exceptions to this rule are that \code{x} can be a list of multiple data frames if the output file format is an OpenDocument Spreadsheet (.ods, .fods), Excel .xlsx workbook, .Rdata file, or HTML file, or a variety of R objects if the output file format is RDS or JSON. See examples.) To export a list of data frames to multiple files, use \code{\link[=export_list]{export_list()}} instead.}

\item{file}{A character string naming a file. Must specify \code{file} and/or \code{format}.}

\item{format}{An optional character string containing the file format, which can be used to override the format inferred from \code{file} or, in lieu of specifying \code{file}, a file with the symbol name of \code{x} and the specified file extension will be created. Must specify \code{file} and/or \code{format}. Shortcuts include: \dQuote{,} (for comma-separated values), \dQuote{;} (for semicolon-separated values), \dQuote{|} (for pipe-separated values), and \dQuote{dump} for \code{\link[base:dump]{base::dump()}}.}

\item{\dots}{Additional arguments for the underlying export functions. This can be used to specify non-standard arguments. See examples.}
}
\value{
The name of the output file as a character string (invisibly).
}
\description{
Write data.frame to a file
}
\details{
This function exports a data frame or matrix into a file with file format based on the file extension (or the manually specified format, if \code{format} is specified).

The output file can be to a compressed directory, simply by adding an appropriate additional extensiont to the \code{file} argument, such as: \dQuote{mtcars.csv.tar}, \dQuote{mtcars.csv.zip}, or \dQuote{mtcars.csv.gz}.

\code{export} supports many file formats. See the documentation for the underlying export functions for optional arguments that can be passed via \code{...}

\itemize{
\item Comma-separated data (.csv), using \code{\link[data.table:fwrite]{data.table::fwrite()}}
\item Pipe-separated data (.psv), using \code{\link[data.table:fwrite]{data.table::fwrite()}}
\item Tab-separated data (.tsv), using \code{\link[data.table:fwrite]{data.table::fwrite()}}
\item SAS (.sas7bdat), using \code{\link[haven:write_sas]{haven::write_sas()}}.
\item SAS XPORT (.xpt), using \code{\link[haven:read_xpt]{haven::write_xpt()}}.
\item SPSS (.sav), using \code{\link[haven:read_spss]{haven::write_sav()}}
\item SPSS compressed (.zsav), using \code{\link[haven:read_spss]{haven::write_sav()}}
\item Stata (.dta), using \code{\link[haven:read_dta]{haven::write_dta()}}. Note that variable/column names containing dots (.) are not allowed and will produce an error.
\item Excel (.xlsx), using \code{\link[writexl:write_xlsx]{writexl::write_xlsx()}}. \code{x} can also be a list of data frames; the list entry names are used as sheet names.
\item R syntax object (.R), using \code{\link[base:dput]{base::dput()}} (by default) or \code{\link[base:dump]{base::dump()}} (if \code{format = 'dump'})
\item Saved R objects (.RData,.rda), using \code{\link[base:save]{base::save()}}. In this case, \code{x} can be a data frame, a named list of objects, an R environment, or a character vector containing the names of objects if a corresponding \code{envir} argument is specified.
\item Serialized R objects (.rds), using \code{\link[base:readRDS]{base::saveRDS()}}. In this case, \code{x} can be any serializable R object.
\item Serialized R objects (.qs2), using \code{\link[qs2:qs_save]{qs2::qs_save()}}.
\item "XBASE" database files (.dbf), using \code{\link[foreign:write.dbf]{foreign::write.dbf()}}
\item Weka Attribute-Relation File Format (.arff), using \code{\link[foreign:write.arff]{foreign::write.arff()}}
\item Fixed-width format data (.fwf), using \code{\link[utils:write.table]{utils::write.table()}} with \code{row.names = FALSE}, \code{quote = FALSE}, and \code{col.names = FALSE}
\item \href{https://github.com/csvy}{CSVY} (CSV with a YAML metadata header) using \code{\link[data.table:fwrite]{data.table::fwrite()}}.
\item Apache Arrow Parquet (.parquet), using \code{\link[nanoparquet:write_parquet]{nanoparquet::write_parquet()}}
\item Feather R/Python interchange format (.feather), using \code{\link[arrow:write_feather]{arrow::write_feather()}}
\item Fast storage (.fst), using \code{\link[fst:write_fst]{fst::write.fst()}}
\item JSON (.json), using \code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}. In this case, \code{x} can be a variety of R objects, based on class mapping conventions in this paper: \url{https://arxiv.org/abs/1403.2805}.
\item Matlab (.mat), using \code{\link[rmatio:write.mat-methods]{rmatio::write.mat()}}
\item OpenDocument Spreadsheet (.ods, .fods), using \code{\link[readODS:write_ods]{readODS::write_ods()}} or \code{\link[readODS:write_ods]{readODS::write_fods()}}.
\item HTML (.html), using a custom method based on \code{\link[xml2:xml_replace]{xml2::xml_add_child()}} to create a simple HTML table and \code{\link[xml2:write_xml]{xml2::write_xml()}} to write to disk.
\item XML (.xml), using a custom method based on \code{\link[xml2:xml_replace]{xml2::xml_add_child()}} to create a simple XML tree and \code{\link[xml2:write_xml]{xml2::write_xml()}} to write to disk.
\item YAML (.yml), using \code{\link[yaml:write_yaml]{yaml::write_yaml()}}, default to write the content with UTF-8. Might not work on some older systems, e.g. default Windows locale for R <= 4.2.
\item Clipboard export (on Windows and Mac OS), using \code{\link[utils:write.table]{utils::write.table()}} with \code{row.names = FALSE}
}

When exporting a data set that contains label attributes (e.g., if imported from an SPSS or Stata file) to a plain text file, \code{\link[=characterize]{characterize()}} can be a useful pre-processing step that records value labels into the resulting file (e.g., \code{export(characterize(x), "file.csv")}) rather than the numeric values.

Use \code{\link[=export_list]{export_list()}} to export a list of dataframes to separate files.
}
\examples{
## For demo, a temp. file path is created with the file extension .csv
csv_file <- tempfile(fileext = ".csv")
## .xlsx
xlsx_file <- tempfile(fileext = ".xlsx")

## create CSV to import
export(iris, csv_file)

## You can certainly export your data with the file name, which is not a variable:
## import(mtcars, "car_data.csv")

## pass arguments to the underlying function
## data.table::fwrite is the underlying function and `col.names` is an argument
export(iris, csv_file, col.names = FALSE)

## export a list of data frames as worksheets
export(list(a = mtcars, b = iris), xlsx_file)

# NOT RECOMMENDED

## specify `format` to override default format
export(iris, xlsx_file, format = "csv") ## That's confusing
## You can also specify only the format; in the following case
## "mtcars.dta" is written [also confusing]

## export(mtcars, format = "stata")
}
\seealso{
\code{\link[=characterize]{characterize()}}, \code{\link[=import]{import()}}, \code{\link[=convert]{convert()}}, \code{\link[=export_list]{export_list()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert.R
\name{convert}
\alias{convert}
\title{Convert from one file format to another}
\usage{
convert(in_file, out_file, in_opts = list(), out_opts = list())
}
\arguments{
\item{in_file}{A character string naming an input file.}

\item{out_file}{A character string naming an output file.}

\item{in_opts}{A named list of options to be passed to \code{\link[=import]{import()}}.}

\item{out_opts}{A named list of options to be passed to \code{\link[=export]{export()}}.}
}
\value{
A character string containing the name of the output file (invisibly).
}
\description{
This function constructs a data frame from a data file using \code{\link[=import]{import()}} and uses \code{\link[=export]{export()}} to write the data to disk in the format indicated by the file extension.
}
\examples{
## For demo, a temp. file path is created with the file extension .dta (Stata)
dta_file <- tempfile(fileext = ".dta")
## .csv
csv_file <- tempfile(fileext = ".csv")
## .xlsx
xlsx_file <- tempfile(fileext = ".xlsx")


## Create a Stata data file
export(mtcars, dta_file)

## convert Stata to CSV and open converted file
convert(dta_file, csv_file)
import(csv_file)

## correct an erroneous file format
export(mtcars, xlsx_file, format = "tsv") ## DON'T DO THIS
## import(xlsx_file) ## ERROR
## convert the file by specifying `in_opts`
convert(xlsx_file, xlsx_file, in_opts = list(format = "tsv"))
import(xlsx_file)

## convert from the command line:
## Rscript -e "rio::convert('mtcars.dta', 'mtcars.csv')"
}
\seealso{
\href{https://lbraglia.github.io/}{Luca Braglia} has created a Shiny app called \href{https://github.com/lbraglia/rioweb}{rioweb} that provides access to the file conversion features of rio through a web browser.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gather_attrs.R
\name{gather_attrs}
\alias{gather_attrs}
\alias{spread_attrs}
\title{Gather attributes from data frame variables}
\usage{
gather_attrs(x)

spread_attrs(x)
}
\arguments{
\item{x}{A data frame.}
}
\value{
\code{x}, with variable-level attributes stored at the data frame level.
}
\description{
\code{gather_attrs} moves variable-level attributes to the data frame level and \code{spread_attrs} reverses that operation.
}
\details{
\code{\link[=import]{import()}} attempts to standardize the return value from the various import functions to the extent possible, thus providing a uniform data structure regardless of what import package or function is used. It achieves this by storing any optional variable-related attributes at the variable level (i.e., an attribute for \code{mtcars$mpg} is stored in \code{attributes(mtcars$mpg)} rather than \code{attributes(mtcars)}). \code{gather_attrs} moves these to the data frame level (i.e., in \code{attributes(mtcars)}). \code{spread_attrs} moves attributes back to the variable level.
}
\seealso{
\code{\link[=import]{import()}}, \code{\link[=characterize]{characterize()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_list.R
\name{import_list}
\alias{import_list}
\title{Import list of data frames}
\usage{
import_list(
  file,
  setclass = getOption("rio.import.class", "data.frame"),
  which,
  rbind = FALSE,
  rbind_label = "_file",
  rbind_fill = TRUE,
  ...
)
}
\arguments{
\item{file}{A character string containing a single file name for a multi-object file (e.g., Excel workbook, zip file, tar archive, or HTML file), or a vector of file paths for multiple files to be imported.}

\item{setclass}{An optional character vector specifying one or more classes
to set on the import. By default, the return object is always a
\dQuote{data.frame}. Allowed values include \dQuote{tbl_df}, \dQuote{tbl}, or
\dQuote{tibble} (if using tibble), \dQuote{arrow}, \dQuote{arrow_table} (if using arrow table; the suggested package \code{arrow} must be installed) or \dQuote{data.table} (if using
data.table). Other values are ignored, such that a data.frame is returned.
The parameter takes precedents over parameters in \dots which set a different class.}

\item{which}{If \code{file} is a single file path, this specifies which objects should be extracted (passed to \code{\link[=import]{import()}}'s \code{which} argument). Ignored otherwise.}

\item{rbind}{A logical indicating whether to pass the import list of data frames through \code{\link[data.table:rbindlist]{data.table::rbindlist()}}.}

\item{rbind_label}{If \code{rbind = TRUE}, a character string specifying the name of a column to add to the data frame indicating its source file.}

\item{rbind_fill}{If \code{rbind = TRUE}, a logical indicating whether to set the \code{fill = TRUE} (and fill missing columns with \code{NA}).}

\item{\dots}{Additional arguments passed to \code{\link[=import]{import()}}. Behavior may be unexpected if files are of different formats.}
}
\value{
If \code{rbind=FALSE} (the default), a list of a data frames. Otherwise, that list is passed to \code{\link[data.table:rbindlist]{data.table::rbindlist()}} with \code{fill = TRUE} and returns a data frame object of class set by the \code{setclass} argument; if this operation fails, the list is returned.
}
\description{
Use \code{\link[=import]{import()}} to import a list of data frames from a vector of file names or from a multi-object file (Excel workbook, .Rdata file, compressed directory in a zip file or tar archive, or HTML file)
}
\details{
When file is a vector of file paths and any files are missing, those files are ignored (with warnings) and this function will not raise any error. For compressed files, the file name must also contain information about the file format of all compressed files, e.g. \code{files.csv.zip} for this function to work.
}
\section{Trust}{
For serialization formats (.R, .RDS, and .RData), please note that you should only load these files from trusted sources. It is because these formats are not necessarily for storing rectangular data and can also be used to store many things, e.g. code. Importing these files could lead to arbitary code execution. Please read the security principles by the R Project (Plummer, 2024). When importing these files via \code{rio}, you should affirm that you trust these files, i.e. \code{trust = TRUE}. See example below. If this affirmation is missing, the current version assumes \code{trust} to be true for backward compatibility and a deprecation notice will be printed. In the next major release (2.0.0), you must explicitly affirm your trust when importing these files.
}

\section{Which}{
For compressed archives (zip and tar, where a compressed file can contain multiple files), it is possible to come to a situation where the parameter \code{which} is used twice to indicate two different concepts. For example, it is unclear for \code{.xlsx.zip}whether \code{which} refers to the selection of an exact file in the archive or the selection of an exact sheet in the decompressed Excel file. In these cases, \code{rio} assumes that \code{which} is only used for the selection of file. After the selection of file with \code{which}, \code{rio} will return the first item, e.g. the first sheet.

Please note, however, \code{.gz} and \code{.bz2} (e.g. \code{.xlsx.gz}) are compressed, but not archive format. In those cases, \code{which} is used the same way as the non-compressed format, e.g. selection of sheet for Excel.
}

\examples{
## For demo, a temp. file path is created with the file extension .xlsx
xlsx_file <- tempfile(fileext = ".xlsx")
export(
    list(
        mtcars1 = mtcars[1:10, ],
        mtcars2 = mtcars[11:20, ],
        mtcars3 = mtcars[21:32, ]
    ),
    xlsx_file
)

# import a single file from multi-object workbook
import(xlsx_file, sheet = "mtcars1")
# import all worksheets, the return value is a list
import_list(xlsx_file)

# import and rbind all worksheets, the return value is a data frame
import_list(xlsx_file, rbind = TRUE)
}
\references{
Plummer, M (2024). Statement on CVE-2024-27322. \url{https://blog.r-project.org/2024/05/10/statement-on-cve-2024-27322/}
}
\seealso{
\code{\link[=import]{import()}}, \code{\link[=export_list]{export_list()}}, \code{\link[=export]{export()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_info}
\alias{get_info}
\alias{get_ext}
\title{Get File Info}
\usage{
get_info(file)

get_ext(file)
}
\arguments{
\item{file}{A character string containing a filename, file path, or URL.}
}
\value{
For \code{\link[=get_info]{get_info()}}, a list is return with the following slots
\itemize{
\item \code{input} file extension or information used to identify the possible file format
\item \code{format} file format, see \code{format} argument of \code{\link[=import]{import()}}
\item \code{type} "import" (supported by default); "suggest" (supported by suggested packages, see \code{\link[=install_formats]{install_formats()}}); "enhance" and "known " are not directly supported; \code{NA} is unsupported
\item \code{format_name} name of the format
\item \code{import_function} What function is used to import this file
\item \code{export_function} What function is used to export this file
\item \code{file} \code{file}
}
For \code{\link[=get_ext]{get_ext()}}, just \code{input} (usually file extension) is returned; retained for backward compatibility.
}
\description{
A utility function to retrieve the file information of a filename, path, or URL.
}
\examples{
get_info("starwars.xlsx")
get_info("starwars.ods")
get_info("https://github.com/ropensci/readODS/raw/v2.1/starwars.ods")
get_info("~/duran_duran_rio.mp3")
get_ext("clipboard") ## "clipboard"
get_ext("https://github.com/ropensci/readODS/raw/v2.1/starwars.ods")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/characterize.R
\name{characterize}
\alias{characterize}
\alias{factorize}
\alias{characterize.default}
\alias{characterize.data.frame}
\alias{factorize.default}
\alias{factorize.data.frame}
\title{Character conversion of labelled data}
\usage{
characterize(x, ...)

factorize(x, ...)

\method{characterize}{default}(x, ...)

\method{characterize}{data.frame}(x, ...)

\method{factorize}{default}(x, coerce_character = FALSE, ...)

\method{factorize}{data.frame}(x, ...)
}
\arguments{
\item{x}{A vector or data frame.}

\item{\dots}{additional arguments passed to methods}

\item{coerce_character}{A logical indicating whether to additionally coerce character columns to factor (in \code{factorize}). Default \code{FALSE}.}
}
\value{
a character vector (for \code{characterize}) or factor vector (for \code{factorize})
}
\description{
Convert labelled variables to character or factor
}
\details{
\code{characterize} converts a vector with a \code{labels} attribute of named levels into a character vector. \code{factorize} does the same but to factors. This can be useful at two stages of a data workflow: (1) importing labelled data from metadata-rich file formats (e.g., Stata or SPSS), and (2) exporting such data to plain text files (e.g., CSV) in a way that preserves information.
}
\examples{
## vector method
x <- structure(1:4, labels = c("A" = 1, "B" = 2, "C" = 3))
characterize(x)
factorize(x)

## data frame method
x <- data.frame(v1 = structure(1:4, labels = c("A" = 1, "B" = 2, "C" = 3)),
                v2 = structure(c(1,0,0,1), labels = c("foo" = 0, "bar" = 1)))
str(factorize(x))
str(characterize(x))

## Application
csv_file <- tempfile(fileext = ".csv")
## comparison of exported file contents
import(export(x, csv_file))
import(export(factorize(x), csv_file))
}
\seealso{
\code{\link[=gather_attrs]{gather_attrs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rio.R
\docType{package}
\name{rio}
\alias{rio-package}
\alias{rio}
\title{A Swiss-Army Knife for Data I/O}
\description{
The aim of rio is to make data file input and output as easy as possible. \code{\link[=export]{export()}} and \code{\link[=import]{import()}} serve as a Swiss-army knife for painless data I/O for data from almost any file format by inferring the data structure from the file extension, natively reading web-based data sources, setting reasonable defaults for import and export, and relying on efficient data import and export packages. An additional convenience function, \code{\link[=convert]{convert()}}, provides a simple method for converting between file types.

Note that some of rio's functionality is provided by \sQuote{Suggests} dependendencies, meaning they are not installed by default. Use \code{\link[=install_formats]{install_formats()}} to make sure these packages are available for use.
}
\examples{
# export
library("datasets")
export(mtcars, csv_file <- tempfile(fileext = ".csv")) # comma-separated values
export(mtcars, rds_file <- tempfile(fileext = ".rds")) # R serialized
export(mtcars, sav_file <- tempfile(fileext = ".sav")) # SPSS

# import
x <- import(csv_file)
y <- import(rds_file)
z <- import(sav_file)

# convert sav (SPSS) to dta (Stata)
convert(sav_file, dta_file <- tempfile(fileext = ".dta"))

# cleanup
unlink(c(csv_file, rds_file, sav_file, dta_file))

}
\references{
\href{https://cran.r-project.org/package=datamods}{datamods} provides Shiny modules for importing data via \code{rio}.

\href{https://github.com/Stan125/GREA}{GREA} provides an RStudio add-in to import data using rio.
}
\seealso{
\code{\link[=import]{import()}}, \code{\link[=import_list]{import_list()}}, \code{\link[=export]{export()}}, \code{\link[=export_list]{export_list()}}, \code{\link[=convert]{convert()}}, \code{\link[=install_formats]{install_formats()}}
}
\author{
\strong{Maintainer}: Chung-hong Chan \email{chainsawtiney@gmail.com} (\href{https://orcid.org/0000-0002-6232-7530}{ORCID})

Authors:
\itemize{
  \item Jason Becker \email{jason@jbecker.co}
  \item David Schoch \email{david@schochastics.net} (\href{https://orcid.org/0000-0003-2952-4812}{ORCID})
  \item Thomas J. Leeper \email{thosjleeper@gmail.com} (\href{https://orcid.org/0000-0003-4097-6326}{ORCID})
}

Other contributors:
\itemize{
  \item Geoffrey CH Chan \email{gefchchan@gmail.com} [contributor]
  \item Christopher Gandrud [contributor]
  \item Andrew MacDonald [contributor]
  \item Ista Zahn [contributor]
  \item Stanislaus Stadlmann [contributor]
  \item Ruaridh Williamson \email{ruaridh.williamson@gmail.com} [contributor]
  \item Patrick Kennedy [contributor]
  \item Ryan Price \email{ryapric@gmail.com} [contributor]
  \item Trevor L Davis \email{trevor.l.davis@gmail.com} [contributor]
  \item Nathan Day \email{nathancday@gmail.com} [contributor]
  \item Bill Denney \email{wdenney@humanpredictions.com} (\href{https://orcid.org/0000-0002-5759-428X}{ORCID}) [contributor]
  \item Alex Bokov \email{alex.bokov@gmail.com} (\href{https://orcid.org/0000-0002-0511-9815}{ORCID}) [contributor]
  \item Hugo Gruson (\href{https://orcid.org/0000-0002-4094-1476}{ORCID}) [contributor]
  \item Jacob Mears [contributor]
}

}
